// A BDD-based ground reasoning solver for LTL cap ACTL synthesis
//
// (c) 2011/2012 Rüdiger Ehlers

#include "BF.h"
#include "bddDump.h"
#include "BFStateToVariableValuationEncoderClassicAutobit.h"
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>
#include <tr1/unordered_map>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <fstream>
#include <map>
#include <error.h>
#include <tr1/unordered_set>
#include <tr1/unordered_map>
#include <boost/algorithm/string.hpp>
#include "UVW.h"
// #include "md5.hpp"
#include "variableAllocationOptimizer.hpp"
#include "advancedVariableInfoContainer.hpp"

//=============================================================
// Local types for Strategy Extraction
//=============================================================

// Default hasher for BFs
namespace std {
        namespace tr1 {
                template<> struct hash<std::pair<bool,BF> > : public std::unary_function<std::pair<bool,BF> , std::size_t> {
                        std::size_t operator()(const std::pair<bool,BF> &b) const { return b.second.getHashCode() + b.first?0:0x1378393; };
                };
        }
};

class BDDGraphNode {
public:
        std::tr1::unordered_set<std::pair<bool,BF> > incoming;
};

typedef enum { PRE, POST, IN, OUT } BddVarType;


//=============================================================
// Algorithm container class
//=============================================================
class Backend : public AdvancedVariableInfoContainer {
private:

    // Information about the instance (read in)
    enum { Undef, Mealy, Moore } semantics;
    std::vector<std::string> inputAPs;
    std::vector<std::string> outputAPs;
    std::string assumptionUVWUnparsed;
    std::string guaranteeUVWUnparsed;

    // After-parse data for the symbolic representation of the gmae
    UVW assumptions;
    UVW guarantees;

    // Variable groups
    VariableGroup varsInputAPs;
    VariableGroup varsOutputAPs;
    VariableGroup varsInputAPsSecondCopy; // Reserved for copying the respective elements to inputsToBeKeptPre later (if needed)
    VariableGroup varsOutputAPsSecondCopy; // Reserved for copying the respective elements to outputsToBeKeptPre later (if needed)

    VariableGroup varsAssumptionUVWStatesPre;
    VariableGroup varsAssumptionUVWStatesPost;
    VariableGroup varsGuaranteeUVWStatesPre;
    VariableGroup varsGuaranteeUVWStatesPost;
    VariableGroup varsAssumptionCounterPre;
    VariableGroup varsAssumptionCounterPost;
    VariableGroup varsGuaranteeCounterPre;
    VariableGroup varsGuaranteeCounterPost;

    // Stuff for the advanced algorithm for mapping some states into the last I/O
    std::vector<BF> guaranteeStateActivatingIO;
    std::vector<BF> assumptionStateActivatingIO;
    std::map<uint,uint> renumberingOfTranditionallyHandledGuarantees;
    std::map<uint,uint> renumberingOfTranditionallyHandledAssumptions;
    BF bfForInitialNoIncomingEdgesState;
    VariableGroup inputsToBeKeptPre;
    VariableGroup outputsToBeKeptPre;


    // Tool functions
    static std::string toString(int i) {
        std::ostringstream os;
        os << i;
        return os.str();
    }

    BFVarCube computeCube(std::vector<uint> const &vec) {
        std::vector<BF> bfs;
        for (auto it = vec.begin();it!=vec.end();it++) bfs.push_back(vars[*it]);
        return mgr.computeCube(bfs);
    }

    BFVarVector computeVarVector(std::vector<uint> const &vec) {
        std::vector<BF> bfs;
        for (auto it = vec.begin();it!=vec.end();it++) bfs.push_back(vars[*it]);
        return mgr.computeVarVector(bfs);
    }

    BFVarVector computeVarVector(std::vector<uint> const &vec1, std::vector<uint> const &vec2, std::vector<uint> const &vec3, std::vector<uint> const &vec4) {
        std::vector<BF> bfs;
        for (auto it = vec1.begin();it!=vec1.end();it++) bfs.push_back(vars[*it]);
        for (auto it = vec2.begin();it!=vec2.end();it++) bfs.push_back(vars[*it]);
        for (auto it = vec3.begin();it!=vec3.end();it++) bfs.push_back(vars[*it]);
        for (auto it = vec4.begin();it!=vec4.end();it++) bfs.push_back(vars[*it]);
        return mgr.computeVarVector(bfs);
    }

    std::set<uint> computeSupport(std::vector<BF> &supportVars,BF data) {
        BF_DUMPANDCHECK(*this,data,NULL,"/tmp/computeSupport.dot")
        std::set<uint> supportSoFar;
        std::set<BF> todo;
        std::set<BF> done;
        todo.insert(data);
        done.insert(data);
        while (todo.size()>0) {
            BF current = *todo.begin();
            todo.erase(current);
            if (!current.isConstant()) {
                uint currentIndex = current.readNodeIndex();
                bool found = false;
                for (uint j=0;j<supportVars.size();j++) {
                    if (supportVars[j].readNodeIndex()==currentIndex) {
                        supportSoFar.insert(j);
                        found = true;
                        BF one = (current & supportVars[j]).ExistAbstractSingleVar(supportVars[j]);
                        BF two = (current & !supportVars[j]).ExistAbstractSingleVar(supportVars[j]);
                        if (done.count(one)==0) {
                            todo.insert(one);
                            done.insert(one);
                        }
                        if (done.count(two)==0) {
                            todo.insert(two);
                            done.insert(two);
                        }
                    }
                }
                if (!found) {
                    throw "Error in computing the support: Unexpected variable.";
                }
            }
        }
        return supportSoFar;
    }


    //=============================================================================
    // XML processing functions
    //=============================================================================
    std::string xmlNodeGetChildrenText(xmlNode *node) {
        std::ostringstream os;
        for (xmlNode *it3 = node->children; it3!=NULL; it3 = it3->next) {
            if (it3->type == XML_TEXT_NODE) {
                os << it3->content;
            } else if (it3->type == XML_ATTRIBUTE_NODE) {
                // OK
            } else if (it3->type == XML_COMMENT_NODE) {
                // OK
            } else {
                std::cerr << "Error in xmlNodeGetChildrenText(xmlNode *node) : given node contains illegal child of type: " << it3->type << std::endl;
                throw "Aborting";
            }
        }
        return os.str();
    }

    void readAPList(xmlNode *firstNodeInChildList, std::vector<std::string> &whereToPutBits) {
        xmlNode *nextNode = firstNodeInChildList;
        while (nextNode!=NULL) {
            if (nextNode->type==XML_ELEMENT_NODE) {
                // Must be a bit!
                if (xmlStrEqual(nextNode->name,BAD_CAST("Bit"))) {
                    std::string bitName = xmlNodeGetChildrenText(nextNode);
                    boost::trim(bitName);
                    whereToPutBits.push_back(bitName);
                }
            } else if (nextNode->type==XML_COMMENT_NODE) {
                // No problem with comments
            } else if (nextNode->type==XML_TEXT_NODE) {
                std::string text = xmlNodeGetChildrenText(nextNode);
                boost::trim(text);
                if (text.length()!=0) {
                    std::cerr << "Error in procedure 'readAPList' - found some unexpexted text: " << text << std::endl;
                    throw "Aborted.";
                }
            } else {
                std::cerr << "Error in procedure 'readAPList' - expected only comments and Bits!" << std::endl;
                std::cerr << "However, I found a node of type " << nextNode->type << std::endl;
                throw "Aborted.";
            }
            nextNode = nextNode->next;
        }
    }

    xmlChar *readAttribute(xmlNode *node, const char *whichOne) {
        for (xmlAttr *it3 = node->properties;node!=NULL;it3 = it3->next) {
            if (it3->type!=XML_ATTRIBUTE_NODE) {
                throw "Only expected attribute node as result of readAttribute command.";
            } else {
                if (!xmlStrEqual(it3->name,BAD_CAST(whichOne))) {
                    // Is ok.
                } else {
                    xmlChar *c = xmlNodeListGetString(it3->doc, it3->children, true);
                    return c;
                }
            }
        }
        return NULL;
    }

    BF parseTransitionString(std::istringstream &is, std::vector<BF> relevantVars, uint position) {
        // Base case
        if (position == relevantVars.size()) return mgr.constantTrue();

        char c;
        is >> c;
        if (is.fail()) throw "Error in function parseTransitionString - cannot read part of the transition string (1)";
        if (c=='0') {
            return !relevantVars[position] & parseTransitionString(is,relevantVars,position+1);
        } else if (c=='1') {
            return relevantVars[position] & parseTransitionString(is,relevantVars,position+1);
        } else if (c=='*') {
            return parseTransitionString(is,relevantVars,position+1);
        } else if (c=='(') {
            BF a = parseTransitionString(is,relevantVars,position+1);
            char c2;
            is >> c2;
            if (is.fail()) throw "Error in function parseTransitionString - cannot read part of the transition string (2)";
            if (c2!=',') throw"Error in function parseTransitionString - cannot read part of the transition string (3)";
            BF b = parseTransitionString(is,relevantVars,position+1);
            is >> c2;
            if (is.fail()) throw "Error in function parseTransitionString - cannot read part of the transition string (4)";
            if (c2!=')') throw "Error in function parseTransitionString - cannot read part of the transition string (5)";
            return (a & !relevantVars[position]) | (b & relevantVars[position]);
        } else {
            std::cerr << "Found: " << c << "\n";
            throw "Error in function parseTransitionString - cannot read part of the transition string (6)";
        }
    }

public:
    //=============================================================================
    // Synthesis setting-up functions
    //=============================================================================
    void prepareGame() {
        // Allocate atomic propositions
        for (auto it = inputAPs.begin();it!=inputAPs.end();it++) {
            vars.push_back(mgr.newVariable());
            varsInputAPs.add(varNames.size());
            varNames.push_back("in_"+*it);
            vars.push_back(mgr.newVariable());
            varsInputAPsSecondCopy.add(varNames.size());
            varNames.push_back("in_"+*it+"_keeppre");
        }
        for (auto it = outputAPs.begin();it!=outputAPs.end();it++) {
            vars.push_back(mgr.newVariable());
            varsOutputAPs.add(varNames.size());
            varNames.push_back("out_"+*it);
            vars.push_back(mgr.newVariable());
            varsOutputAPsSecondCopy.add(varNames.size());
            varNames.push_back("out_"+*it+"_keeppre");
        }

        // Parse UVWs - Both assumptions and guarantees
        for (uint i=0;i<2;i++) {
            std::string &src = (i==0)?assumptionUVWUnparsed:guaranteeUVWUnparsed;
            UVW &dest = (i==0)?assumptions:guarantees;
            std::istringstream is(src);
            std::cerr << "SRC UVW: " << src << "\n";

            // Read atomic propositions
            int nofAPs;
            is >> nofAPs;
            if (is.fail()) throw "Error while obtaining #APs from UVW";
            std::vector<BF> theseAPs;
            for (int i=0;i<nofAPs;i++) {
                std::string apName;
                std::string originalAPName;
                is >> apName;
                if (is.fail()) throw "Error while obtaining AP name from UVW";
                if (apName[0]!='v') throw "Error while obtaining AP name from UVW: Assumption that all variable names start with 'v' is not satisfied!";
                if ((apName.length()% 2)!=1) throw "Error while obtaining AP name from UVW: Assumption that all variable names start with 'v' is not satisfied (case 2)!";
                for (uint pos=1;pos<apName.length();pos+=2) {
                    size_t index1 = std::string("0123456789abcdef").find(apName[pos]);
                    if (index1==std::string::npos) throw "Error while obtaining AP name from UVW: Assumption that all variable names are hex-encoded is not fulfilled (case 1)!";
                    size_t index2 = std::string("0123456789abcdef").find(apName[pos+1]);
                    if (index2==std::string::npos) throw "Error while obtaining AP name from UVW: Assumption that all variable names are hex-encoded is not fulfilled (case 2)!";
                    originalAPName += char(index1*16+index2);
                }
                std::cerr << "Using AP " << originalAPName << "\n";
                // Find atomic proposition in the list
                bool found = false;
                for (uint j=0;(j<inputAPs.size()) && !found;j++) {
                    if (originalAPName == inputAPs[j]) {
                        found = true;
                        theseAPs.push_back(varsInputAPs.getVarBF(j));
                    }
                }
                for (uint j=0;(j<outputAPs.size()) && !found;j++) {
                    if (originalAPName == outputAPs[j]) {
                        found = true;
                        theseAPs.push_back(varsOutputAPs.getVarBF(j));
                    }
                }
                if (!found) {
                    std::cerr << "Error while parsing automata: the atomic proposition '" << originalAPName << "' used in the specification is neither listed as input nor output atomic proposition.\n";
                    throw "Aborting";
                }
            }

            // Read states
            assert(dest.states.size()==0);
            int nofStates;
            is >> nofStates;
            if (is.fail()) throw "Error: Cannot read number of states from UVW.";
            dest.states.resize(nofStates);
            for (int i=0;i<nofStates;i++) {

                // IsInitial?
                int isInitial;
                is >> isInitial;
                if (is.fail()) throw "Error while parsing automata: Cannot parse isInitial flag";
                if (isInitial==0) {
                    dest.states[i].isInitial = false;
                } else if (isInitial==1) {
                    dest.states[i].isInitial = true;
                } else {
                    std::cerr << "Found: " << isInitial << std::endl;
                    throw "Error while parsing automata: Illegal isInitial value";
                }

                int isFinal;
                is >> isFinal;
                if (is.fail()) throw "Error while parsing automata: Cannot parse isFinal flag";
                if (isFinal==0) {
                    dest.states[i].isFinal = false;
                } else if (isFinal==1) {
                    dest.states[i].isFinal = true;
                } else {
                    std::cerr << "Found: " << isFinal << std::endl;
                    throw "Error while parsing automata: Illegal isFinal value";
                }

                //Read transitions
                uint nofTransitions;
                is >> nofTransitions;
                if (is.fail()) throw "Error: Cannot read number of transitions in a UVW state.";
                for (uint j=0;j<nofTransitions;j++) {
                    uint targetNr;
                    is >> targetNr;
                    if (is.fail()) throw "Error: cannot read transition target number";
                    dest.states[i].transitions.push_back(std::pair<uint,BF>(targetNr,parseTransitionString(is,theseAPs,0)));
                }
            }
        }
    }

    void allocatePairOfVariables(std::string preName, VariableGroup &preTargetVector, VariableGroup &postTargetVector) {
        vars.push_back(mgr.newVariable());
        preTargetVector.add(varNames.size());
        varNames.push_back(preName);
        vars.push_back(mgr.newVariable());
        postTargetVector.add(varNames.size());
        varNames.push_back(preName+"'");
        std::vector<BF> group;
        group.push_back(vars[vars.size()-2]);
        group.push_back(vars[vars.size()-1]);
        mgr.groupVariables(group);
    }

    // This procedure fills a transition-relation-computation-operations-to-perform-list.
    // The procedure is only called from the main synthesis procedure
    void fillTransitionOperationWithAutomatonTransitionsAssumptions(std::vector<std::pair<bool, BF> > &transitionOperations, std::vector<uint> rejectingStateEncoding) {
        UVW &automaton = assumptions;
        VariableGroup &varsCounterEncodingPre = varsAssumptionCounterPre;
        VariableGroup &varsCounterEncodingPost = varsAssumptionCounterPost;
        VariableGroup &varsStateBitsPre = varsAssumptionUVWStatesPre;
        VariableGroup &varsStateBitsPost = varsAssumptionUVWStatesPost;

        // First post-bit represents whether the state represented by the pre-counter has been left
        BF cases = mgr.constantFalse();
        BF matchCases = mgr.constantFalse();
        for (uint i=0;i<rejectingStateEncoding.size();i++) {
            uint currentState = rejectingStateEncoding[i];
            assert(currentState<automaton.states.size());
            BF thisCase = mgr.constantTrue();
            for (uint j=0;j<varsCounterEncodingPre.size()-1;j++) {
                if (((i >> j)&1)!=0) {
                    thisCase &= varsCounterEncodingPre.getVarBF(j+1);
                } else {
                    thisCase &= !varsCounterEncodingPre.getVarBF(j+1);
                }
            }
            matchCases |= thisCase;
            cases |= thisCase & (!automaton.getSelfLoop(currentState) | !varsStateBitsPre.getVarBF(currentState));
        }
        cases |= !matchCases;

        transitionOperations.push_back(std::pair<bool,BF>(false,!(cases ^ varsCounterEncodingPost.getVarBF(0))));

        // Compute post-states
        // Perform a topological sorting on-the-fly
        std::set<uint> doneList;
        uint round = 0;
        while (doneList.size()!=renumberingOfTranditionallyHandledAssumptions.size()) {

            round++;
            if ((doneList.size()+5)<round) throw "Error: One of the UVWs is not very weak.";

            // Print
            //std::cerr << "sizeAut: " << automaton.states.size() << ": ";
            //for (auto it = doneList.begin();it!=doneList.end();it++) {
            //    std::cerr << *it << " ";
            //}
            //std::cerr << std::endl;

            // Process automata that aren't already bottom-most
            for (uint i=0;i<automaton.states.size();i++) {
                if (renumberingOfTranditionallyHandledAssumptions.count(i)>0) {
                    // Check if state is bottom
                    bool bottom = true;
                    for (auto it = automaton.states[i].transitions.begin();it!=automaton.states[i].transitions.end();it++) {
                        bottom &= ((it->first == i) || (doneList.count(it->first)>0) || (renumberingOfTranditionallyHandledAssumptions.count(it->first)==0));
                    }
                    if (bottom && (doneList.count(i)==0)) {
                        BF cases = mgr.constantFalse();
                        // Get incoming edges
                        for (uint j=0;j<automaton.states.size();j++) {
                            for (auto it = automaton.states[j].transitions.begin();it!=automaton.states[j].transitions.end();it++) {
                                if (it->first == i) {
                                    if (renumberingOfTranditionallyHandledAssumptions.count(j)>0) {
                                        cases |= varsStateBitsPre.getVarBF(renumberingOfTranditionallyHandledAssumptions[j]) & it->second;
                                    } else {
                                        cases |= (!bfForInitialNoIncomingEdgesState) & assumptionStateActivatingIO[j] & it->second;
                                    }
                                }
                            }
                        }
                        transitionOperations.push_back(std::pair<bool,BF>(false,!(cases ^ varsStateBitsPost.getVarBF(i))));
                        transitionOperations.push_back(std::pair<bool,BF>(true,varsStateBitsPost.getVarBF(i)));
                        doneList.insert(i);

                    }
                }
            }
        }

        // For the assumptions, the counter is (possibly) reset during the move of the system player. In every case, the post bits are abstracted from

        for (uint i=1;i<varsCounterEncodingPost.size();i++) {
            transitionOperations.push_back(std::pair<bool,BF>(false,varsCounterEncodingPost.getVarBF(0) | !(varsCounterEncodingPost.getVarBF(i) ^ varsCounterEncodingPre.getVarBF(i))));
            transitionOperations.push_back(std::pair<bool,BF>(true,varsCounterEncodingPost.getVarBF(i)));
        }
        transitionOperations.push_back(std::pair<bool,BF>(true,varsCounterEncodingPost.getVarBF(0)));
    }

    // This procedure fills a transition-relation-computation-operations-to-perform-list.
    // The procedure is only called from the main synthesis procedure
    void fillTransitionOperationWithAutomatonTransitionsGuarantees(std::vector<std::pair<bool, BF> > &transitionOperations, std::vector<uint> rejectingStateEncoding) {
        UVW &automaton = guarantees;
        VariableGroup &varsCounterEncodingPre = varsGuaranteeCounterPre;
        VariableGroup &varsCounterEncodingPost = varsGuaranteeCounterPost;
        VariableGroup &varsStateBitsPre = varsGuaranteeUVWStatesPre;
        VariableGroup &varsStateBitsPost = varsGuaranteeUVWStatesPost;

        // First post-bit represents whether the state represented by the pre-counter has been left
        BF cases = mgr.constantFalse();
        BF matchCases = mgr.constantFalse();
        for (uint i=0;i<rejectingStateEncoding.size();i++) {
            uint currentState = rejectingStateEncoding[i];
            assert(currentState<automaton.states.size());
            BF thisCase = mgr.constantTrue();
            for (uint j=0;j<varsCounterEncodingPre.size()-1;j++) {
                if (((i >> j)&1)!=0) {
                    thisCase &= varsCounterEncodingPre.getVarBF(j+1);
                } else {
                    thisCase &= !varsCounterEncodingPre.getVarBF(j+1);
                }
            }
            matchCases |= thisCase;
            cases |= thisCase & (!automaton.getSelfLoop(currentState) | !varsStateBitsPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[currentState]));
        }
        cases |= !matchCases;

        // Guarantee special case: Retain pre-bit
        cases |= varsCounterEncodingPre.getVarBF(0);
        transitionOperations.push_back(std::pair<bool,BF>(false,!(cases ^ varsCounterEncodingPost.getVarBF(0))));

        // Compute post-states
        // Perform a topological sorting on-the-fly
        std::set<uint> doneList;
        uint round = 0;
        while (doneList.size()!=renumberingOfTranditionallyHandledGuarantees.size()) {

            round++;
            if ((doneList.size()+5)<round) throw "Error: One of the UVWs is not very weak.";

            // Print
            //std::cerr << "sizeAut: " << automaton.states.size() << ": ";
            //for (auto it = doneList.begin();it!=doneList.end();it++) {
            //    std::cerr << *it << " ";
            //}
            //std::cerr << std::endl;

            // Process automata that aren't already bottom-most
            for (uint i=0;i<automaton.states.size();i++) {
                if (renumberingOfTranditionallyHandledGuarantees.count(i)>0) {
                    // Check if state is bottom
                    bool bottom = true;
                    for (auto it = automaton.states[i].transitions.begin();it!=automaton.states[i].transitions.end();it++) {
                        bottom &= ((it->first == i) || (doneList.count(it->first)>0) || (renumberingOfTranditionallyHandledGuarantees.count(it->first)==0));
                    }
                    if (bottom && (doneList.count(i)==0)) {
                        BF cases = mgr.constantFalse();
                        // Get incoming edges
                        for (uint j=0;j<automaton.states.size();j++) {
                            for (auto it = automaton.states[j].transitions.begin();it!=automaton.states[j].transitions.end();it++) {
                                if (it->first == i) {
                                    if (renumberingOfTranditionallyHandledGuarantees.count(j)>0) {
                                        cases |= varsStateBitsPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[j]) & it->second;
                                    } else {
                                        cases |= (!bfForInitialNoIncomingEdgesState) & guaranteeStateActivatingIO[j] & it->second;
                                    }
                                }
                            }
                        }
                        transitionOperations.push_back(std::pair<bool,BF>(false,!(cases ^ varsStateBitsPost.getVarBF(renumberingOfTranditionallyHandledGuarantees[i]))));
                        transitionOperations.push_back(std::pair<bool,BF>(true,varsStateBitsPost.getVarBF(renumberingOfTranditionallyHandledGuarantees[i])));
                        doneList.insert(i);
                    }
                }
            }
        }

        // For the assumptions, the counter is (possibly) reset during the move of the system player. In every case, the post bits are abstracted from
        for (uint i=0;i<varsCounterEncodingPost.size();i++) {
            // Case 0 is already contained in the case above
            if (i!=0) transitionOperations.push_back(std::pair<bool,BF>(false,!(varsCounterEncodingPost.getVarBF(i) ^ varsCounterEncodingPre.getVarBF(i))));
            transitionOperations.push_back(std::pair<bool,BF>(true,varsCounterEncodingPost.getVarBF(i)));
        }
    }


    //======================================================
    // Main synthesis function
    //
    // Uses an advanced optimization technique that tries to map
    // states to last input/output to perform well on "classical GR(1) specifications"
    // and those that have many such components.
    //======================================================
    void performSynthesis(bool classico, bool noImplementationBuilding) {

        // Analyse for which states it is reasonable to use the previoux input/output for checking if we are in a state of not.
        // - first, check if we have a state that tells us if we are initial
        // also find on-true self looping initial states
        int stateThatIsOnlyInitial = -1;
        bool stateThatIsOnlyInitialIsAssumption = false;
        std::set<std::pair<uint,bool /* IsAssumption */> > onTrueSelfLoopingInitialStates;
        for (uint i=0;i<assumptions.states.size();i++) {
            if (assumptions.states[i].isInitial) {
                bool isOk = true;
                bool selfLoopsOnTrue = false;
                bool onlySelfLoops = true;
                for (uint k=0;k<assumptions.states.size();k++) {
                    for (uint j=0;j<assumptions.states[k].transitions.size();j++) {
                        isOk &= assumptions.states[k].transitions[j].first!=i;
                        onlySelfLoops &= (assumptions.states[k].transitions[j].first!=i) || (i==k);
                        if (i==k) {
                            if ((assumptions.states[i].transitions[j].first==i) && (assumptions.states[i].transitions[j].second==mgr.constantTrue())) {
                                selfLoopsOnTrue = true;
                            }
                        }
                    }
                }

                if (isOk) {
                    stateThatIsOnlyInitialIsAssumption = true;
                    stateThatIsOnlyInitial = i;
                    std::cerr << "IsOnlyInitial: Asssumption " << i << std::endl;
                }
                if (selfLoopsOnTrue && onlySelfLoops) {
                    onTrueSelfLoopingInitialStates.insert(std::pair<uint,bool>(i,true));
                    std::cerr << "Self-loop initial state: Assumption " << i << std::endl;
                }
            }
        }
        for (uint i=0;i<guarantees.states.size();i++) {
            if (guarantees.states[i].isInitial) {
                bool isOk = true;
                bool selfLoopsOnTrue = false;
                bool onlySelfLoops = true;
                for (uint k=0;k<guarantees.states.size();k++) {
                    for (uint j=0;j<guarantees.states[k].transitions.size();j++) {
                        isOk &= guarantees.states[k].transitions[j].first!=i;
                        onlySelfLoops &= (guarantees.states[k].transitions[j].first!=i) || (i==k);
                        if (i==k) {
                            if ((guarantees.states[i].transitions[j].first==i) && (guarantees.states[i].transitions[j].second==mgr.constantTrue())) {
                                selfLoopsOnTrue = true;
                            }
                        }
                    }

                }

                if (isOk) {
                    stateThatIsOnlyInitialIsAssumption = false;
                    stateThatIsOnlyInitial = i;
                    std::cerr << "IsOnlyInitial: Guantees " << i << std::endl;
                }
                if (selfLoopsOnTrue && onlySelfLoops) {
                    onTrueSelfLoopingInitialStates.insert(std::pair<uint,bool>(i,false));
                    std::cerr << "Self-loop initial state: Guantees " << i << std::endl;
                }
            }
        }

        // Analyse connections between inputs/outputs and states - prepare I/O variables
        if (classico) stateThatIsOnlyInitial = -1;
        std::vector<BF> varsInputAndOutputs;
        for (uint i=0;i<varsInputAPs.size();i++) varsInputAndOutputs.push_back(varsInputAPs.getVarBF(i));
        for (uint i=0;i<varsOutputAPs.size();i++) varsInputAndOutputs.push_back(varsOutputAPs.getVarBF(i));
        VariableAllocationOptimizer variableAllocationOptimizer(varsInputAndOutputs.size(),1,1);

        // Analyse connections between inputs/outputs and states - add transitions
        for (uint i=0;i<assumptions.states.size();i++) {
            BF transitionsFromSelfLoopingInitialState = mgr.constantFalse();

            // Only mark states as candidate for which there are no other incoming edges
            // ...also only do so if we have a designated state that is only active in the initial round.
            bool isCandidate = stateThatIsOnlyInitial >= 0;

            for (uint j=0;j<assumptions.states.size();j++) {
                if (onTrueSelfLoopingInitialStates.count(std::pair<uint,bool>(j,true))==0) {
                    for (auto it = assumptions.states[j].transitions.begin();it!=assumptions.states[j].transitions.end();it++) {
                        if ((it->first==i) && (it->second != mgr.constantFalse()))
                            isCandidate = false;
                    }
                }
            }
            if (isCandidate) {
                for (auto it = onTrueSelfLoopingInitialStates.begin();it!=onTrueSelfLoopingInitialStates.end();it++) {
                    if ((it->second)) {
                        for (uint j=0;j<assumptions.states[it->first].transitions.size();j++) {
                            if (assumptions.states[it->first].transitions[j].first==i) {
                                transitionsFromSelfLoopingInitialState |= assumptions.states[it->first].transitions[j].second;
                            }
                        }
                    }
                }
                assumptionStateActivatingIO.push_back(transitionsFromSelfLoopingInitialState);
            } else {
                assumptionStateActivatingIO.push_back(BF());
            }
            std::cerr << "Assumption self-loop incoming edge: " << i << ", support: ";
            std::set<uint> varSupport = computeSupport(varsInputAndOutputs,transitionsFromSelfLoopingInitialState);
            for (auto it = varSupport.begin();it!=varSupport.end();it++) {
                std::cerr << *it << " ";
            }
            std::cerr << std::endl;
            variableAllocationOptimizer.addCombination(varSupport);
        }
        for (uint i=0;i<guarantees.states.size();i++) {
            BF transitionsFromSelfLoopingInitialState = mgr.constantFalse();

            // Only mark states as candidate for which there are no other incoming edges
            // ...also only do so if we have a designated state that is only active in the initial round.
            bool isCandidate = stateThatIsOnlyInitial >= 0;

            for (uint j=0;j<guarantees.states.size();j++) {
                if (onTrueSelfLoopingInitialStates.count(std::pair<uint,bool>(j,false))==0) {
                    for (auto it = guarantees.states[j].transitions.begin();it!=guarantees.states[j].transitions.end();it++) {
                        if ((it->first==i) && (it->second != mgr.constantFalse()))
                            isCandidate = false;
                    }
                }
            }
            if (isCandidate) {
                for (auto it = onTrueSelfLoopingInitialStates.begin();it!=onTrueSelfLoopingInitialStates.end();it++) {
                    if (!(it->second)) {
                        for (uint j=0;j<guarantees.states[it->first].transitions.size();j++) {
                            if (guarantees.states[it->first].transitions[j].first==i) {
                                transitionsFromSelfLoopingInitialState |= guarantees.states[it->first].transitions[j].second;
                            }
                        }
                    }
                }
                guaranteeStateActivatingIO.push_back(transitionsFromSelfLoopingInitialState);
            } else {
                guaranteeStateActivatingIO.push_back(BF());
            }
            std::cerr << "Guarantee self-loop incoming edge: " << i << ", support: ";
            std::set<uint> varSupport = computeSupport(varsInputAndOutputs,transitionsFromSelfLoopingInitialState);
            for (auto it = varSupport.begin();it!=varSupport.end();it++) {
                std::cerr << *it << " ";
            }
            std::cerr << std::endl;
            variableAllocationOptimizer.addCombination(varSupport);
        }

        // Solve the variable allocation
        variableAllocationOptimizer.solve();
        for (uint i=0;i<varsInputAndOutputs.size();i++) {
            std::cerr << "Choice: " << variableAllocationOptimizer.getSolutionVariables()[i] << std::endl;
        }

        // Mark which inputs/outputs/guarantee states/assumption states take
        // part in the sate-encoding-into-i/o-scehem
        std::vector<bool> inputsToBeKept = variableAllocationOptimizer.getSolutionVariables();
        std::vector<bool> outputsToBeKept;
        VariableGroup inputsToBeKeptPost(this);
        VariableGroup outputsToBeKeptPost(this);
        for (uint i=0;i<varsInputAPs.size();i++) {
            if (inputsToBeKept[i]) {
                inputsToBeKeptPost.add(varsInputAPs[i]);
                vars.push_back(mgr.newVariable());
                inputsToBeKeptPre.add(varsInputAPsSecondCopy[i]);
                // varNames.push_back(varNames[varsInputAPs[i]]+"_keeppre");
                std::vector<BF> group;
                group.push_back(varsInputAPs.getVarBF(i));
                group.push_back(varsInputAPsSecondCopy.getVarBF(i));
                // mgr.groupVariables(group);
            }
        }
        for (uint i=0;i<varsOutputAPs.size();i++) outputsToBeKept.push_back(inputsToBeKept[varsInputAPs.size()+i]);
        for (uint i=0;i<varsOutputAPs.size();i++) {
            if (outputsToBeKept[i]) {
                outputsToBeKeptPost.add(varsOutputAPs[i]);
                vars.push_back(mgr.newVariable());
                outputsToBeKeptPre.add(varsOutputAPsSecondCopy[i]);
                // varNames.push_back(varNames[varsOutputAPs[i]]+"_keeppre");
                std::vector<BF> group;
                group.push_back(varsOutputAPs.getVarBF(i));
                group.push_back(varsOutputAPsSecondCopy.getVarBF(i));
                // mgr.groupVariables(group);
            }
        }

        // Now make the "traditional" states are at the beginning.
        for (uint i=0;i<assumptions.states.size();i++) {
            bool encodeIntoLastIO = true;
            std::set<uint> support = variableAllocationOptimizer.getCombination(i);
            encodeIntoLastIO &= (support.size()>0);
            for (auto it = support.begin();it!=support.end();it++) {
                encodeIntoLastIO &= variableAllocationOptimizer.getSolutionVariables()[*it];
            }
            if (!encodeIntoLastIO) {
                uint n = renumberingOfTranditionallyHandledAssumptions.size();
                renumberingOfTranditionallyHandledAssumptions[i] = n;
                std::cerr << "R-A: " << i << ": " << renumberingOfTranditionallyHandledAssumptions[i] << std::endl;
            }
        }

        for (uint i=0;i<guarantees.states.size();i++) {
            bool encodeIntoLastIO = true;
            std::set<uint> support = variableAllocationOptimizer.getCombination(i+assumptions.states.size());
            encodeIntoLastIO &= (support.size()>0);
            for (auto it = support.begin();it!=support.end();it++) {
                encodeIntoLastIO &= variableAllocationOptimizer.getSolutionVariables()[*it];
            }
            if (!encodeIntoLastIO) {
                uint n = renumberingOfTranditionallyHandledGuarantees.size();
                renumberingOfTranditionallyHandledGuarantees[i] = n;
                // std::cerr << "R-G: " << i << ": " << renumberingOfTranditionallyHandledGuarantees[i] << std::endl;
            }
        }

        // Allocate pre and Post-variables
        for (uint i=0;i<assumptions.states.size();i++) {
            if (renumberingOfTranditionallyHandledAssumptions.count(i)>0) {
                std::ostringstream os;
                os << "assumption_" << i;
                allocatePairOfVariables(os.str(),varsAssumptionUVWStatesPre,varsAssumptionUVWStatesPost);
            }
        }
        for (uint i=0;i<guarantees.states.size();i++) {
            if (renumberingOfTranditionallyHandledGuarantees.count(i)>0) {
                std::ostringstream os;
                os << "guarantee_" << i;
                allocatePairOfVariables(os.str(),varsGuaranteeUVWStatesPre,varsGuaranteeUVWStatesPost);
            }
        }

        // Compute a BF (over pre-variables) that marks states that are initial and only initial
        if (stateThatIsOnlyInitial>-1) {
            if (stateThatIsOnlyInitialIsAssumption) {
                bfForInitialNoIncomingEdgesState = varsAssumptionUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledAssumptions[stateThatIsOnlyInitial]);
            } else {
                //std::cerr << "stateThatIsOnlyInitial: " << stateThatIsOnlyInitial << std::endl;
                //std::cerr << "Count: " << renumberingOfTranditionallyHandledGuarantees[stateThatIsOnlyInitial] << std::endl;
                //std::cerr << "Max: " << varsGuaranteeUVWStatesPre.size() << "," << renumberingOfTranditionallyHandledGuarantees.size() << std::endl;
                bfForInitialNoIncomingEdgesState = varsGuaranteeUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[stateThatIsOnlyInitial]);
            }
        }

        // Convert post-i/o in the guaranteeStateActivatingIO and AssumptionStateActivatingIO to pre-i/o
        for (uint i = 0;i<guaranteeStateActivatingIO.size();i++) {
            if (guaranteeStateActivatingIO[i].isValid()) {
                guaranteeStateActivatingIO[i] = guaranteeStateActivatingIO[i].SwapVariables(inputsToBeKeptPre.getVector(),inputsToBeKeptPost.getVector());
                guaranteeStateActivatingIO[i] = guaranteeStateActivatingIO[i].SwapVariables(outputsToBeKeptPre.getVector(),outputsToBeKeptPost.getVector());
                BF_DUMPANDCHECK(*this,guaranteeStateActivatingIO[i],NULL,"/tmp/guaIO.dot")
            }
        }
        for (uint i = 0;i<assumptionStateActivatingIO.size();i++) {
            if (assumptionStateActivatingIO[i].isValid()) {
                assumptionStateActivatingIO[i] = assumptionStateActivatingIO[i].SwapVariables(inputsToBeKeptPre.getVector(),inputsToBeKeptPost.getVector());
                assumptionStateActivatingIO[i] = assumptionStateActivatingIO[i].SwapVariables(outputsToBeKeptPre.getVector(),outputsToBeKeptPost.getVector());
                BF_DUMPANDCHECK(*this,assumptionStateActivatingIO[i],NULL,"/tmp/guaIO.dot")
            }
        }


        // Allocate pre/post for the obligation counters - Reserve an additional counter for
        // whether the counter has just changed or the
        std::vector<uint> assumptionRejectingStates;
        std::vector<uint> guaranteeRejectingStates;
        for (uint i=0;i<assumptions.states.size();i++) {
            if (assumptions.states[i].isFinal) {
                assumptionRejectingStates.push_back(i);
                if (renumberingOfTranditionallyHandledAssumptions.count(i)==0) {
                    throw "Internal Error: Transient assumption states must not be rejecting!";
                }
            }
        }
        for (uint i=0;i<guarantees.states.size();i++) {
            if (guarantees.states[i].isFinal) {
                guaranteeRejectingStates.push_back(i);
                if (renumberingOfTranditionallyHandledGuarantees.count(i)==0) {
                    throw "Inernal Error: Transient guarantee states must not be rejecting!";
                }
            }
        }
        std::cerr << "NofAssumptionRejectingStates: " << assumptionRejectingStates.size() << "\n";
        allocatePairOfVariables("assumptionCounter_0",varsAssumptionCounterPre,varsAssumptionCounterPost);
        for (int i=assumptionRejectingStates.size()-1;i>0;i = i >> 1) {
            std::ostringstream os;
            os << "assumptionCounter_" << varsAssumptionCounterPre.size();
            allocatePairOfVariables(os.str(),varsAssumptionCounterPre,varsAssumptionCounterPost);
        }
        std::cerr << "NofGuaranteeRejectingStates: " << guaranteeRejectingStates.size() << "\n";
        allocatePairOfVariables("guaranteeCounter_0",varsGuaranteeCounterPre,varsGuaranteeCounterPost);
        for (int i=guaranteeRejectingStates.size()-1;i>0;i = i >> 1) {
            std::ostringstream os;
            os << "guaranteeCounter_" << varsGuaranteeCounterPre.size();
            allocatePairOfVariables(os.str(),varsGuaranteeCounterPre,varsGuaranteeCounterPost);
        }

        // Compute Conjuncts and existential abstractions for next-state relation.
        // We pre-compute a list of operations to perform.
        // The include two types of operations: & and ExistAbstract
        // We use pairs in which the first element states whether it is & (false) or ExistAbstract (true)
        std::vector<std::pair<bool, BF> > transitionOperations;
        fillTransitionOperationWithAutomatonTransitionsAssumptions(transitionOperations,assumptionRejectingStates);
        fillTransitionOperationWithAutomatonTransitionsGuarantees(transitionOperations,guaranteeRejectingStates);

        // Debug Check
#ifndef NDEBUG
        int transitionOperationDumpDotNumber = 0;
        for (auto it = transitionOperations.begin();it!=transitionOperations.end();it++) {
            std::ostringstream os;
            os << "/tmp/transitionOperation" << transitionOperationDumpDotNumber++ << "-";
            if (it->first) os << "ExistAbstract";
            else os << "And";
            os << ".dot";
            BF_DUMPANDCHECK(*this,it->second,"guaranteeCounterPre guaranteeUVWStatesPre assumptionUVWStatesPre assumptionCounterPre inputsToBeKeptPre outputsToBeKeptPre inputAPs outputAPs assumptionCounterPost guaranteeCounterPost assumptionUVWStatesPost guaranteeUVWStatesPost",os.str())
        }
#endif

        // Sanity check
        assert(vars.size()==varNames.size());

        // Perform fixed point computation
        //
        // Formula: mu X_2. nu X_1. mu X_0.
        //          vee_{i=0,i=1,i=2} (col_i & V_0 & <>X_i) || (col_i & V_1 & []X_i)
        //
        // See Dawar and Graedel: The descriptive complexity of parity games
        //
        // Sys = Player 0 - That player always has colour 0.
        //                  Thus, at the outer levels, we do not need to compute
        //                  (col_i & V_1 & <>X_i) for i>0
        // Thus, we compute:
        // Formula: nu X_2. mu X_1. nu X_0.
        //          (col_2 & V_0 & []X_2) || (col_1 & V_0 & []X_1) || (col_0 & V_0 & []X_0) || (col_0 & V_1 & <>X_0)

        if (semantics!=Mealy) throw "Only Mealy-type semantics supported so far.";
        BF outerFixedPointEnv = mgr.constantTrue();
        BF outerFixedPointSys = mgr.constantTrue();
        BF outerFixedPointPureEnv = mgr.constantTrue(); // Only the ones with <> color_2
        BF middleFixedPointEnv = mgr.constantFalse();
        BF middleFixedPointSys = mgr.constantFalse();
        BF middleFixedPointPureEnv = mgr.constantFalse(); // Only the ones with <> color_1
        BF innerFixedPointEnv = mgr.constantTrue();
        BF innerFixedPointPureEnv = mgr.constantTrue(); // Only the ones with <> color_0
        BF innerFixedPointSys = mgr.constantTrue();

        BF outerFixedPointEnvOld;
        BF lastSetOfSystemStates = mgr.constantTrue();

        // Determine initial state
        BF initialStatePre = mgr.constantTrue();
        for (uint i=0;i<assumptions.states.size();i++) {
            if (renumberingOfTranditionallyHandledAssumptions.count(i)>0) {
                initialStatePre &= (assumptions.states[i].isInitial)?varsAssumptionUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledAssumptions[i]):!varsAssumptionUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledAssumptions[i]);
            } else {
                assert(!(assumptions.states[i].isInitial));
            }
        }
        for (uint i=0;i<guarantees.states.size();i++) {
            if (renumberingOfTranditionallyHandledGuarantees.count(i)>0) {
                initialStatePre &= (guarantees.states[i].isInitial)?varsGuaranteeUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[i]):!varsGuaranteeUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[i]);
            } else {
                assert(!(guarantees.states[i].isInitial));
            }
        }
        BF_DUMPANDCHECK(*this,initialStatePre,"assumptionUVWStatesPre guaranteeUVWStatesPre","/tmp/initialState.dot");

        // Prepare variable vectors
        BFVarVector gameSolvingSwapVectorPre;
        BFVarVector gameSolvingSwapVectorPost;

        {
            std::vector<BF> pre;
            std::vector<BF> post;

#define ADD_ALL(a,b) for (auto it = b.begin();it!=b.end();it++) a.push_back(vars[*it]);
            ADD_ALL(pre,varsAssumptionUVWStatesPre)
            ADD_ALL(pre,varsGuaranteeUVWStatesPre)
            ADD_ALL(pre,varsAssumptionCounterPre)
            ADD_ALL(pre,varsGuaranteeCounterPre)
            ADD_ALL(post,varsAssumptionUVWStatesPost)
            ADD_ALL(post,varsGuaranteeUVWStatesPost)
            ADD_ALL(post,varsAssumptionCounterPost)
            ADD_ALL(post,varsGuaranteeCounterPost)

            ADD_ALL(pre,inputsToBeKeptPre)
            ADD_ALL(post,inputsToBeKeptPost)
            ADD_ALL(pre,outputsToBeKeptPre)
            ADD_ALL(post,outputsToBeKeptPost)
#undef ADD_ALL

            gameSolvingSwapVectorPre = mgr.computeVarVector(pre);
            gameSolvingSwapVectorPost = mgr.computeVarVector(post);
            assert(pre.size()==post.size());
        }



        // Outermost fixed point
        //
        // The change of the guarantee counter is done by the environment,
        // The change of the assumption counter is performed by the system.
        //
        // Thus, The environment nodes have only pre variables and no input/outputs
        //       The System nodes have preAssumptionCounter, postGuaranteeCounter, inputs
        //
        // Note that there are two ways to set the change-bit in the guarantee counter:
        // Changing the guarantee (in the Env->Sys transition) and satisfying it (in the Sys->Env transition)
        //
        // Thus, after applying the transition relation in the Sys->Env transition, we must swap Pre and Post here.
        // The environment player may always reset the guarantee changed-Bit if no change has occurred.
#ifndef NDEBUG
        static const char *SYSTEM_POSITION_BF_SUPPORT_VARIABLES = "guaranteeCounterPre inputsToBeKeptPre outputsToBeKeptPre inputAPs assumptionUVWStatesPre guaranteeUVWStatesPre assumptionCounterPre";
        static const char *ENVIRONMENT_POSITION_BF_SUPPORT_VARIABLES = "inputsToBeKeptPre outputsToBeKeptPre assumptionUVWStatesPre guaranteeUVWStatesPre assumptionCounterPre guaranteeCounterPre";
#endif
        //BFVarVector varVectorAllPre = computeVarVector(varsAssumptionUVWStatesPre,varsGuaranteeUVWStatesPre,varsAssumptionCounterPre,varsGuaranteeCounterPre,);
        //BFVarVector varVectorAllPost = computeVarVector(varsAssumptionUVWStatesPost,varsGuaranteeUVWStatesPost,varsAssumptionCounterPost,varsGuaranteeCounterPost);

        // Compute reachable states
#ifndef NO_COMPUTATION_OF_REACHABLE_STATES
        BF reachableStates = initialStatePre;
        BF oldReachableStates = mgr.constantFalse();
        while (reachableStates!=oldReachableStates) {


#ifdef DEBUG_REACHABLE_STATES
        {
                    uint nofBitsToConsider = varsAssumptionCounterPre.size() + varsGuaranteeCounterPre.size() /*+ inputAPs.size()*/ + assumptions.states.size() + guarantees.states.size();
                    if (nofBitsToConsider > 30) throw "Too many states for this debugging version!";
                    std::ostringstream occurrenceString;
                    std::cerr << "#Bits: " << nofBitsToConsider << "\n";
                    // for (uint i=0;i<(1 <<  nofBitsToConsider);i++) {
                    for (uint i=17761;i<=17761;i++) {

                        BF candidate = reachableStates;
                        uint thisElement = i;
                        // if ((i%1000000)==0) std::cerr << ".";

                        // Process inputAPs
                        /*for (auto it = varsInputAPs.begin();it!=varsInputAPs.end();it++) {
                            if ((thisElement&1)>0) candidate &= vars[*it];
                            else candidate &= !vars[*it];
                            thisElement = thisElement >> 1;
                        }*/

                        // Process counters
                        for (auto it = varsAssumptionCounterPre.begin();it!=varsAssumptionCounterPre.end();it++) {
                            if ((thisElement&1)>0) candidate &= vars[*it];
                            if ((thisElement&1)>0) std::cerr << varNames[*it] << std::endl;
                            else candidate &= !vars[*it];
                            thisElement = thisElement >> 1;
                        }
                        for (auto it = varsGuaranteeCounterPre.begin();it!=varsGuaranteeCounterPre.end();it++) {
                            if ((thisElement&1)>0) candidate &= vars[*it];
                            if ((thisElement&1)>0) std::cerr << varNames[*it] << std::endl;
                            else candidate &= !vars[*it];
                            thisElement = thisElement >> 1;
                        }

                        // Process assumptions
                        for (auto it = varsAssumptionUVWStatesPre.begin();it!=varsAssumptionUVWStatesPre.end();it++) {
                            if ((thisElement&1)>0) candidate &= vars[*it];
                            if ((thisElement&1)>0) std::cerr << varNames[*it] << std::endl;
                            else candidate &= !vars[*it];
                            thisElement = thisElement >> 1;
                        }

                        // Speedup.

                        // Process guarantees
                        for (uint j=0;j<guarantees.states.size();j++) {
                            //BF_DUMPANDCHECK(*this,candidate,NULL,"/tmp/candidate.dot");
                            if ((thisElement&1)>0) {
                                // if ((thisElement&1)>0) std::cerr << "Gua state: " << j << std::endl;
                                if (renumberingOfTranditionallyHandledGuarantees.count(j)>0) {
                                    candidate &= varsGuaranteeUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[j]);
                                } else {
                                    candidate &= (!bfForInitialNoIncomingEdgesState) & guaranteeStateActivatingIO[j];
                                }
                            } else {
                                if (renumberingOfTranditionallyHandledGuarantees.count(j)>0) {
                                    candidate &= !varsGuaranteeUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[j]);
                                } else {
                                    candidate &= !((!bfForInitialNoIncomingEdgesState) & guaranteeStateActivatingIO[j]);
                                }
                            }
                            thisElement = thisElement >> 1;
                        }

                        if (candidate==mgr.constantFalse()) {
                            // std::cerr << "T" << i << ": 0" << std::endl;
                            occurrenceString << "0";
                        } else {
                            // std::cerr << "T" << i << ": 1" << std::endl;
                            // throw 77;
                            occurrenceString << "1";
                        }

                        // if (reachableStates != initialStatePre) throw 15;
                    }
                    std::cerr << "Reachable states FINAL MD5: " << md5(occurrenceString.str()) << std::endl;
                    std::cerr << "MD5 length: " << occurrenceString.str().length() << std::endl;
                }
#endif
            BF prefixPoint = reachableStates;

            for (auto it = transitionOperations.begin();it!=transitionOperations.end();it++) {
                if (!it->first) {
                    prefixPoint &= it->second;
                    //std::cerr << ",";
                    std::cerr.flush();
                }
            }

            //BF_DUMPANDCHECK3(*this,prefixPoint & varsAssumptionCounterPost.getVarBF(0) & !varsGuaranteeCounterPost.getVarBF(0) & varsGuaranteeUVWStatesPost.getVarBF(renumberingOfTranditionallyHandledGuarantees[0])& !varsGuaranteeUVWStatesPost.getVarBF(renumberingOfTranditionallyHandledGuarantees[6]) & !varsGuaranteeUVWStatesPost.getVarBF(renumberingOfTranditionallyHandledGuarantees[7]) & varsGuaranteeUVWStatesPost.getVarBF(renumberingOfTranditionallyHandledGuarantees[8]) & !varsGuaranteeUVWStatesPost.getVarBF(renumberingOfTranditionallyHandledGuarantees[9]) ,NULL,"/tmp/reachableStatesTransition",stateThatIsOnlyInitial,".dot")
            BF_DUMPANDCHECK3(*this,prefixPoint.ExistAbstract(varsGuaranteeCounterPost.getCube()).ExistAbstract(varsGuaranteeCounterPre.getCube()).ExistAbstract(varsAssumptionCounterPre.getCube()) ,NULL,"/tmp/reachableStatesTransition",stateThatIsOnlyInitial,".dot")

            // Now increment counter whenever guarantee is satisfied.
            BF filtered = prefixPoint & !vars[varsGuaranteeCounterPost[0]];
            for (int j=0;j<(1 << (varsGuaranteeCounterPost.size()-1));j++) {

                //std::cerr << ".";
                std::cerr.flush();

                BF thisCase = prefixPoint & vars[varsGuaranteeCounterPost[0]];
                for (int k=0;k<(int)varsGuaranteeCounterPost.size()-1;k++) {
                    if ((j & (1 << k))!=0)
                        thisCase &= vars[varsGuaranteeCounterPost[k+1]];
                    else
                        thisCase &= !vars[varsGuaranteeCounterPost[k+1]];
                }
                thisCase = thisCase.ExistAbstract(computeCube(varsGuaranteeCounterPost));
                thisCase &= !vars[varsGuaranteeCounterPost[0]];
                for (int k=0;k<(int)varsGuaranteeCounterPost.size()-1;k++) {
                    if (((j+1) & (1 << k))!=0)
                        thisCase &= vars[varsGuaranteeCounterPost[k+1]];
                    else
                        thisCase &= !vars[varsGuaranteeCounterPost[k+1]];
                }
                filtered |= thisCase;
            }

            // Make sure that last I/O is correctly transferred
            for (uint k=0;k<inputsToBeKeptPre.size();k++) {
                filtered &= !(vars[inputsToBeKeptPre[k]] ^ vars[inputsToBeKeptPost[k]]);
            }
            for (uint k=0;k<outputsToBeKeptPre.size();k++) {
                filtered &= !(vars[outputsToBeKeptPre[k]] ^ vars[outputsToBeKeptPost[k]]);
            }

            BF nextOne = filtered.ExistAbstract(computeCube(varsAssumptionCounterPre))
                              .ExistAbstract(computeCube(varsGuaranteeCounterPre))
                              .ExistAbstract(computeCube(varsAssumptionUVWStatesPre))
                              .ExistAbstract(computeCube(varsGuaranteeUVWStatesPre));

            nextOne = nextOne.SwapVariables(gameSolvingSwapVectorPre,gameSolvingSwapVectorPost).ExistAbstract(computeCube(varsInputAPs))
                      .ExistAbstract(computeCube(varsOutputAPs));
            oldReachableStates = reachableStates;
            reachableStates |= nextOne;
        }
        BF_DUMPANDCHECK3(*this,reachableStates,"assumptionCounterPre guaranteeCounterPre assumptionUVWStatesPre guaranteeUVWStatesPre inputsToBeKeptPre outputsToBeKeptPre","/tmp/reachableStates",stateThatIsOnlyInitial,".dot")
#endif


        int outerIteration = 0;
        std::vector<BF> envPrefixPoints; // For the strategy extraction
        std::vector<BF> fixedPointAcceleratorSys;
        std::vector<BF> fixedPointAcceleratorEnv;

        do {
           outerIteration++;
           outerFixedPointEnvOld = outerFixedPointEnv;
           envPrefixPoints.clear();
           std::vector<BF> nextFixedPointAcceleratorSys;
           std::vector<BF> nextFixedPointAcceleratorEnv;

           // Compute outer pure color 2 states
           BF newEnv = outerFixedPointSys;
           // Two cases: Counter changes or not
           BF case1 = (newEnv & !vars[varsGuaranteeCounterPre[0]]).ExistAbstract(vars[varsGuaranteeCounterPre[0]]);
           BF case2 = (newEnv & vars[varsGuaranteeCounterPre[0]]).ExistAbstract(vars[varsGuaranteeCounterPre[0]]).UnivAbstract(computeCube(varsGuaranteeCounterPre));
           outerFixedPointPureEnv = (case1 & case2).UnivAbstract(varsInputAPs.getCube()) & vars[varsGuaranteeCounterPre[0]];
           BF_DUMPANDCHECK(*this,outerFixedPointPureEnv,ENVIRONMENT_POSITION_BF_SUPPORT_VARIABLES,"/tmp/CurrentOuterFixedPointEnv.dot")

           // Middle fixed point
           middleFixedPointEnv = mgr.constantFalse();
           middleFixedPointSys = mgr.constantFalse();
           middleFixedPointPureEnv = mgr.constantFalse();
           BF middleFixedPointEnvOld;
           int middleIteration = 0;

           do {
                middleFixedPointEnvOld = middleFixedPointEnv;
                middleIteration++;

                // Compute middle pure color 1 states
                newEnv = middleFixedPointSys;
                // Two cases: Counter changes or not
                BF case1 = (newEnv & !vars[varsGuaranteeCounterPre[0]]).ExistAbstract(vars[varsGuaranteeCounterPre[0]]);
                BF case2 = (newEnv & vars[varsGuaranteeCounterPre[0]]).ExistAbstract(vars[varsGuaranteeCounterPre[0]]).UnivAbstract(computeCube(varsGuaranteeCounterPre));
                middleFixedPointPureEnv = (case1 & case2).UnivAbstract(varsInputAPs.getCube()) & !vars[varsGuaranteeCounterPre[0]] & vars[varsAssumptionCounterPre[0]];
                BF_DUMPANDCHECK(*this,middleFixedPointPureEnv,ENVIRONMENT_POSITION_BF_SUPPORT_VARIABLES,"/tmp/CurrentMiddleFixedPointEnv.dot")

                // Inner fixed point
                if ((int)fixedPointAcceleratorEnv.size()>=middleIteration) {
                    innerFixedPointPureEnv = fixedPointAcceleratorEnv[middleIteration-1];
                    innerFixedPointSys = fixedPointAcceleratorSys[middleIteration-1];
                    innerFixedPointEnv = (outerFixedPointPureEnv | middleFixedPointPureEnv | innerFixedPointPureEnv);
                } else {
                    innerFixedPointEnv = mgr.constantTrue();
                    innerFixedPointPureEnv = mgr.constantTrue();
                    innerFixedPointSys = mgr.constantTrue();
                }
                BF innerFixedPointSysOld;
                BF innerFixedPointEnvOld;
                int innerIteration = 0;

                do {
                    innerFixedPointSysOld = innerFixedPointSys;
                    innerFixedPointEnvOld = innerFixedPointEnv;
                    innerIteration++;

                    std::cerr << "Processing iteration: " << outerIteration << "," << middleIteration << "," << innerIteration << std::endl;

                    // Compute transitions from environment to system state.
                    // Allow changing the counter values
                    BF newEnv = innerFixedPointSys;
                    // Two cases: Counter changes or not
                    BF case1 = (newEnv & !vars[varsGuaranteeCounterPre[0]]).ExistAbstract(vars[varsGuaranteeCounterPre[0]]);
                    BF case2 = (newEnv & vars[varsGuaranteeCounterPre[0]]).ExistAbstract(vars[varsGuaranteeCounterPre[0]]).UnivAbstract(computeCube(varsGuaranteeCounterPre));
                    innerFixedPointPureEnv = (case1 & case2).UnivAbstract(varsInputAPs.getCube()) & !vars[varsGuaranteeCounterPre[0]] & !vars[varsAssumptionCounterPre[0]];
                    BF_DUMPANDCHECK(*this,innerFixedPointPureEnv,ENVIRONMENT_POSITION_BF_SUPPORT_VARIABLES,"/tmp/CurrentInnerFixedPointEnv.dot")

                    // Computer system elemeent
                    innerFixedPointSys = (outerFixedPointPureEnv | middleFixedPointPureEnv | innerFixedPointPureEnv).SwapVariables(gameSolvingSwapVectorPre,gameSolvingSwapVectorPost);
                    BF_DUMPANDCHECK(*this,innerFixedPointSys,NULL,"/tmp/InnerSysTarget.dot")
                    //BF_DUMPANDCHECK(*this,innerFixedPointSys,NULL,"/tmp/test.dot")
                    for (auto it = transitionOperations.begin();it!=transitionOperations.end();it++) {
                        if (it->first) {
                            innerFixedPointSys = innerFixedPointSys.ExistAbstractSingleVar(it->second);
                        } else {
                            innerFixedPointSys &= it->second;
                        }
                        BF_DUMPANDCHECK(*this,innerFixedPointSys,"inputsToBeKeptPre outputsToBeKeptPre guaranteeCounterPre guaranteeCounterPost inputAPs outputAPs assumptionUVWStatesPre assumptionUVWStatesPost guaranteeUVWStatesPre guaranteeUVWStatesPost assumptionCounterPre assumptionCounterPost","/tmp/test.dot")
                    }
                    // For reference, the following version of the transitionOperations applyer first computes a conjunction, and then abstracts existentially:
                    /*
                    for (auto it = transitionOperations.begin();it!=transitionOperations.end();it++) {
                        if (!it->first) {
                            innerFixedPointSys &= it->second;
                        }
                    }
                    BF_DUMPANDCHECK(*this,innerFixedPointSys,"assumptionUVWStatesPre guaranteeUVWStatesPre assumptionCounterPre guaranteeCounterPre inputAPs outputAPs assumptionUVWStatesPost guaranteeUVWStatesPost assumptionCounterPost guaranteeCounterPost","/tmp/InnerSysTargetIntermediate.dot")
                    for (auto it = transitionOperations.begin();it!=transitionOperations.end();it++) {
                        if (it->first) {
                            innerFixedPointSys = innerFixedPointSys.ExistAbstractSingleVar(it->second);
                        }
                    }*/

                    BF_DUMPANDCHECK(*this,innerFixedPointSys,NULL,"/tmp/InnerSysPreAbstract.dot")
                    innerFixedPointSys = innerFixedPointSys.ExistAbstract(computeCube(varsOutputAPs));
                    BF_DUMPANDCHECK(*this,innerFixedPointSys,SYSTEM_POSITION_BF_SUPPORT_VARIABLES,"/tmp/InnerSys.dot")

                    // Compute hash for the set of state/input combinations
#ifdef DEBUG_CHARACTERIZE_INNER_SYS
                    uint nofBitsToConsider = varsAssumptionCounterPre.size() + varsGuaranteeCounterPre.size() + inputAPs.size() + assumptions.states.size() + guarantees.states.size();
                    if (nofBitsToConsider > 30) throw "Too many states for this debugging version!";
                    std::ostringstream occurrenceString;
                    std::cerr << "#Bits: " << nofBitsToConsider << "\n";
                    BF theBFWeCareAbout = innerFixedPointSys & reachableStates;
                    //for (uint i=0;i<(1 <<  nofBitsToConsider);i++) {
                    for (uint i=2248832;i<2248833;i++) {

                        BF candidate = theBFWeCareAbout;
                        uint thisElement = i;
                        if ((i%1000000)==0) std::cerr << ".";

                        // Process inputAPs
                        for (auto it = varsInputAPs.begin();it!=varsInputAPs.end();it++) {
                            if ((thisElement&1)>0) candidate &= vars[*it];
                            if ((thisElement&1)>0) std::cerr << varNames[*it] << std::endl;
                            else candidate &= !vars[*it];
                            thisElement = thisElement >> 1;
                        }

                        // Process counters
                        for (auto it = varsAssumptionCounterPre.begin();it!=varsAssumptionCounterPre.end();it++) {
                            if ((thisElement&1)>0) candidate &= vars[*it];
                            if ((thisElement&1)>0) std::cerr << varNames[*it] << std::endl;
                            else candidate &= !vars[*it];
                            thisElement = thisElement >> 1;
                        }
                        for (auto it = varsGuaranteeCounterPre.begin();it!=varsGuaranteeCounterPre.end();it++) {
                            if ((thisElement&1)>0) candidate &= vars[*it];
                            if ((thisElement&1)>0) std::cerr << varNames[*it] << std::endl;
                            else candidate &= !vars[*it];
                            thisElement = thisElement >> 1;
                        }

                        // Process assumptions
                        for (auto it = varsAssumptionUVWStatesPre.begin();it!=varsAssumptionUVWStatesPre.end();it++) {
                            if ((thisElement&1)>0) candidate &= vars[*it];
                            if ((thisElement&1)>0) std::cerr << varNames[*it] << std::endl;
                            else candidate &= !vars[*it];
                            thisElement = thisElement >> 1;
                        }

                        // Process guarantees
                        for (uint j=0;j<guarantees.states.size();j++) {
                            if ((thisElement&1)>0) {
                                if ((thisElement&1)>0) std::cerr << "Guarantee State " << j << std::endl;
                                if (renumberingOfTranditionallyHandledGuarantees.count(j)>0) {
                                    candidate &= varsGuaranteeUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[j]);
                                } else {
                                    candidate &= (!bfForInitialNoIncomingEdgesState) & guaranteeStateActivatingIO[j];
                                }
                            } else {
                                if (renumberingOfTranditionallyHandledGuarantees.count(j)>0) {
                                    candidate &= !varsGuaranteeUVWStatesPre.getVarBF(renumberingOfTranditionallyHandledGuarantees[j]);
                                } else {
                                    candidate &= !((!bfForInitialNoIncomingEdgesState) & guaranteeStateActivatingIO[j]);
                                }
                            }
                            thisElement = thisElement >> 1;
                        }

                        std::cerr << "T" << i << ": ";
                        if (candidate==mgr.constantFalse()) {
                            std::cerr << "0\n";
                            occurrenceString << "0";
                        } else {
                            std::cerr << "1\n";
                            occurrenceString << "1";
                        }

                        BF_DUMPANDCHECK(*this,candidate,NULL,"/tmp/candidate.dot");
                        //throw 77;

                        // inputsToBeKeptPre outputsToBeKeptPre inputAPs assumptionUVWStatesPre guaranteeUVWStatesPre assumptionCounterPre
                        //std::cerr << "InnerMD5: " << md5(occurrenceString.str()) << std::endl;

                    }
                    std::cerr << "\nInnerMD5: " << md5(occurrenceString.str()) << std::endl;
                    std::cerr << "MD5 length: " << occurrenceString.str().length() << std::endl;
#endif

                    innerFixedPointEnv = (outerFixedPointPureEnv | middleFixedPointPureEnv | innerFixedPointPureEnv);
                    BF_DUMPANDCHECK(*this,innerFixedPointEnv,ENVIRONMENT_POSITION_BF_SUPPORT_VARIABLES,"/tmp/InnerEnv.dot")


                } while ((innerFixedPointSysOld!=innerFixedPointSys) || (innerFixedPointEnvOld!=innerFixedPointEnv));

                //std::cerr << "Writing to envPrefixPoints(" << innerIteration << ")\n";
                envPrefixPoints.push_back(innerFixedPointPureEnv | middleFixedPointPureEnv | outerFixedPointPureEnv);

                middleFixedPointEnv = (outerFixedPointPureEnv | middleFixedPointPureEnv | innerFixedPointPureEnv);
                BF_DUMPANDCHECK(*this,middleFixedPointEnv,ENVIRONMENT_POSITION_BF_SUPPORT_VARIABLES,"/tmp/MiddleEnv.dot")
                middleFixedPointSys = innerFixedPointSys;

                fixedPointAcceleratorEnv.push_back(innerFixedPointPureEnv);
                fixedPointAcceleratorSys.push_back(innerFixedPointSys);

           } while (middleFixedPointEnv!=middleFixedPointEnvOld);

           outerFixedPointEnv = (outerFixedPointPureEnv | middleFixedPointPureEnv | innerFixedPointPureEnv);
           BF_DUMPANDCHECK(*this,outerFixedPointEnv,ENVIRONMENT_POSITION_BF_SUPPORT_VARIABLES,"/tmp/OuterEnv.dot")
           outerFixedPointSys = middleFixedPointSys;

           fixedPointAcceleratorSys = nextFixedPointAcceleratorSys;
           fixedPointAcceleratorEnv = nextFixedPointAcceleratorEnv;

           if ((outerFixedPointEnv & initialStatePre)==mgr.constantFalse()) {
               std::cout << "\n-- Result: Specification is unrealisable" << std::endl;
               std::cerr << "\n-- Result: Specification is unrealisable" << std::endl;
               return;
           }

        } while (outerFixedPointEnv!=outerFixedPointEnvOld);

        // Winning set should not depend on counters - in theory.
        BF_DUMPANDCHECK(*this,outerFixedPointEnv,"inputsToBeKeptPre outputsToBeKeptPre assumptionUVWStatesPre guaranteeUVWStatesPre","/tmp/winningSet.dot");

        // Now check if initial state is contained.
        if ((outerFixedPointEnv & initialStatePre)!=mgr.constantFalse()) {
            std::cout << "\n-- Result: Specification is realisable" << std::endl;
            std::cerr << "\n-- Result: Specification is realisable" << std::endl;
        } else {
            std::cout << "\n-- Result: Specification is unrealisable" << std::endl;
            std::cerr << "\n-- Result: Specification is unrealisable" << std::endl;
            return;
        }

        // No strategy extraction
        if (noImplementationBuilding) return;

        //==========================================
        // Extract strategy - testing the Anti-chain method
        //==========================================
#ifdef ANTICHAINS_ANALYSIS
        std::cerr << "Proving infos about antichains: \n";
        for (int i=0;i<(int)envPrefixPoints.size();i++) {
            BF toDecompose = envPrefixPoints.at(i);
            std::cerr << "Tier " << i << ":\n";
            while (toDecompose!=mgr.constantFalse()) {

                BF toRemove = toDecompose;
                BF choice = toDecompose;
                // Assumption states
                std::cerr << "  ";
                for (auto it = varsAssumptionUVWStatesPre.begin();it!=varsAssumptionUVWStatesPre.end();it++) {
                    if ((choice & !vars[*it])==mgr.constantFalse()) {
                        choice &= vars[*it];
                        toRemove &= vars[*it];
                        std::cerr << "1";
                    } else {
                        choice &= !vars[*it];
                        std::cerr << "0";
                    }
                }
                std::cerr << ",";
                for (auto it = varsGuaranteeUVWStatesPre.begin();it!=varsGuaranteeUVWStatesPre.end();it++) {
                    if ((choice & vars[*it])==mgr.constantFalse()) {
                        toRemove &= !vars[*it];
                        choice &= !vars[*it];
                        std::cerr << "0";
                    } else {
                        std::cerr << "1";
                        choice &= vars[*it];
                    }
                }
                std::cerr << ",";
                for (auto it = varsAssumptionCounterPre.begin();it!=varsAssumptionCounterPre.end();it++) {
                    if ((choice & vars[*it])==mgr.constantFalse()) {
                        toRemove &= !vars[*it];
                        choice &= !vars[*it];
                        std::cerr << "0";
                    } else {
                        toRemove &= vars[*it];
                        choice &= vars[*it];
                        std::cerr << "1";
                    }
                }
                for (auto it = varsGuaranteeCounterPre.begin();it!=varsGuaranteeCounterPre.end();it++) {
                    if ((choice & vars[*it])==mgr.constantFalse()) {
                        toRemove &= !vars[*it];
                        choice &= !vars[*it];
                        std::cerr << "0";
                    } else {
                        toRemove &= vars[*it];
                        choice &= vars[*it];
                        std::cerr << "1";
                    }
                }
                std::cerr << "\n";
                toDecompose &= !toRemove;

            }
        }
#endif

        //==========================================
        // Extract strategy
        //==========================================
        std::cerr << "Extracting implementation";
        BF transitionRelationSoFar = mgr.constantFalse();
        BF casesCovered = mgr.constantFalse();
        for (int i=0;i<(int)envPrefixPoints.size();i++) {

            // Compute Color 0,1,2-Attractor to this prefix point
            // std::cerr << "i+1: " << i+1 << "\n";

            BF prefixPoint = envPrefixPoints.at(i);
            // std::cerr << "BDD Id " << prefixPoint.getHashCode() << std::endl;

            // Color-2 attractor
            prefixPoint = prefixPoint.SwapVariables(gameSolvingSwapVectorPre,gameSolvingSwapVectorPost);
            for (auto it = transitionOperations.begin();it!=transitionOperations.end();it++) {
                if (!it->first) {
                    prefixPoint &= it->second;
                    std::cerr << ",";
                    std::cerr.flush();
                }
                // No deletion of formerly unneded variables.
            }

            // Now increment counter whenever guarantee is satisfied.
            BF filtered = prefixPoint & !vars[varsGuaranteeCounterPost[0]];
            for (int j=0;j<(1 << (varsGuaranteeCounterPost.size()-1));j++) {

                std::cerr << ".";
                std::cerr.flush();

                BF thisCase = prefixPoint & vars[varsGuaranteeCounterPost[0]];
                for (int k=0;k<(int)varsGuaranteeCounterPost.size()-1;k++) {
                    if ((j & (1 << k))!=0)
                        thisCase &= vars[varsGuaranteeCounterPost[k+1]];
                    else
                        thisCase &= !vars[varsGuaranteeCounterPost[k+1]];
                }
                thisCase = thisCase.ExistAbstract(computeCube(varsGuaranteeCounterPost));
                thisCase &= !vars[varsGuaranteeCounterPost[0]];
                for (int k=0;k<(int)varsGuaranteeCounterPost.size()-1;k++) {
                    if (((j+1) & (1 << k))!=0)
                        thisCase &= vars[varsGuaranteeCounterPost[k+1]];
                    else
                        thisCase &= !vars[varsGuaranteeCounterPost[k+1]];
                }
                filtered |= thisCase;
            }

            BF_DUMPANDCHECK(*this,filtered,NULL,"/tmp/attractor.dot");

            // Compute relevance set
            BF relevanceSet = filtered.ExistAbstract(computeCube(varsAssumptionCounterPost))
                              .ExistAbstract(computeCube(varsGuaranteeCounterPost))
                              .ExistAbstract(computeCube(varsAssumptionUVWStatesPost))
                              .ExistAbstract(computeCube(varsGuaranteeUVWStatesPost))
                              .ExistAbstract(computeCube(varsOutputAPs));
            relevanceSet &= !casesCovered;
            transitionRelationSoFar |= filtered & relevanceSet;
            casesCovered |= relevanceSet;
        }
        BF_DUMPANDCHECK(*this,transitionRelationSoFar,NULL,"/tmp/nondeterministic-strategy.dot");
        std::cerr << "\n";

        //==================================================
        // -1. Prepare variable lists
        //==================================================
        std::vector<uint> varsAllPre;
        std::vector<uint> varsAllPost;
        #define ADD_ALL_NUM(a,b) for (auto it = b.begin();it!=b.end();it++) a.push_back(*it);
        ADD_ALL_NUM(varsAllPre,varsAssumptionUVWStatesPre)
        ADD_ALL_NUM(varsAllPre,varsGuaranteeUVWStatesPre)
        ADD_ALL_NUM(varsAllPre,varsAssumptionCounterPre)
        ADD_ALL_NUM(varsAllPre,varsGuaranteeCounterPre)
        ADD_ALL_NUM(varsAllPre,inputsToBeKeptPre)
        ADD_ALL_NUM(varsAllPre,outputsToBeKeptPre)
        ADD_ALL_NUM(varsAllPost,varsAssumptionUVWStatesPost)
        ADD_ALL_NUM(varsAllPost,varsGuaranteeUVWStatesPost)
        ADD_ALL_NUM(varsAllPost,varsAssumptionCounterPost)
        ADD_ALL_NUM(varsAllPost,varsGuaranteeCounterPost)
        ADD_ALL_NUM(varsAllPost,inputsToBeKeptPost)
        ADD_ALL_NUM(varsAllPost,outputsToBeKeptPost)

        //======================================
        // 0. Compute vectors & VarTypes
        //======================================
        std::tr1::unordered_map<int,BddVarType> varTypesExplicit;
        std::tr1::unordered_map<int,BF> varsExplicit;
        std::tr1::unordered_map<int,std::string> varsExplicitNames;
        std::vector<BF> _pre;
        std::vector<BF> _post;
        std::vector<BF> _input;
        std::vector<BF> _output;
        std::vector<BF> _preInputAndOutput;
        std::vector<BF> _postInputAndOutput;
        std::vector<BF> _postAndOutput;

        for (std::vector<uint>::iterator it = varsAllPre.begin();it!=varsAllPre.end();it++) {
                _pre.push_back(vars[*it]);
                _preInputAndOutput.push_back(_pre.back());
                varTypesExplicit[_pre.back().readNodeIndex()] = PRE;
                varsExplicit[_pre.back().readNodeIndex()] = _pre.back();
                varsExplicitNames[_pre.back().readNodeIndex()] = varNames[*it];
        }

        for (std::vector<uint>::iterator it = varsAllPost.begin();it!=varsAllPost.end();it++) {
                _post.push_back(vars[*it]);
                _postInputAndOutput.push_back(vars[*it]);
                _postAndOutput.push_back(vars[*it]);
                varTypesExplicit[vars[*it].readNodeIndex()] = POST;
                varsExplicit[vars[*it].readNodeIndex()] = vars[*it];
                varsExplicitNames[vars[*it].readNodeIndex()] = varNames[*it];
        }
        for (auto it = varsInputAPs.begin();it!=varsInputAPs.end();it++) {
                _input.push_back(vars[*it]);
                _preInputAndOutput.push_back(vars[*it]);
                _postInputAndOutput.push_back(vars[*it]);
                varTypesExplicit[vars[*it].readNodeIndex()] = IN;
                varsExplicit[vars[*it].readNodeIndex()] = vars[*it];
                varsExplicitNames[vars[*it].readNodeIndex()] = varNames[*it];
        }
        for (auto it = varsOutputAPs.begin();it!=varsOutputAPs.end();it++) {
                _output.push_back(vars[*it]);
                _preInputAndOutput.push_back(vars[*it]);
                _postInputAndOutput.push_back(vars[*it]);
                _postAndOutput.push_back(vars[*it]);
                varTypesExplicit[vars[*it].readNodeIndex()] = OUT;
                varsExplicit[vars[*it].readNodeIndex()] = vars[*it];
                varsExplicitNames[vars[*it].readNodeIndex()] = varNames[*it];
        }
        BFVarVector postStateVec = mgr.computeVarVector(_post);
        BFVarCube postStateCube = mgr.computeCube(_post);
        BFVarVector preStateVec = mgr.computeVarVector(_pre);
        BFVarCube preStateCube = mgr.computeCube(_pre);
        BFVarCube inputCube = mgr.computeCube(_input);
        BFVarCube outputCube = mgr.computeCube(_output);
        BFVarCube preStateInputAndOutputCube = mgr.computeCube(_preInputAndOutput);
        BFVarCube postStateInputAndOutputCube = mgr.computeCube(_postInputAndOutput);
        BFVarCube postStateAndOutputCube = mgr.computeCube(_postAndOutput);

        //==================================================
        // 4. Write the NuSMV model - Initial stuff
        //==================================================
        std::cout << "---------NuSMV model:---------\n";
        std::cout << "MODULE main\n";
        std::cout << "IVAR\n";
        for (uint i=0;i<varsInputAPs.size();i++) {
                std::cout << "  " << varNames[varsInputAPs.getVarNumber(i)] << " : boolean;\n";
        }
        std::cout << "VAR\n";
        for (uint i=0;i<varsAllPre.size();i++) {
                std::cout << "  " << varNames[varsAllPre[i]] << " : boolean;\n";
        }
        for (uint i=0;i<varsOutputAPs.size();i++) {
                std::cout << "  " << varNames[varsOutputAPs[i]] << " : boolean;\n";
        }
        std::cout << "ASSIGN\n";
        // Initial states
        for (uint i=0;i<varsAllPre.size();i++) {
                int varNumber = varsAllPre[i];
                BF current = vars[varNumber];
                std::cout << "  init(" << varNames[varNumber] << ") := ";
                if ((current & initialStatePre)==mgr.constantFalse()) {
                        std::cout << "FALSE;\n";
                } else if (((!current) & initialStatePre)==mgr.constantFalse()) {
                        std::cout << "TRUE;\n";
                } else {
                        std::cout << "FALSE; -- Does not matter\n";
                }
        }

        //==================================================
        // 5. Write the NuSMV model - Recurse
        //==================================================

        // First: Build a graph
        std::tr1::unordered_map<BF,BDDGraphNode> graph;
        std::tr1::unordered_map<BF,int> idNumber;
        {
                std::tr1::unordered_set<BF> todo;
                std::tr1::unordered_set<BF> done;
                todo.insert(transitionRelationSoFar);
                graph[transitionRelationSoFar] = BDDGraphNode();
                while (todo.size()>0) {
                        BF current = *(todo.begin());
                        todo.erase(todo.begin());
                        done.insert(current);
                        idNumber[current] = idNumber.size();
                        if (!current.isConstant()) {
                                assert(varTypesExplicit.count(current.readNodeIndex())>0);
                                BF thenSucc = (current & varsExplicit[current.readNodeIndex()]).ExistAbstractSingleVar(varsExplicit[current.readNodeIndex()]);
                                BF elseSucc = (current & (!varsExplicit[current.readNodeIndex()])).ExistAbstractSingleVar(varsExplicit[current.readNodeIndex()]);
                                graph[thenSucc].incoming.insert(std::pair<bool,BF>(true,current));
                                graph[elseSucc].incoming.insert(std::pair<bool,BF>(false,current));
                                if (done.count(thenSucc)==0) todo.insert(thenSucc);
                                if (done.count(elseSucc)==0) todo.insert(elseSucc);
                        }
                }
        }

        std::tr1::unordered_set<BF> todo; // Is not being erased!
        std::tr1::unordered_set<BF> done;
        todo.insert(transitionRelationSoFar);

        std::cout << "DEFINE\n";

        // Forward steps.
        bool hadInitialNode = false;
        while (todo.size()>0) {
                BF current = *todo.begin();
                todo.erase(todo.begin());
                if (done.count(current)==0) {
                        done.insert(current);

                        // Iterate over the incoming edges
                        std::tr1::unordered_set<std::pair<bool,BF> > &in = graph[current].incoming;
                        if (in.size()==0) {
                                // Initial node
                                std::cout << " firstPass_" << idNumber[current] << "_in:= TRUE;\n";
                                if (hadInitialNode) throw "Already had initial node here!";
                                hadInitialNode = true;
                        } else {
                                std::cout << " firstPass_" << idNumber[current] << "_in:= FALSE";
                                for (std::tr1::unordered_set<std::pair<bool,BF> >::iterator it = in.begin();it!=in.end();it++) {
                                        std::cout << " | firstPass_" << idNumber[it->second] << "_" << (it->first?1:0);
                                }
                                std::cout << ";\n";
                        }

                        // The rest is only for non-const nodes.
                        if (!current.isConstant()) {
                                // Register successors
                                BF thenSucc = (current & varsExplicit[current.readNodeIndex()]).ExistAbstractSingleVar(varsExplicit[current.readNodeIndex()]);
                                BF elseSucc = (current & (!varsExplicit[current.readNodeIndex()])).ExistAbstractSingleVar(varsExplicit[current.readNodeIndex()]);

                                // Add to TODO-list if necessary.
                                todo.insert(thenSucc);
                                todo.insert(elseSucc);
                                assert(varTypesExplicit.count(current.readNodeIndex())>0);

                                switch (varTypesExplicit[current.readNodeIndex()]) {
                                        case OUT:
                                        case POST:
                                                // Sum up all incoming edges for the first pass
                                                std::cout << " firstPass_" << idNumber[current] << "_0:= firstPass_" << idNumber[current] << "_in;\n";
                                                std::cout << " firstPass_" << idNumber[current] << "_1:= firstPass_" << idNumber[current] << "_in;\n";
                                                break;
                                        case IN:
                                        case PRE:
                                                std::cout << " firstPass_" << idNumber[current] << "_0:= firstPass_" << idNumber[current] << "_in & !" << varsExplicitNames[current.readNodeIndex()] << ";\n";
                                                std::cout << " firstPass_" << idNumber[current] << "_1:= firstPass_" << idNumber[current] << "_in & " << varsExplicitNames[current.readNodeIndex()] << ";\n";
                                                break;
                                }
                        }
                }
        }

        // Backward steps
        // ....
        todo.clear();
        done.clear();
        todo.insert(mgr.constantTrue());

        while (todo.size()>0) {
                BF current = *(todo.begin());
                todo.erase(todo.begin());
                done.insert(current);
                if (current==mgr.constantTrue()) {
                        std::cout << " secondPass_" << idNumber[current] << " := TRUE;\n";
                } else {
                        // Collect the results from downwards
                        std::cout << " secondPass_" << idNumber[current] << " := FALSE";

                        BF thenSucc = (current & varsExplicit[current.readNodeIndex()]).ExistAbstractSingleVar(varsExplicit[current.readNodeIndex()]);
                        BF elseSucc = (current & (!varsExplicit[current.readNodeIndex()])).ExistAbstractSingleVar(varsExplicit[current.readNodeIndex()]);

                        if (thenSucc!=mgr.constantFalse()) {
                                std::cout << " | secondPass_" << idNumber[current] << "_" << idNumber[thenSucc];
                        }
                        if (elseSucc!=mgr.constantFalse()) {
                                std::cout << " | secondPass_" << idNumber[current] << "_" << idNumber[elseSucc];
                        }
                        std::cout << ";\n";

                }

                // Propagate upwards
                std::ostringstream constraintsSoFar;
                constraintsSoFar << " secondPass_" << idNumber[current];
                std::tr1::unordered_set<std::pair<bool,BF> > &incoming = graph[current].incoming;
                for (std::tr1::unordered_set<std::pair<bool,BF> >::iterator it = incoming.begin();it!=incoming.end();it++) {
                        std::cout << " secondPass_" << idNumber[it->second] << "_" << idNumber[current] << " := ";
                        std::cout << constraintsSoFar.str() << " & firstPass_" << idNumber[it->second] << "_" << (it->first?"1":"0") << ";\n";
                        constraintsSoFar << " & (!firstPass_" << idNumber[it->second] << "_" << (it->first?"1":"0") << ")";
                        if (done.count(it->second)==0) todo.insert(it->second);
                }
        }

        // Finally make the outward connections
        std::cout << "ASSIGN\n";
        for (uint i=0;i<varsOutputAPs.size();i++) {
                int outputVar = varsOutputAPs[i];
                BF outputVarBF = vars[outputVar];
                std::cout << "  next(" << varNames[varsOutputAPs[i]] << ") := FALSE";
                for (std::tr1::unordered_map<BF,int>::iterator it = idNumber.begin();it!=idNumber.end();it++) {
                        BF current = it->first;
                        if (current.readNodeIndex()==outputVarBF.readNodeIndex()) {
                                BF thenSucc = (current & varsExplicit[current.readNodeIndex()]).ExistAbstractSingleVar(varsExplicit[current.readNodeIndex()]);
                                if (thenSucc!=mgr.constantFalse()) {
                                        std::cout << " | secondPass_" << idNumber[current]  << "_" << idNumber[thenSucc];
                                }
                        }

                }
                std::cout << ";\n";
        }


        for (uint i=0;i<varsAllPost.size();i++) {
                int outputVar = varsAllPost[i];
                BF outputVarBF = vars[outputVar];

                std::cout << "  next(" << varNames[varsAllPre[i]] << ") := FALSE";
                for (std::tr1::unordered_map<BF,int>::iterator it = idNumber.begin();it!=idNumber.end();it++) {
                        BF current = it->first;
                        if (current.readNodeIndex()==outputVarBF.readNodeIndex()) {
                                BF thenSucc = (current & varsExplicit[current.readNodeIndex()]).ExistAbstractSingleVar(varsExplicit[current.readNodeIndex()]);
                                if (thenSucc!=mgr.constantFalse()) {
                                        std::cout << " | secondPass_" << idNumber[current]  << "_" << idNumber[thenSucc];
                                }
                        }

                }
                std::cout << ";\n";
        }


        // Write no specification
        std::cout << "\n\n";
    }

    Backend(const char *inputFilename) : semantics(Undef), varsInputAPs(this),
            varsOutputAPs(this), varsInputAPsSecondCopy(this), varsOutputAPsSecondCopy(this), varsAssumptionUVWStatesPre(this), varsAssumptionUVWStatesPost(this),
            varsGuaranteeUVWStatesPre(this), varsGuaranteeUVWStatesPost(this), varsAssumptionCounterPre(this),
            varsAssumptionCounterPost(this), varsGuaranteeCounterPre(this), varsGuaranteeCounterPost(this),
            inputsToBeKeptPre(this), outputsToBeKeptPre(this)  {

        xmlDoc *xmlDoc = xmlReadFile(inputFilename, NULL, 0);
        if (xmlDoc==NULL) {
        std::cerr << "Error reading XML file: " << inputFilename << "\n";
            throw "Aborting.";
        }

        // Process all Vermicelli
        xmlNode *root = xmlDocGetRootElement(xmlDoc);
        bool rootAlreadyFound = false;
        for (xmlNode *it = root; it!=NULL; it = it->next) {
            if (!rootAlreadyFound) {
                rootAlreadyFound = true;

                bool foundInputAPs;
                bool foundOutputAPs;

                // Parse node
                for (xmlNode *it2 = it->children; it2!=NULL; it2 = it2->next) {
                    if (it2->type==XML_ELEMENT_NODE) {
                        // std::cerr << "Name:" << it2->name << "\n";

                        if (xmlStrEqual(it2->name,BAD_CAST("Title"))) {
                            std::cerr << "Title of the scenario: " << xmlNodeGetChildrenText(it2) << std::endl;
                        } else if (xmlStrEqual(it2->name,BAD_CAST("Description"))) {
                            std::cerr << "Description of the scenario: " << xmlNodeGetChildrenText(it2) << std::endl;
                        } else if (xmlStrEqual(it2->name,BAD_CAST("Inputs"))) {
                            foundInputAPs = true;
                            readAPList(it2->children,inputAPs);
                        } else if (xmlStrEqual(it2->name,BAD_CAST("Outputs"))) {
                            foundOutputAPs = true;
                            readAPList(it2->children,outputAPs);
                        } else if (xmlStrEqual(it2->name,BAD_CAST("Semantics"))) {
                            // Read type
                            xmlChar *c = readAttribute(it2,"type");
                            if (c==NULL) {
                                throw "Error in semantics node: no type attribute found.";
                            } else if (xmlStrEqual(c,BAD_CAST("Mealy"))) {
                                semantics = Mealy;
                            } else if (xmlStrEqual(c,BAD_CAST("Moore"))) {
                                semantics = Moore;
                            } else {
                                std::cerr << "Error: Cannot identify Semantics mode: " << c << std::endl;
                                xmlFree(c);
                                throw "Aborted";
                            }
                            xmlFree(c);
                        } else if (xmlStrEqual(it2->name,BAD_CAST("Specification"))) {
                            bool inAssumptionsGuaranteesMode = false;
                            uint nofUVWReadSoFar = 0;
                            for (xmlNode *it3 = it2->children; it3!=NULL; it3 = it3->next) {
                                if (it3->type==XML_TEXT_NODE) {
                                    // Check if empty
                                    std::string s = (const char *)it3->content;
                                    boost::trim(s);
                                    if (s.length()>0) {
                                        throw "Error in specification node: Didn't expect text.";
                                    }
                                } else if (it3->type==XML_COMMENT_NODE) {
                                    // Nothing to care about.
                                } else if (it3->type==XML_ELEMENT_NODE) {
                                    // Determine type. Is this UVW?
                                    if (xmlStrEqual(it3->name,BAD_CAST("PropertyGroup"))) {
                                        // Check type
                                        xmlChar *type = readAttribute(it3,"type");
                                        if (type==NULL) {
                                            throw "Error in PropertyGroup node: no type attribute found.";
                                        } else if (xmlStrEqual(type,BAD_CAST("Assumption-Guarantee"))) {
                                            // Ok
                                        } else {
                                            std::cerr << "Error: Unsupported type of PropertyGroup: " << type << std::endl;
                                            xmlFree(type);
                                            throw "Aborted";
                                        }
                                        xmlFree(type);
                                        inAssumptionsGuaranteesMode = true;
                                        // In the next round, consider the children
                                        it3 = (xmlNode *)((size_t)it3 - (size_t)(&(it3->next)) + (size_t)(&(it3->children)));
                                    } else if (xmlStrEqual(it3->name,BAD_CAST("UVW"))) {
                                        // Read UVW
                                        std::string uvwString = xmlNodeGetChildrenText(it3);
                                        boost::trim(uvwString);
                                        // std::cerr << "Parsing UVW: " << uvwString << std::endl;
                                        if ((nofUVWReadSoFar==0) && inAssumptionsGuaranteesMode) {
                                            assumptionUVWUnparsed = uvwString;
                                        } else if ((nofUVWReadSoFar==1) && inAssumptionsGuaranteesMode) {
                                            guaranteeUVWUnparsed = uvwString;
                                        } else if (nofUVWReadSoFar==0 && !inAssumptionsGuaranteesMode) {
                                            guaranteeUVWUnparsed = uvwString;
                                        } else if (nofUVWReadSoFar==1 && !inAssumptionsGuaranteesMode) {
                                            std::cerr << "Error: This specification is not in this tool's input fragment";
                                            throw "Aborting";
                                        } else {
                                            std::cerr << "Internal error in line " << __LINE__ << "\n";
                                            throw "Aborting";
                                        }
                                        nofUVWReadSoFar++;
                                    }
                                } else {
                                    std::cerr << "Error: Unexpected node type: " << it3->type << "\n";
                                }
                            }
                        } else {
                            std::cerr << "Error reading input XML File. Found an unexpected tag: " << it2->name << std::endl;
                            throw "Aborted.";
                        }
                    }
                }

                if (!foundInputAPs) {
                    throw "Error reading input XML file: Didn't find the input atomic propositions\n";
                } else if (!foundOutputAPs) {
                    throw "Error reading input XML file: Didn't find the output atomic propositions\n";
                } else if (semantics==Undef) {
                    throw "Error reading input XML file: No Semantics type (Mealy or Moore) given.";
                }

            } else {
                std::cerr << "Error: Multiple root nodes found in XML input file.";
                throw "Aborting";
            }
         }

         // Done.
         xmlFreeDoc(xmlDoc);
    }

    //==================================================================
    // BDD DOT dumping helper functions
    //==================================================================
    virtual void getVariableTypes(std::vector<std::string> &types) const {
        types.push_back("inputAPs");
        types.push_back("outputAPs");
        types.push_back("assumptionUVWStatesPre");
        types.push_back("assumptionUVWStatesPost");
        types.push_back("guaranteeUVWStatesPre");
        types.push_back("guaranteeUVWStatesPost");
        types.push_back("assumptionCounterPre");
        types.push_back("assumptionCounterPost");
        types.push_back("guaranteeCounterPre");
        types.push_back("guaranteeCounterPost");
        types.push_back("inputsToBeKeptPre");
        types.push_back("outputsToBeKeptPre");
    }

    virtual void getVariableNumbersOfType(std::string type, std::vector<uint> &nums) const {
        std::vector<BF> const *myvector;
        if (false) {}
        else {
            if (type=="inputAPs") {
                nums.insert(nums.end(),varsInputAPs.begin(),varsInputAPs.end());
                return;
            } else if (type=="outputAPs") {
                nums.insert(nums.end(),varsOutputAPs.begin(),varsOutputAPs.end());
                return;
            } else if (type=="assumptionUVWStatesPre") {
                nums.insert(nums.end(),varsAssumptionUVWStatesPre.begin(),varsAssumptionUVWStatesPre.end());
                return;
            } else if (type=="assumptionUVWStatesPost") {
                nums.insert(nums.end(),varsAssumptionUVWStatesPost.begin(),varsAssumptionUVWStatesPost.end());
                return;
            } else if (type=="guaranteeUVWStatesPre") {
                nums.insert(nums.end(),varsGuaranteeUVWStatesPre.begin(),varsGuaranteeUVWStatesPre.end());
                return;
            } else if (type=="guaranteeUVWStatesPost") {
                nums.insert(nums.end(),varsGuaranteeUVWStatesPost.begin(),varsGuaranteeUVWStatesPost.end());
                return;
            } else if (type=="assumptionCounterPre") {
                nums.insert(nums.end(),varsAssumptionCounterPre.begin(),varsAssumptionCounterPre.end());
                return;
            } else if (type=="assumptionCounterPost") {
                nums.insert(nums.end(),varsAssumptionCounterPost.begin(),varsAssumptionCounterPost.end());
                return;
            } else if (type=="guaranteeCounterPre") {
                nums.insert(nums.end(),varsGuaranteeCounterPre.begin(),varsGuaranteeCounterPre.end());
                return;
            } else if (type=="guaranteeCounterPost") {
                nums.insert(nums.end(),varsGuaranteeCounterPost.begin(),varsGuaranteeCounterPost.end());
                return;
            } else if (type=="inputsToBeKeptPre") {
                nums.insert(nums.end(),inputsToBeKeptPre.begin(),inputsToBeKeptPre.end());
                return;
            } else if (type=="outputsToBeKeptPre") {
                nums.insert(nums.end(),outputsToBeKeptPre.begin(),outputsToBeKeptPre.end());
                return;
            }

            // No hit
            std::cerr << "Error: Could not idenfity variable type in virtual void getVariableNumbersOfType(std::string type, std::vector<uint> &nums) const:\n" << type << "\n";
            throw "Terminating.";
        }
        for (std::vector<BF>::const_iterator it = myvector->begin();it!=myvector->end();it++) {
            BF thisOne = *it;
            for (uint i=0;i<vars.size();i++) {
                if (thisOne==vars[i]) nums.push_back(i);
            }
        }
    }

    virtual BF getVariableBF(uint number) const {
        return vars[number];
    }

    virtual std::string getVariableName(uint number) const {
        return varNames[number];
    }
};

void showHelp() {
    std::cerr << "Usage: Either: \n   converter --single vermicelliFile\n  converter specFile.xml\n";
}


int main(int argN, const char **args) {
    try {
        if (argN<2) {
            std::cerr << "Error: No input file given!\n";
            return 100;
        }

        std::string filename;
        bool isClassic = false;
        bool noImplementationBuilding = false;
 
        for (int i=1;i<argN;i++) {
            if (args[i][0]=='-') {
                if (std::string(args[i])=="--help") {
                    showHelp();
                    return 0;
                } else if (std::string(args[i])=="--classic") {
                    isClassic = true;
                } else if (std::string(args[i])=="--nosyn") {
                    noImplementationBuilding = true;
                } else {
                    showHelp();
                    return 100;
                }
            } else {
                if (filename=="") {
                    filename = args[i];
                } else {
                    std::cerr << "Error: More than one input file name given.\n";
                    return 100;
                }
            }
        }

        LIBXML_TEST_VERSION

        Backend algorithm(filename.c_str());
        algorithm.prepareGame();
        algorithm.performSynthesis(isClassic,noImplementationBuilding);

        xmlCleanupParser();

        return 0;
    }  catch (const char *c) {
        std::cerr << "Error: " << c << "\n";
        return 100;
    } catch (int i) {
        std::cerr << "Integer error: " << i<< "\n";
        return 100;
    } /* catch (BFDumpDotException e) {
        std::cerr << "BFDumpDotException: " << e.getMessage()<< "\n";
        return 100;
    } */
}
