//
// A variable allocation optimier.
// Solves the following optimization problem:
//
// Given a set X and a Multiset Y of subsets of X
// Find a subset S of X of minimal cost
// The cost is defined as the sum of
// - b*|S|
// - c for every element of Y for which not all of its elements are in S
//
// The solution is based on network flow/minimum cut analysis

#include "variableAllocationOptimizer.hpp"
#include <iostream>
#include <map>
#include <cassert>

void VariableAllocationOptimizer::solve() {

    // Allocate current solution (in max-flow problem)
    std::vector<bool> selectedVariables;
    for (uint i=0;i<nofVariables;i++) selectedVariables.push_back(false);

    std::vector<bool> selectedCombinations;
    for (uint i=0;i<combinations.size();i++) selectedCombinations.push_back(false);

    std::map<std::pair<uint /*Combination*/ ,uint /* to varible*/ >, int > residualBackEdgeWeigths;

    // Compute according to the Edmonds–Karp algorithm
    do {

        // Search for a new path
        // Allocate variables
        std::vector<std::pair<uint,uint> > todoListVars; // Includes the path costs as the second value
        std::map<uint,int> doneListVarsAndPredecessors;

        std::vector<std::pair<uint,uint> > todoListCombinations; // Includes the path costs as the second value
        std::map<uint,int> doneListCombinationsAndPredecessors;

        // Start off from the source node
        for (uint i=0;i<selectedVariables.size();i++) {
            if (!selectedVariables[i]) {
                todoListVars.push_back(std::pair<uint,uint>(i,penaltyUseVar));
                doneListVarsAndPredecessors[i] = -1;
            }
        }

        // Alternate between vars and combinations node
        bool pathFound = false;
        while ((todoListVars.size()>0) && (!pathFound)) {

            // Compute combinations-successors for variable nodes
            while (todoListVars.size()>0) {
                std::pair<uint,uint> thisOne = todoListVars.back();
                todoListVars.pop_back();
                for (uint j=0;j<combinations.size();j++) {
                    if (combinations[j].count(thisOne.first)>0) {
                        if (doneListCombinationsAndPredecessors.count(j)==0) {
                            todoListCombinations.push_back(std::pair<uint,uint>(j,thisOne.second));
                            doneListCombinationsAndPredecessors[j] = thisOne.first;
                        }
                    }
                }
            }

            // Now compute transitions from combinations nodes
            while ((todoListCombinations.size()>0) && (!pathFound)) {
                std::pair<uint,uint> thisOne = todoListCombinations.back();
                todoListCombinations.pop_back();

                // Winning combination?
                if (!selectedCombinations[thisOne.first]) {
                    pathFound = true;

                    // Modify the residualBackEdgeWeigths
                    int backTrackCombination = thisOne.first;
                    int costOfPath = std::min((int)thisOne.second,penaltyNotUseCombination);
                    selectedCombinations[backTrackCombination] = true;

                    while (backTrackCombination!=-1) {
                        uint sourceVariable = doneListCombinationsAndPredecessors.at(backTrackCombination);

                        // Add to reverse edges
                        auto it2 = residualBackEdgeWeigths.find(std::pair<uint,uint>(backTrackCombination,sourceVariable));
                        if (it2!=residualBackEdgeWeigths.end()) {
                            it2->second += costOfPath;
                        } else {
                            residualBackEdgeWeigths[std::pair<uint,uint>(backTrackCombination,sourceVariable)] = costOfPath;
                        }

                        // What now - We are at a source variable
                        int nextSource = doneListVarsAndPredecessors[sourceVariable];
                        if (nextSource==-1) {
                            backTrackCombination = -1;
                            selectedVariables[sourceVariable] = true;
                        } else {
                            auto it2 = residualBackEdgeWeigths.find(std::pair<uint,uint>(nextSource,sourceVariable));
                            assert(it2!=residualBackEdgeWeigths.end());
                            assert(it2->second>= costOfPath);
                            it2->second -= costOfPath;
                            backTrackCombination = nextSource;
                        }

                    }
                } else {
                    // Check for residual back-edges
                    for (auto it = combinations[thisOne.first].begin();it!=combinations[thisOne.first].end();it++) {
                        auto it2 = residualBackEdgeWeigths.find(std::pair<uint,uint>(thisOne.first,*it));
                        if (it2!=residualBackEdgeWeigths.end()) {
                            if (it2->second>0) {
                                if (doneListVarsAndPredecessors.count(*it)==0) {
                                    todoListVars.push_back(std::pair<uint,uint>(*it,std::min(it2->second,(int)thisOne.second)));
                                    doneListVarsAndPredecessors[*it] = thisOne.first;
                                }
                            }
                        }
                    }
                }
            }
        }

        if (!pathFound) {
            // Done
            // Compute the set of combinations for which all preconditions are fulfilled. From these, re-compute the variables then
            solutionCombinations.clear();
            solutionVariables.clear();
            for (uint j=0;j<selectedVariables.size();j++) solutionVariables.push_back(false);
            for (uint j=0;j<combinations.size();j++) {
                bool ok = true;
                for (auto it = combinations[j].begin();it!=combinations[j].end();it++) {
                    ok &= selectedVariables[*it];
                }
                solutionCombinations.push_back(ok);
                if (ok) {
                    for (auto it = combinations[j].begin();it!=combinations[j].end();it++) {
                        solutionVariables[*it] = true;
                    }
                }
            }
            return;
        }

    } while (true);
}



/*
int main(int argN, const char **args) {

    (void) argN;
    (void) args;

    VariableAllocationOptimizer opt(5,4,3);
    #define ADD_COMBI_2(a,b) { std::vector<uint> v; v.push_back(a); v.push_back(b); opt.addCombination(v); }
    #define ADD_COMBI_3(a,b,c) { std::vector<uint> v; v.push_back(a); v.push_back(b); v.push_back(c); opt.addCombination(v); }
    ADD_COMBI_2(0,1);
    ADD_COMBI_2(1,2);
    ADD_COMBI_2(1,0);
    ADD_COMBI_2(3,4);

    opt.solve();
    std::vector<bool> const &sln = opt.getSolutionVariables();
    for (auto it = sln.begin();it!=sln.end();it++) {
        std::cout << "Solution variable: " << *it << std::endl;
    }
    std::vector<bool> const &sln2 = opt.getSolutionCombinations();
    for (auto it = sln2.begin();it!=sln2.end();it++) {
        std::cout << "Solution combination: " << *it << std::endl;
    }
} */
