/*
 * BFHelper.cpp
 *
 *  Created on: 12.10.2010
 *      Author: ehlers
 */

#include "BFHelper.h"
#include <stdexcept>

BFHelper::BFHelper(const BFManager& manager) :
	m_manager(manager) {
}

BF BFHelper::bitStringLess(const std::vector<BF>& a, const std::vector<BF>& b) const {
	const uint bits = std::max(a.size(), b.size());
	BF soFar = m_manager.constantFalse();
	for (uint bit = 0; bit < bits; ++bit) {
		if (bit >= a.size()) {
			soFar |= b[bit];
		} else if (bit >= b.size()) {
			soFar &= !a[bit];
		} else {
			soFar = ((!a[bit]) & b[bit]) | ((!(a[bit] ^ b[bit])) & soFar);
		}
	}
	return soFar;
}

/*static std::vector<BF> bitsEquals(const std::vector<BF>& a, const std::vector<BF>& b) {
 const uint bits = std::max(a.size(), b.size());
 std::vector<BF> bitPreds;
 for (uint bit = 0; bit < bits; ++bit) {
 if (bit >= a.size()) {
 bitPreds.push_back(!b[bit]);
 } else if (bit >= b.size()) {
 bitPreds.push_back(!a[bit]);
 } else {
 bitPreds.push_back(!(a[bit] ^ b[bit]));
 }
 }
 assert(bitPreds.size() > 0);
 return bitPreds;
 }*/

BF BFHelper::bitStringEquals(const std::vector<BF>& a, const std::vector<BF>& b) const {
	const uint bits = std::max(a.size(), b.size());
	BF equalExpression = m_manager.constantTrue();
	for (uint bit = 0; bit < bits; ++bit) {
		if (bit >= a.size()) {
			equalExpression &= !b[bit];
		} else if (bit >= b.size()) {
			equalExpression &= !a[bit];
		} else {
			equalExpression &= (!(a[bit] ^ b[bit]));
		}
	}
	return equalExpression;
}

std::vector<BF> BFHelper::bitStringAdd(const std::vector<BF>& a, const std::vector<BF>& b) const {

	assert(a.size()>0);
	assert(b.size()>0);

	uint bits = std::max(a.size(), b.size());
	BF carry = m_manager.constantFalse();
	std::vector<BF> result;
	for (uint bit = 0; bit < bits; ++bit) {
		if (bit >= a.size()) {
			result.push_back(b[bit] ^ carry);
			carry = (carry & b[bit]);
		} else if (bit >= b.size()) {
			result.push_back(a[bit] ^ carry);
			carry = (carry & a[bit]);
		} else {
			result.push_back(a[bit] ^ b[bit] ^ carry);
			carry = (a[bit] & b[bit]) | (carry & b[bit]) | (a[bit] & carry);
		}
	}
	result.push_back(carry);
	return result;
}

std::vector<BF> BFHelper::bitStringSubtract(const std::vector<BF>& a, const std::vector<BF>& b) const {

	assert(a.size()>0);
	assert(b.size()>0);

	uint bits = std::max(a.size(), b.size());
	BF carry = m_manager.constantFalse();
	std::vector<BF> result;
	for (uint bit = 0; bit < bits; ++bit) {
		if (bit >= a.size()) {
			result.push_back(b[bit] ^ carry);
			carry = (carry | b[bit]);
		} else if (bit >= b.size()) {
			result.push_back(a[bit] ^ carry);
			carry = (carry) & (!a[bit]);
		} else {
			result.push_back(a[bit] ^ b[bit] ^ carry);
			carry = (carry | b[bit]) & ((!a[bit]) | (b[bit] & carry));
		}
	}
	result.push_back(carry);
	// Now add truncating semantics
	BF noTruncate = !(bitStringLess(a, b));
	for (uint i = 0; i < result.size(); i++) {
		result[i] &= noTruncate;
	}

	// Get rid of non-necessary constantZero entries:
	while ((result.back() == m_manager.constantFalse()) && (result.size() > 1))
		result.erase(result.end() - 1);

	return result;
}

std::vector<BF> BFHelper::signedBitStringSaturate(const uint nof_target_bits, std::vector<BF> a, const int min, const int max) const {

	// Check if function can yield a sane result
	assert(max>=min);
	assert((nof_target_bits >= sizeof(int)*8) || (max<0) || ((1 << (nof_target_bits-1)) > max));
	assert((nof_target_bits >= sizeof(int)*8) || (min>=0) || ((1 << (nof_target_bits-1)) > -min));

	std::vector<BF> rest(a.begin() + 1, a.end());
	if (min < 0) {
		if (max < 0) {
			std::vector<BF> pos = bitStringSaturate(nof_target_bits - 1, rest, -max, -min);
			std::vector<BF> result;
			result.push_back(m_manager.constantTrue());
			for (uint i = 0; i < nof_target_bits - 1; i++) {
				result.push_back((((-max) & (1 << i)) != 0) ? ((!a[0]) | ((a[0]) & pos[i])) : (a[0] & pos[i]));
			}
			return result;
		} else {
			std::vector<BF> ifNeg = bitStringSaturate(nof_target_bits - 1, rest, 0, -min);
			std::vector<BF> ifPos = bitStringSaturate(nof_target_bits - 1, rest, 0, max);
			std::vector<BF> result;
			result.push_back(a[0]);
			for (uint i = 0; i < nof_target_bits - 1; i++) {
				result.push_back((a[0] & ifNeg[i]) | ((!a[0]) & ifPos[i]));
			}
			return result;
		}
	} else {
		if (max < 0) {
			throw std::runtime_error("BFHelper::signedBitStringSaturate Error: maximum value is negative but minimum value is positive.");
		} else {
			//std::cout << "HERE!\n";
			std::vector<BF> pos = bitStringSaturate(nof_target_bits - 1, rest, min, max);
			std::vector<BF> result;
			result.push_back(m_manager.constantFalse());
			for (uint i = 0; i < nof_target_bits - 1; i++) {
				result.push_back(((min & (1 << i)) != 0) ? (a[0] | ((!a[0]) & pos[i])) : ((!a[0]) & pos[i]));
			}
			return result;
		}
	}
}

std::vector<BF> BFHelper::bitStringSaturate(const uint nof_target_bits, std::vector<BF> a, const int min, const int max) const {

	assert(max>=min);
	assert((nof_target_bits >= sizeof(int)*8) || ((1 << (nof_target_bits)) > max));

	while (a.size() < nof_target_bits) {
		a.push_back(m_manager.constantFalse());
	}

	// Allocate space for results
	std::vector<BF> maxSaturated;

	// Make vector large enough
	if ((a.size() < sizeof(uint) * 8)) {
		while (max < (1 << a.size())) {
			a.push_back(m_manager.constantFalse());
		}
	}
	// Create BF stating whether the maximum allowed value was exceeded
	BF exceedsMax = m_manager.constantFalse();
	for (uint i = 0; i < a.size(); i++) {
		if ((i < (sizeof(uint) * 8)) && ((max & ((uint) 1 << i)) != 0)) {
			exceedsMax &= a[i];
		} else {
			exceedsMax |= a[i];
		}
	}
	for (uint i = 0; i < a.size(); i++) {
		if ((i < (sizeof(uint) * 8)) && ((max & ((uint) 1 << i)) != 0)) {
			//std::cerr << "U";
			maxSaturated.push_back(((!exceedsMax) & a[i]) | exceedsMax);
		} else {
			//std::cerr << "A";
			maxSaturated.push_back(((!exceedsMax) & a[i]));
		}
	}

	// Now from below
	if (min == 0) return maxSaturated;
	// Create BF stating whether the maximum allowed value was exceeded
	BF exceedsMin = m_manager.constantFalse();
	std::vector<BF> minSaturated;
	for (uint i = 0; i < a.size(); i++) {
		if ((i < (sizeof(uint) * 8)) && ((min & ((uint) 1 << i)) != 0)) {
			exceedsMin |= !maxSaturated[i];
		} else {
			exceedsMin &= !maxSaturated[i];
		}
	}
	for (uint i = 0; i < a.size(); i++) {
		if ((i < (sizeof(uint) * 8)) && ((min & ((uint) 1 << i)) != 0)) {
			minSaturated.push_back(((!exceedsMin) & maxSaturated[i]) | exceedsMin);
		} else {
			minSaturated.push_back(((!exceedsMin) & maxSaturated[i]));
		}
	}
	return minSaturated;
}

std::vector<BF> BFHelper::bitStringMultiply(const std::vector<BF>& a, const std::vector<BF>& b) const {

	assert(a.size()>0);
	assert(b.size()>0);

	std::vector<BF> soFar;
	soFar.push_back(m_manager.constantFalse());
	uint offset = 0;
	for (std::vector<BF>::const_iterator it = a.begin(); it != a.end(); it++) {
		std::vector<BF> thisElement;
		for (uint i = 0; i < offset; i++) {
			thisElement.push_back(m_manager.constantFalse());
		}
		offset++;
		for (std::vector<BF>::const_iterator it2 = b.begin(); it2 != b.end(); it2++) {
			thisElement.push_back(*it2 & *it);
		}
		soFar = bitStringAdd(soFar, thisElement);
	}

	// Get rid of non-necessary constantZero entries:
	while ((soFar.back() == m_manager.constantFalse()) && (soFar.size() > 1))
		soFar.erase(soFar.end() - 1);

	return soFar;
}

std::vector<BF> BFHelper::bitSelect(const BF selector, const std::vector<BF>& a, const std::vector<BF>& b) const {
	std::vector<BF> result;
	for (uint i = 0; i < std::min(a.size(), b.size()); i++) {
		result.push_back((selector & a[i]) | ((!selector) & b[i]));
	}
	for (uint i = a.size(); i < b.size(); i++) {
		result.push_back((!selector) & b[i]);
	}
	for (uint i = b.size(); i < a.size(); i++) {
		result.push_back((selector) & a[i]);
	}
	return result;
}

std::vector<BF> BFHelper::bitStringDivide(const std::vector<BF>& a, const std::vector<BF>& b) const {
	assert(a.size()>0);
	assert(b.size()>0);

	std::vector<BF> soFar = a;
	std::vector<BF> result;
	result.resize(a.size());
	for (int offset = a.size() - 1; offset >= 0; offset--) {

		// Compute element to compare
		std::vector<BF> thisElement;
		for (int i = 0; i < offset; i++) {
			thisElement.push_back(m_manager.constantFalse());
		}
		thisElement.insert(thisElement.end(), b.begin(), b.end());
		BF fits = !bitStringLess(soFar, thisElement);
		result[offset] = fits;
		soFar = bitSelect(fits, bitStringSubtract(soFar, thisElement), soFar);
	}

	// Divide by 0: set to 0 then
	std::vector<BF> zeroString;
	zeroString.push_back(m_manager.constantFalse());
	for (std::vector<BF>::iterator it = result.begin(); it != result.end(); it++) {
		*it &= !bitStringEquals(b, zeroString);
	}

	// Get rid of non-necessary constantZero entries:
	while ((result.back() == m_manager.constantFalse()) && (result.size() > 1))
		result.erase(result.end() - 1);

	return result;
}

std::vector<BF> BFHelper::bitStringModulo(const std::vector<BF>& a, const std::vector<BF>& b) const {
	return bitStringSubtract(a, bitStringMultiply(bitStringDivide(a, b), b));
}

std::vector<BF> BFHelper::signedBitStringModulo(const std::vector<BF>& a, const std::vector<BF>& b) const {
	return signedBitStringSubtract(a, signedBitStringMultiply(signedBitStringDivide(a, b), b));
}

BF BFHelper::signedBitStringLess(const std::vector<BF>& a, const std::vector<BF>& b) const {
	std::vector<BF> unsignedPartA(a.begin() + 1, a.end());
	std::vector<BF> unsignedPartB(b.begin() + 1, b.end());
	BF unsignedLess = bitStringLess(unsignedPartA, unsignedPartB);
	return (a[0] & (!b[0])) | ((!a[0]) & (!b[0]) & unsignedLess) | (a[0] & b[0] & (!unsignedLess) & (!bitStringEquals(unsignedPartA, unsignedPartB)));
}

std::vector<BF> BFHelper::signedBitStringAdd(const std::vector<BF>& a, const std::vector<BF>& b) const {
	std::vector<BF> unsignedPartA(a.begin() + 1, a.end());
	std::vector<BF> unsignedPartB(b.begin() + 1, b.end());
	std::vector<BF> result;
	BF b_less_a = bitStringLess(unsignedPartB, unsignedPartA);
	BF a_less_b = bitStringLess(unsignedPartA, unsignedPartB);
	result.push_back((a[0] & b[0]) | (a[0] & (!b[0]) & b_less_a) | ((!a[0]) & b[0] & a_less_b));
	std::vector<BF> variantA = bitStringAdd(unsignedPartA, unsignedPartB);
	std::vector<BF> variantB = bitStringSubtract(unsignedPartA, unsignedPartB);
	std::vector<BF> variantC = bitStringSubtract(unsignedPartB, unsignedPartA);
	uint nofBits = std::max(variantA.size(), std::max(variantB.size(), variantC.size()));
	for (uint i = 0; i < nofBits; i++) {
		BF _a = (i < variantA.size()) ? variantA[i] : m_manager.constantFalse();
		BF _b = (i < variantB.size()) ? variantB[i] : m_manager.constantFalse();
		BF _c = (i < variantC.size()) ? variantC[i] : m_manager.constantFalse();
		result.push_back(((!(a[0] ^ b[0])) & _a) | ((a[0] ^ b[0]) & a_less_b & _c) | ((a[0] ^ b[0]) & b_less_a & _b));
	}
	return result;
}

std::vector<BF> BFHelper::signedBitStringSubtract(const std::vector<BF>& a, const std::vector<BF>& b) const {
	std::vector<BF> secondArgument(b);
	secondArgument[0] = !secondArgument[0];
	return signedBitStringAdd(a, secondArgument);
}

std::vector<BF> BFHelper::signedBitStringMultiply(const std::vector<BF>& a, const std::vector<BF>& b) const {
	std::vector<BF> unsignedPartA(a.begin() + 1, a.end());
	std::vector<BF> unsignedPartB(b.begin() + 1, b.end());
	std::vector<BF> normalResult = bitStringMultiply(unsignedPartA, unsignedPartB);
	normalResult.insert(normalResult.begin(), a[0] ^ b[0]);
	return normalResult;
}

std::vector<BF> BFHelper::signedBitStringDivide(const std::vector<BF>& a, const std::vector<BF>& b) const {
	std::vector<BF> unsignedPartA(a.begin() + 1, a.end());
	std::vector<BF> unsignedPartB(b.begin() + 1, b.end());
	std::vector<BF> normalResult = bitStringDivide(unsignedPartA, unsignedPartB);
	normalResult.insert(normalResult.begin(), a[0] ^ b[0]);
	return normalResult;

}
