/*
 * BFStateToVariableValuationEncoderClassicAutobit.cpp
 *
 *  Created on: 17.08.2010
 *      Author: ehlers
 */

#include "BFStateToVariableValuationEncoderClassicAutobit.h"
#include <cassert>
#include <iostream>

BFStateToVariableValuationEncoderClassicAutobit::BFStateToVariableValuationEncoderClassicAutobit() :
	m_assignment(0) {
}

void BFStateToVariableValuationEncoderClassicAutobit::reset(uint numberOfStates) {
	assert(numberOfStates > 0);

	// Clean up
	nofStates = numberOfStates;
	transitions.clear();
	for (std::vector<char*>::iterator it = grayCodes.begin(); it != grayCodes.end(); it++)
		delete[] *it;
	grayCodes.clear();

	// Allocate space for the transition relation
	transitions.resize(nofStates);

	// Compute number of bits & gray codes
	nofBits = 0;
	{
		int num = nofStates - 1;
		while (num != 0) {
			nofBits++;
			num = num >> 1;
		}
	}
	// TODO handle nofBits == 0 case
	for (int i = 0; i < (1 << nofBits); i++) {
		char *current = new char[nofBits];
		int num = i;
		for (int j = 0; j < nofBits; j++) {
			int pos = num % 4;
			if ((pos == 1) || (pos == 2)) {
				current[j] = 1;
			} else {
				current[j] = 0;
			}
			num = num >> 1;
		}
		grayCodes.push_back(current);
	}

	// Prepare the rest of the data structures
	m_assignment = new uint[nofStates];
	for (uint i = 0; i < nofStates; i++) {
		m_assignment[i] = i;
	}
}

BFStateToVariableValuationEncoderClassicAutobit::~BFStateToVariableValuationEncoderClassicAutobit() {
	if (m_assignment != 0) delete[] m_assignment; // The m_assignment computed
	for (std::vector<char*>::iterator it = grayCodes.begin(); it != grayCodes.end(); it++) {
		delete[] *it;
	}
}

/**
 * This function computes the cost of the edges from one state in the encoding
 */
int BFStateToVariableValuationEncoderClassicAutobit::computeCostForState(uint stateNr) {
	int code1 = m_assignment[stateNr];
	char *gray1 = grayCodes[code1];
	std::set<uint> &c = transitions[stateNr];
	int cost = 0;
	for (std::set<uint>::iterator it = c.begin(); it != c.end(); it++) {
		int code2 = m_assignment[*it];
		char *gray2 = grayCodes[code2];
		for (int i = 0; i < nofBits; i++) {
			if (gray1[i] != gray2[i]) cost++;
		}
	}
	return cost;
}

void BFStateToVariableValuationEncoderClassicAutobit::run() {
	currentCosts = new int[nofStates];

	int totalCostsSoFar = 0;
	for (uint i = 0; i < nofStates; i++) {
		currentCosts[i] = computeCostForState(i);
		totalCostsSoFar += currentCosts[i];
	}

	// Optimize
	int bestSoFar;
	int bestSoSwap1 = -1;
	int bestSoSwap2 = -1;
	do {
		bestSoFar = 0;
		for (uint i = 0; i < nofStates; i++) {
			for (uint j = i + 1; j < nofStates; j++) {
				int base = currentCosts[i] + currentCosts[j];
				uint tmp = m_assignment[i];
				m_assignment[i] = m_assignment[j];
				m_assignment[j] = tmp;
				base -= computeCostForState(i) + computeCostForState(j);
				if (base > bestSoFar) {
					bestSoFar = base;
					bestSoSwap1 = i;
					bestSoSwap2 = j;
				}
				tmp = m_assignment[i];
				m_assignment[i] = m_assignment[j];
				m_assignment[j] = tmp;
			}
		}
		if (bestSoFar != 0) {
			totalCostsSoFar -= currentCosts[bestSoSwap1] + currentCosts[bestSoSwap2];
			uint tmp = m_assignment[bestSoSwap1];
			m_assignment[bestSoSwap1] = m_assignment[bestSoSwap2];
			m_assignment[bestSoSwap2] = tmp;
			currentCosts[bestSoSwap1] = computeCostForState(bestSoSwap1);
			currentCosts[bestSoSwap2] = computeCostForState(bestSoSwap2);
			totalCostsSoFar += currentCosts[bestSoSwap1] + currentCosts[bestSoSwap2];
		}
	} while (bestSoFar != 0);

	delete[] currentCosts;
}

void BFStateToVariableValuationEncoderClassicAutobit::addTransition(uint from, uint to) {
	// Make everything reflexive for the algorithm here
	transitions[from].insert(to);
	transitions[to].insert(from);
}

/***
 * Returns the bitvector for a state m_assignment
 */
uint BFStateToVariableValuationEncoderClassicAutobit::getAssignment(uint stateNr) const {
	assert(stateNr < nofStates);
	assert(m_assignment[stateNr] < nofStates);
        // std::cout << "Returning for state " << stateNr << ": " << m_assignment[stateNr] << "\n";
	return m_assignment[stateNr];

}

int BFStateToVariableValuationEncoderClassicAutobit::getNumberOfBits() const {
	return nofBits;
}
