// A program that computes a UVW representation of a given deterministic
// Buechi automaton.
//
// Diverges if a given DBA has no UVW representation.
//
// (c) 2011 Rüdiger Ehlers

#include "buechi.hpp"
#include "BF.h"
#include "bddDump.h"
#include "BFStateToVariableValuationEncoderClassicAutobit.h"
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>
#include <tr1/unordered_map>

//=============================================================
// Algorithm container class
//=============================================================

class ConverterAlgorithm : public VariableInfoContainer {
private:

    // The automaton and its encodings
    DBW dbw;
    BFStateToVariableValuationEncoderClassicAutobit stateBFEncoder;

    // BF stuff
    BFManager mgr;
    std::vector<BF> vars;
    std::vector<std::string> varNames;
    std::vector<BF> varsAlphabet; // One variable per alphabet symbol
    std::vector<BF> varsSingleStatePre; // Binary encoding of a single state in the automaton
    std::vector<BF> varsSingleStatePost; // Binary encoding of a single state in the automaton
    BFVarCube cubeVarsSingleStatePre;
    BFVarCube cubeVarsSingleStatePost;
    BFVarCube cubeVarsAlphabet;
    BFVarVector vectorVarsSingleStatePre;
    BFVarVector vectorVarsSingleStatePost;

    // BF stuff - the following variables are only allocated during the computeVermicelli() function
    std::vector<BF> varsCurrentVermicelliElementCharacterSelectionPre;
    std::vector<BF> varsCurrentVermicelliElementCharacterSelectionPost;
    std::vector<BF> varsCurrentVermicelliElementCharacterSelectionPlusTypePre;
    std::vector<BF> varsCurrentVermicelliElementCharacterSelectionPlusTypePost;
    BF varVermicelliTypeSelectorPre;
    BF varVermicelliTypeSelectorPost;
    std::vector<BF> varsPowerSetOfStatesPre;
    std::vector<BF> varsPowerSetOfStatesPost;
    std::vector<BF> varsAutomatonRepresentingVermicelliSoFarPre;
    std::vector<BF> varsAutomatonRepresentingVermicelliSoFarPost;

    BFVarCube cubeCurrentVermicelliElementCharacterSelectionPre;
    BFVarCube cubeCurrentVermicelliElementCharacterSelectionPost;
    BFVarCube cubeCurrentVermicelliElementCharacterSelectionPlusTypePre;
    BFVarCube cubeCurrentVermicelliElementCharacterSelectionPlusTypePost;

    BFVarCube cubePowerSetOfStatesPre;
    BFVarCube cubePowerSetOfStatesPost;
    BFVarVector vectorCurrentVermicelliElementCharacterSelectionPre;
    BFVarVector vectorCurrentVermicelliElementCharacterSelectionPost;
    BFVarVector vectorCurrentVermicelliElementCharacterSelectionPlusTypePre;
    BFVarVector vectorCurrentVermicelliElementCharacterSelectionPlusTypePost;
    BFVarVector vectorPowerSetOfStatesPre;
    BFVarVector vectorPowerSetOfStatesPost;

    // Results and intermediate results
    BF rejectingEndComponents; // over varsAlphabet and varsSingleStatePre

    // Tool functions
    static std::string toString(int i) {
        std::ostringstream os;
        os << i;
        return os.str();
    }

    BF determinise(BF whichOne, std::vector<BF> const &vars, bool preferredValue = true) {
        if (whichOne == mgr.constantFalse()) throw "Error in function 'determinise' - the given BF is equivalent to FALSE.";
        BF cand = whichOne;
        for (std::vector<BF>::const_iterator it = vars.begin();it!=vars.end();it++) {
            if (preferredValue==false) {
                BF candZero = cand & !*it;
                if (candZero==mgr.constantFalse()) {
                    cand = cand & *it;
                } else {
                    cand = candZero;
                }
            } else {
                BF candOne= cand & *it;
                if (candOne==mgr.constantFalse()) {
                    cand = cand & !*it;
                } else {
                    cand = candOne;
                }
            }
        }
        assert(cand!=mgr.constantFalse());
        return cand;
    }

    // A special data type for the Vermicelli-so-far-representing automata
    // Always uses the set of variables "varsRepresentingVermicelliAutomatonSoFar"
    //
    // In this class, we represent the chains as a trie.
    // The trie is represented as an indexed vector to pairs, where the first element in every
    // pair describes whether the chain so far is accepting, and the second element contains the
    // transitions
    //
    // After every added vermicelli, the transition relation must be updated.
    //
    // We use one BF variable for every node in the trie. The first one is only
    // set to true initially.
    //
    // If a node is later marked as unneeded, we do not recycle BF variables, but rather just
    // remove their usage. The same goes for slots in the trieNodes array.
    class VermicelliSoFarAutomaton {

        ConverterAlgorithm &parent;
        std::vector<std::pair<bool,std::tr1::unordered_map<BF, uint> > > trieNodes;
        std::vector<BF> transitionFunctions;
        BF acceptingCombinationsAsPre;
        BF acceptingCombinationsAsPost;
        BFManager &mgr;

        void recursive_add(uint trieIndex, std::vector<BF> const &verm, uint arrayIndex) {
            if (arrayIndex==verm.size()) {
                trieNodes[trieIndex].first = true;
                assert(trieIndex < parent.varsAutomatonRepresentingVermicelliSoFarPre.size());
                acceptingCombinationsAsPre |= parent.varsAutomatonRepresentingVermicelliSoFarPre[trieIndex];
                acceptingCombinationsAsPost |= parent.varsAutomatonRepresentingVermicelliSoFarPost[trieIndex];
                return;
            } else {
                assert(trieIndex<trieNodes.size());
                std::tr1::unordered_map<BF, uint> &currentMap = trieNodes[trieIndex].second;
                BF searchedFor = verm[arrayIndex];
                std::tr1::unordered_map<BF, uint>::iterator it = currentMap.find(searchedFor);
                if (it==currentMap.end()) {

                    BF thisCharacterSet = verm[arrayIndex];
                    BF_DUMPANDCHECK(parent,thisCharacterSet,"currentVermicelliElementPre","/tmp/vermicelli-current-element.dot");

                    uint new_index = trieNodes.size();
                    trieNodes.push_back(std::pair<bool,std::tr1::unordered_map<BF, uint> >());
                    trieNodes.back().first = false;
                    trieNodes[trieIndex].second[thisCharacterSet] = new_index;

                    // allocate new variables if necessary
                    if (parent.varsAutomatonRepresentingVermicelliSoFarPre.size()<trieNodes.size()) {
                        parent.varsAutomatonRepresentingVermicelliSoFarPre.push_back(parent.mgr.newVariable());
                        parent.vars.push_back(parent.varsAutomatonRepresentingVermicelliSoFarPre.back());
                        parent.varsAutomatonRepresentingVermicelliSoFarPost.push_back(parent.mgr.newVariable());
                        parent.vars.push_back(parent.varsAutomatonRepresentingVermicelliSoFarPost.back());
                        parent.varNames.push_back("vermicelliSoFarAutomaton-"+toString(parent.varsAutomatonRepresentingVermicelliSoFarPre.size()-1));
                        parent.varNames.push_back("vermicelliSoFarAutomaton-"+toString(parent.varsAutomatonRepresentingVermicelliSoFarPre.size()-1)+"'");
                    }

                    for (int b=0;b<parent.dbw.getSizeSigma();b++) {

                        // Compute transition relation to this new trie state. We have two cases: incoming transition and self-loops
                        BF cases;
                        // Incoming transitions
                        if ((thisCharacterSet & parent.varsCurrentVermicelliElementCharacterSelectionPre[b])==parent.mgr.constantFalse()) {
                            cases = parent.mgr.constantFalse();
                        } else {

                            // Take disjunction of all previous states that can reach this state only via star transitions
                            cases = parent.mgr.constantFalse();
                            int prevIndex = trieIndex;
                            do {
                                assert(prevIndex>=0);
                                cases |= parent.varsAutomatonRepresentingVermicelliSoFarPre[prevIndex];
                                int prevNext = -1;
                                for (uint i=0;i<trieNodes.size();i++) {
                                    for (auto it = trieNodes[i].second.begin();it!=trieNodes[i].second.end();it++) {
                                        if (it->second == (uint)prevIndex) {
                                            if ((parent.varVermicelliTypeSelectorPre & it->first)!=parent.mgr.constantFalse()) {
                                                prevNext = i;
                                            }
                                        }
                                    }
                                }
                                prevIndex = prevNext;
                            } while (prevIndex != -1);

                            // Self-loops
                            // Assert that whether we have a looping node or not is unambiguously defined
                            assert(((thisCharacterSet & parent.varVermicelliTypeSelectorPre)==parent.mgr.constantFalse()) ^ ((thisCharacterSet & (!parent.varVermicelliTypeSelectorPre))==parent.mgr.constantFalse()));

                            if ((thisCharacterSet & parent.varVermicelliTypeSelectorPre)!=parent.mgr.constantFalse()) {
                                cases |= parent.varsAutomatonRepresentingVermicelliSoFarPre[trieNodes.size()-1];
                            }
                        }

                        transitionFunctions[b] &= !(parent.varsAutomatonRepresentingVermicelliSoFarPost[trieNodes.size()-1] ^ cases);
                        BF_DUMPANDCHECK(parent,transitionFunctions[b],"varsAutomatonRepresentingVermicelliSoFarPre varsAutomatonRepresentingVermicelliSoFarPost","/tmp/vermicelli-current-transition-relation.dot");
                    }

                    // Continue recursion
                    recursive_add(new_index,verm,arrayIndex+1);

                } else {
                    recursive_add(it->second,verm,arrayIndex+1);
                }
            }
        }

        void recurse_print(uint trieIndex, uint indent) {
            for (uint i=0;i<indent;i++) std::cerr << " ";
            if (trieNodes[trieIndex].first) {
                std::cerr << "(" << trieIndex << ") a\n";
            } else {
                std::cerr << "(" << trieIndex << ") na\n";
            }
            std::tr1::unordered_map<BF, uint> &currentMap = trieNodes[trieIndex].second;
            for (auto it = currentMap.begin();it!=currentMap.end();it++) {
                BF currentElement = it->first;
                for (uint i=0;i<indent;i++) std::cerr << " ";
                for (uint i=0;i<parent.varsCurrentVermicelliElementCharacterSelectionPre.size();i++) {
                    if ((currentElement & parent.varsCurrentVermicelliElementCharacterSelectionPre[i])!=parent.mgr.constantFalse()) {
                        std::cerr << i << " ";
                    }
                }
                if ((currentElement & parent.varVermicelliTypeSelectorPre)!=parent.mgr.constantFalse()) {
                    std::cerr << "*";
                } else {
                    std::cerr << "t";
                }
                std::cerr << ":\n";
                recurse_print(it->second,indent+1);
            }
        }

        void recurse_print_output_encoded(uint trieIndex, std::ostringstream &os) {
            os << ((trieNodes[trieIndex].first)?"1 ":"0 ");
            std::tr1::unordered_map<BF, uint> &currentMap = trieNodes[trieIndex].second;
            os << currentMap.size() << " ";
            for (auto it = currentMap.begin();it!=currentMap.end();it++) {
                BF currentElement = it->first;
                for (uint i=0;i<parent.varsCurrentVermicelliElementCharacterSelectionPre.size();i++) {
                    if ((currentElement & parent.varsCurrentVermicelliElementCharacterSelectionPre[i])!=parent.mgr.constantFalse()) {
                        os << "1";
                    } else {
                        os << "0";
                    }
                }
                if ((currentElement & parent.varVermicelliTypeSelectorPre)!=parent.mgr.constantFalse()) {
                    os << "* ";
                } else {
                    os << "t ";
                }
                recurse_print_output_encoded(it->second,os);
            }
        }

        /**
          Remove vermicellis that are subsumed by the given one - to be called before adding
          the given vermicelli to reduce their total number.

          Returns whether something has been removed
          */
        bool recurse_remove_contained_vermicellis(uint trieIndex, int parentIndex, std::vector<BF> const &verm, uint arrayIndex) {
            bool somethingHasAtSomePointChanged = false;
            if (arrayIndex < verm.size()) {
                for (auto it = trieNodes[trieIndex].second.begin();(it!=trieNodes[trieIndex].second.end());) {
                    bool somethingChanged = false;

                    // Can skip in the reference vermicelli?
                    if ((verm[arrayIndex] & parent.varVermicelliTypeSelectorPre)!=mgr.constantFalse()) {
                        somethingChanged |= recurse_remove_contained_vermicellis(trieIndex,parentIndex,verm,arrayIndex+1);
                    }
                    if (!somethingChanged) {
                        // Try to move the left element
                        assert(arrayIndex<verm.size());
                        assert(trieNodes[trieIndex].second.size()>0);
                        bool leftMovePossible = ((it->first & parent.varVermicelliTypeSelectorPre)==mgr.constantFalse()) | ((verm[arrayIndex] & parent.varVermicelliTypeSelectorPre)!=mgr.constantFalse());
                        for (int b=0;b<parent.dbw.getSizeSigma();b++) {
                            if ((it->first & parent.varsCurrentVermicelliElementCharacterSelectionPre[b])!=mgr.constantFalse()) {
                                leftMovePossible &= ((verm[arrayIndex] & parent.varsCurrentVermicelliElementCharacterSelectionPre[b])!=mgr.constantFalse());
                            }
                        }
                        if (leftMovePossible) {
                            somethingChanged |= recurse_remove_contained_vermicellis(it->second,trieIndex,verm,arrayIndex+1);
                            if ((!somethingChanged) && ((it->first & parent.varVermicelliTypeSelectorPre)==mgr.constantFalse())) {
                                // We can also only move the left part if the right one is a suitable star-type, so let's try this out!
                                somethingChanged |= recurse_remove_contained_vermicellis(it->second,trieIndex,verm,arrayIndex);
                            }
                        }
                    }
                    if (somethingChanged) {
                        it = trieNodes[trieIndex].second.begin();
                        somethingHasAtSomePointChanged = true;
                    } else {
                        it++;
                    }
                }
            }

                // Remove this node if we don't need it anymore
                if (parentIndex!=-1) {
                    if ((trieNodes[trieIndex].second.size()==0) && ((!trieNodes[trieIndex].first) || (arrayIndex==verm.size()))) {
                        // Delete node from parent
                        for (auto it = trieNodes[parentIndex].second.begin();it!=trieNodes[parentIndex].second.end();) {
                            if (it->second == trieIndex) {
                                trieNodes[parentIndex].second.erase(it);
                                it = trieNodes[parentIndex].second.end();
                            } else {
                                it++;
                            }
                        }

                        // Doesn't need to be accepting anymore
                        acceptingCombinationsAsPre = acceptingCombinationsAsPre.UnivAbstractSingleVar(parent.varsAutomatonRepresentingVermicelliSoFarPre[trieIndex]);
                        acceptingCombinationsAsPost = acceptingCombinationsAsPost.UnivAbstractSingleVar(parent.varsAutomatonRepresentingVermicelliSoFarPost[trieIndex]);
                        trieNodes[trieIndex].first = false;

                        // Remove from the transition relation
                        for (int b=0;b<parent.dbw.getSizeSigma();b++) {
                            transitionFunctions[b] = transitionFunctions[b].ExistAbstractSingleVar(parent.varsAutomatonRepresentingVermicelliSoFarPost[trieIndex]);
                        }
                        return true;
                    }
                }
                return somethingHasAtSomePointChanged;
                // }
        }

    public:
        VermicelliSoFarAutomaton(ConverterAlgorithm &_parent) : parent(_parent), mgr(_parent.mgr) {

            // Allocate first pre/post variables if not yet done.
            if (parent.varsAutomatonRepresentingVermicelliSoFarPre.size()==0) {
                parent.varsAutomatonRepresentingVermicelliSoFarPre.push_back(parent.mgr.newVariable());
                parent.vars.push_back(parent.varsAutomatonRepresentingVermicelliSoFarPre.back());
                parent.varsAutomatonRepresentingVermicelliSoFarPost.push_back(parent.mgr.newVariable());
                parent.vars.push_back(parent.varsAutomatonRepresentingVermicelliSoFarPost.back());
                parent.varNames.push_back("vermicelliSoFarAutomaton-0");
                parent.varNames.push_back("vermicelliSoFarAutomaton-0'");
            }

            for (int b=0;b<parent.dbw.getSizeSigma();b++) {
                // The initial state is only set in the initial round.
                transitionFunctions.push_back(!parent.varsAutomatonRepresentingVermicelliSoFarPost[0]);
            }

            // Allocate root
            trieNodes.push_back(std::pair<bool,std::tr1::unordered_map<BF, uint> >());
            trieNodes.back().first = false;

            acceptingCombinationsAsPre = parent.mgr.constantFalse();
            acceptingCombinationsAsPost = parent.mgr.constantFalse();
        }

        ~VermicelliSoFarAutomaton() {}

        void add_vermicelli(std::vector<BF> const &verm) {
            while (recurse_remove_contained_vermicellis(0,-1,verm,0)) {};
            recursive_add(0,verm,0);
        }

        BF getTransitionRelation(int alphabet_character)  {
            return transitionFunctions[alphabet_character];
        }

        BF getAcceptingCombinationsAsPre() const { return acceptingCombinationsAsPre; }
        BF getAcceptingCombinationsAsPost() const { return acceptingCombinationsAsPost; }
        BF getInitialState() const {
            BF initialState = parent.varsAutomatonRepresentingVermicelliSoFarPre[0];
            for (uint i=1;i<parent.varsAutomatonRepresentingVermicelliSoFarPre.size();i++) {
                initialState &= !parent.varsAutomatonRepresentingVermicelliSoFarPre[i];
            }
            return initialState;
        }

        void print() {
            recurse_print(0,0);
        }

        std::string getOutputEncoding() {
            std::ostringstream os;
            recurse_print_output_encoded(0,os);
            return os.str();
        }

    };

public:

    //==================================================================
    // Setting up functions
    //==================================================================
    ConverterAlgorithm(const char *inputFilename) : dbw(inputFilename), mgr(4096/*,1.05*/) {

        // Encode the DBW into the stateBFEncoder
        stateBFEncoder.reset(dbw.getNofStates());
        for (int i=0;i<dbw.getNofStates();i++) {
            for (int b=0;b<dbw.getSizeSigma();b++) {
                stateBFEncoder.addTransition(i,dbw.getTransition(i,b));
            }
        }
        stateBFEncoder.run();

        // Reserve BDD variables for the alphabet characters
        for (int i=0;i<dbw.getSizeSigma();i++) {
            vars.push_back(mgr.newVariable());
            varsAlphabet.push_back(vars.back());
            varNames.push_back(dbw.getAlphabetCharacterName(i));
            std::cerr << "A: " << varNames.back() << "\n";
        }

        // Reserve variables for representing single states (not: sets
        // of states).
        for (int i=0;i<stateBFEncoder.getNumberOfBits();i++) {
            vars.push_back(mgr.newVariable());
            varsSingleStatePre.push_back(vars.back());
            varNames.push_back(std::string("single_state")+(char)('0'+i)+std::string("_pre"));
            vars.push_back(mgr.newVariable());
            varsSingleStatePost.push_back(vars.back());
            varNames.push_back(std::string("single_state")+(char)('0'+i)+std::string("_post"));
        }

        // Compute Var Vectors and Var Cubes
        cubeVarsSingleStatePre = mgr.computeCube(varsSingleStatePre);
        cubeVarsSingleStatePost = mgr.computeCube(varsSingleStatePost);
        cubeVarsAlphabet = mgr.computeCube(varsAlphabet);
        vectorVarsSingleStatePre = mgr.computeVarVector(varsSingleStatePre);
        vectorVarsSingleStatePost = mgr.computeVarVector(varsSingleStatePost);

    }

    //==================================================================
    // BDD DOT dumping helper functions
    //==================================================================
    virtual void getVariableTypes(std::vector<std::string> &types) const {
        types.push_back("characters");
        types.push_back("singleStatePre");
        types.push_back("singleStatePost");
        types.push_back("currentVermicelliElementPre");
        types.push_back("currentVermicelliElementPost");
        types.push_back("powerSetOfStatesPre");
        types.push_back("powerSetOfStatesPost");
        types.push_back("varsAutomatonRepresentingVermicelliSoFarPre");
        types.push_back("varsAutomatonRepresentingVermicelliSoFarPost");

    }

    virtual void getVariableNumbersOfType(std::string type, std::vector<uint> &nums) const {
        std::vector<BF> const *myvector;
        if (type=="characters") myvector = &varsAlphabet;
        else if (type=="singleStatePre") myvector = &varsSingleStatePre;
        else if (type=="singleStatePost") myvector = &varsSingleStatePost;
        else if (type=="currentVermicelliElementPre") myvector = &varsCurrentVermicelliElementCharacterSelectionPlusTypePre;
        else if (type=="currentVermicelliElementPost") myvector = &varsCurrentVermicelliElementCharacterSelectionPlusTypePost;
        else if (type=="powerSetOfStatesPre") myvector = &varsPowerSetOfStatesPre;
        else if (type=="powerSetOfStatesPost") myvector = &varsPowerSetOfStatesPost;
        else if (type=="varsAutomatonRepresentingVermicelliSoFarPre") myvector = &varsAutomatonRepresentingVermicelliSoFarPre;
        else if (type=="varsAutomatonRepresentingVermicelliSoFarPost") myvector = &varsAutomatonRepresentingVermicelliSoFarPost;
        else {
            std::cerr << "Error: Could not idenfity variable type in virtual void getVariableNumbersOfType(std::string type, std::vector<uint> &nums) const:\n" << type << "\n";
            throw "Terminating.";
        }
        for (std::vector<BF>::const_iterator it = myvector->begin();it!=myvector->end();it++) {
            BF thisOne = *it;
            for (uint i=0;i<vars.size();i++) {
                if (thisOne==vars[i]) nums.push_back(i);
            }
        }
    }

    virtual BF getVariableBF(uint number) const {
        return vars[number];
    }

    virtual std::string getVariableName(uint number) const {
        return varNames[number];
    }

    //==================================================================
    // Compute state x P(alphabet symbols) that represents precisely the
    // maximum set such that for every element (a,b) in it, if we start
    // from state a and only use transitions belonging to alphabet symbols
    // in b, then the run we obtain will be rejecting
    //
    // The implementation is Büchi fixedpoint-like. The previous simple
    // attractor-based version does not necessarily work, as it under-approximates the
    // actual solution in some cases.
    //==================================================================
    void computeRejectingEndTuples() {

        // First compute a BF for the non-accepting states
        BF nonacceptingStates = mgr.constantFalse();
        BF allStatesPre = mgr.constantFalse();
        for (int i=0;i<dbw.getNofStates();i++) {
            uint a = stateBFEncoder.getAssignment(i);
            BF thisState = mgr.constantTrue();
            std::cerr << "Encoding State " << i << ": ";
            for (int j=0;j<stateBFEncoder.getNumberOfBits();j++) {
                bool neg = ((a >> j) & 1)==0;
                std::cerr << (neg?"0":"1");
                thisState &= neg?(!varsSingleStatePre[j]):varsSingleStatePre[j];
            }
            allStatesPre |= thisState;
            if (!dbw.isAcceptingState(i)) {
                std::cerr << " - not accepting\n";
                nonacceptingStates |= thisState;
            } else {
                std::cerr << " - accepting\n";
            }
        }

        BF_DUMPANDCHECK(*this,nonacceptingStates,"characters singleStatePre","/tmp/nonAccStates.dot")

        // Do a Büchi-game like construction
        BF outerFixedPoint = allStatesPre;
        BF previousOuterFixedPoint = (!nonacceptingStates) & allStatesPre;
        while (previousOuterFixedPoint!=outerFixedPoint) {

            BF_DUMPANDCHECK(*this,outerFixedPoint,"characters singleStatePre","/tmp/outerfix.dot")

            previousOuterFixedPoint = outerFixedPoint;

            BF attractor = mgr.constantFalse();
            BF previousAttractor = mgr.constantTrue();
            while (attractor!=previousAttractor) {

                previousAttractor = attractor;

                // Saturate the attractor
                for (int i=0;i<dbw.getNofStates();i++) {
                    for (int b=0;b<dbw.getSizeSigma();b++) {

                        int successorState = dbw.getTransition(i,b);
                        uint predecessorEncoding = stateBFEncoder.getAssignment(i);
                        uint successorEncoding = stateBFEncoder.getAssignment(successorState);

                        // Take the character combinations that are already in the attractor
                        BF thisTransition = mgr.constantTrue();
                        for (uint j=0;j<varsSingleStatePre.size();j++) {
                            bool neg = ((successorEncoding >> j) & 1)==0;
                            thisTransition &= neg?(!varsSingleStatePre[j]):varsSingleStatePre[j];
                        }
                        BF_DUMPANDCHECK(*this,thisTransition,"characters singleStatePre","/tmp/thisTransition.dot")
                        thisTransition &= (attractor | (!nonacceptingStates));
                        BF_DUMPANDCHECK(*this,thisTransition,"characters singleStatePre","/tmp/thisTransition.dot")
                        thisTransition &= previousOuterFixedPoint;
                        BF_DUMPANDCHECK(*this,thisTransition,"characters singleStatePre","/tmp/thisTransition.dot")
                        thisTransition = thisTransition.ExistAbstract(cubeVarsSingleStatePre);
                        BF_DUMPANDCHECK(*this,thisTransition,"characters singleStatePre","/tmp/thisTransition.dot")

                        // Compute predecessor character combinations
                        for (uint j=0;j<varsSingleStatePre.size();j++) {
                            bool neg = ((predecessorEncoding >> j) & 1)==0;
                            thisTransition &= neg?(!varsSingleStatePre[j]):varsSingleStatePre[j];
                        }
                        thisTransition &= previousOuterFixedPoint;
                        thisTransition &= varsAlphabet[b];
                        attractor |= thisTransition;

                        BF_DUMPANDCHECK(*this,thisTransition,"characters singleStatePre","/tmp/thisTransition.dot")

                    }
                }
            }

            outerFixedPoint = attractor;
        }

        BF_DUMPANDCHECK(*this,outerFixedPoint,"characters singleStatePre","/tmp/finalInfinitySet-neg.dot")
        BF_DUMPANDCHECK(*this,!outerFixedPoint & allStatesPre,"characters singleStatePre","/tmp/finalInfinitySet.dot")

        rejectingEndComponents = !outerFixedPoint & allStatesPre;
    }


    //==================================================================
    // Finally compute the UVW by obtaining a set of Vermicelli that
    // represents the rejecting runs.
    //
    // For this, we build a symbolically represented automaton DBA "U" that represents the vermicelli found so far,
    // and compute, for every version of U (starting with the empty language)
    //==================================================================
    void computeVermicelli() {

        // Step 1: Antichain decomposition
        std::vector<std::pair<std::set<uint>,BF> > endComponentDecomposition;

        BF leftToDecompose = rejectingEndComponents;
        while (leftToDecompose!=mgr.constantFalse()) {

            std::set<uint> charactersSet;
            BF compositor = mgr.constantTrue();
            BF to_remove = mgr.constantFalse();
            for (uint i=0;i<(uint)dbw.getSizeSigma();i++) {
                BF candidate = compositor & varsAlphabet[i];
                if ((candidate & leftToDecompose)!=mgr.constantFalse()) {
                    compositor = candidate;
                    charactersSet.insert(i);
                } else {
                    to_remove |= varsAlphabet[i];
                }
            }

            if (charactersSet.size()==0) { // Nothing left to find.
                leftToDecompose = mgr.constantFalse();
            } else {
                BF respectiveStates = (compositor & leftToDecompose).ExistAbstract(cubeVarsAlphabet);
                endComponentDecomposition.push_back(std::pair<std::set<uint>,BF>(charactersSet,respectiveStates));
#ifndef NDEBUG
                {
                    std::ostringstream os;
                    os << "/tmp/antichain-" << endComponentDecomposition.size() << "---";
                    for (std::set<uint>::const_iterator it = charactersSet.begin();it!=charactersSet.end();it++) {
                        os << "-" << *it;
                    }
                    os << ".dot";
                    BF_newDumpDot(*this,respectiveStates,"characters singleStatePre",os.str());
                }
#endif

                // Remove stuff already taken
                // Only allow subsets for which there are more states to be reached
                BF moreToCome = (leftToDecompose & !respectiveStates).ExistAbstract(cubeVarsSingleStatePre);
                leftToDecompose &= to_remove | moreToCome;
            }
        }

        // Step 2: For every anti-chain, compute the vermicelli that are rejecting
        //
        // Compute a graph with the following structure:
        //   (Reference pos in the DBA) x (current element in the vermicelli) x (Vermicelli element star-type) x (Where could I be for the vermicelli elements so far?)
        // x (State in Vermicelli-so-far representing automaton)

        // 1. Reserve additional BDD variables (except for the vermicelli-so-far-automaton)

        // 1.1. Vermicelli element Pre/Post
        for (int i=0;i<dbw.getSizeSigma();i++) {
            // pre
            BF nextVar = mgr.newVariable();
            vars.push_back(nextVar);
            varsCurrentVermicelliElementCharacterSelectionPre.push_back(nextVar);
            varsCurrentVermicelliElementCharacterSelectionPlusTypePre.push_back(nextVar);
            std::ostringstream varName;
            varName << "v" << i;
            varNames.push_back(varName.str());

            // post
            nextVar = mgr.newVariable();
            vars.push_back(nextVar);
            varsCurrentVermicelliElementCharacterSelectionPost.push_back(nextVar);
            varsCurrentVermicelliElementCharacterSelectionPlusTypePost.push_back(nextVar);
            varName << "\'";
            varNames.push_back(varName.str());
        }

        // 1.2 Vermicelli type: with Star or without star. True=With star
        varVermicelliTypeSelectorPre = mgr.newVariable();
        vars.push_back(varVermicelliTypeSelectorPre);
        varsCurrentVermicelliElementCharacterSelectionPlusTypePre.push_back(varVermicelliTypeSelectorPre);
        varNames.push_back("vermicelli_type_pre");
        varVermicelliTypeSelectorPost = mgr.newVariable();
        vars.push_back(varVermicelliTypeSelectorPost);
        varsCurrentVermicelliElementCharacterSelectionPlusTypePost.push_back(varVermicelliTypeSelectorPost);
        varNames.push_back("vermicelli_type_post");

        // 1.3 Power sets of states (for in which states one could be for the vermicelli part seen so far)
        for (int i=0;i<dbw.getNofStates();i++) {
            // pre
            BF nextVar = mgr.newVariable();
            vars.push_back(nextVar);
            varsPowerSetOfStatesPre.push_back(nextVar);
            std::ostringstream varName;
            varName << "ps" << i;
            varNames.push_back(varName.str());

            // post
            nextVar = mgr.newVariable();
            vars.push_back(nextVar);
            varsPowerSetOfStatesPost.push_back(nextVar);
            varName << "\'";
            varNames.push_back(varName.str());
        }

        // 1.4 - Compute cubes and vectors
        cubeCurrentVermicelliElementCharacterSelectionPre = mgr.computeCube(varsCurrentVermicelliElementCharacterSelectionPre);
        cubeCurrentVermicelliElementCharacterSelectionPost = mgr.computeCube(varsCurrentVermicelliElementCharacterSelectionPost);
        cubeCurrentVermicelliElementCharacterSelectionPlusTypePre = mgr.computeCube(varsCurrentVermicelliElementCharacterSelectionPlusTypePre);
        cubeCurrentVermicelliElementCharacterSelectionPlusTypePost = mgr.computeCube(varsCurrentVermicelliElementCharacterSelectionPlusTypePost);
        vectorCurrentVermicelliElementCharacterSelectionPre = mgr.computeVarVector(varsCurrentVermicelliElementCharacterSelectionPre);
        vectorCurrentVermicelliElementCharacterSelectionPost = mgr.computeVarVector(varsCurrentVermicelliElementCharacterSelectionPost);
        vectorCurrentVermicelliElementCharacterSelectionPlusTypePre = mgr.computeVarVector(varsCurrentVermicelliElementCharacterSelectionPlusTypePre);
        vectorCurrentVermicelliElementCharacterSelectionPlusTypePost = mgr.computeVarVector(varsCurrentVermicelliElementCharacterSelectionPlusTypePost);
        cubePowerSetOfStatesPre = mgr.computeCube(varsPowerSetOfStatesPre);
        cubePowerSetOfStatesPost = mgr.computeCube(varsPowerSetOfStatesPost);
        vectorPowerSetOfStatesPre = mgr.computeVarVector(varsPowerSetOfStatesPre);
        vectorPowerSetOfStatesPost = mgr.computeVarVector(varsPowerSetOfStatesPost);

        // 2. Compute set of initial states without vermicelli-so-far-representing automaton
        // We always assume that the vermicelli information represents the *previous* and possibly
        // still active vermicelli. Thus, we can easily start with a zero-encoding that only
        // has the initial state as possible state to be in.
        BF initialState = mgr.constantTrue();
        uint encodingInitialState = stateBFEncoder.getAssignment(0);
        for (int i=0;i<stateBFEncoder.getNumberOfBits();i++) {
            if (((encodingInitialState >> i)&1)==0) {
                initialState &= !varsSingleStatePre[i];
            } else {
                initialState &= varsSingleStatePre[i];
            }
        }
        initialState &= varsPowerSetOfStatesPre[0];
        for (int i=1;i<dbw.getNofStates();i++) {
            initialState &= !varsPowerSetOfStatesPre[i];
        }
        // Select the "empty vermicelli prefix" to start with.
        for (uint i=0;i<varsCurrentVermicelliElementCharacterSelectionPlusTypePre.size();i++) {
            initialState &= !varsCurrentVermicelliElementCharacterSelectionPlusTypePre[i];
        }
        BF_DUMPANDCHECK(*this,initialState,"singleStatePre currentVermicelliElementPre powerSetOfStatesPre","/tmp/initialState.dot")

        // 3. Compute transition relation without vermicelli-so-far-representing automaton
        //    and without reference automaton
        //  -> There are always two transition relations: one with a change in the vermicelli element
        //     to a non-star element, and one to a star-element. The transition relation for Non-changes
        //     need not be built.

        // 3.1. Compute the transition relation for transitions to non-star-vermicelli positions
        BF transitionRelationToNonStarVermicelli = !varVermicelliTypeSelectorPost;
        for (int targetState=0;targetState<dbw.getNofStates();targetState++) {
            BF transitionToTargetState = mgr.constantFalse();
            for (int i=0;i<dbw.getNofStates();i++) {
                for (int b=0;b<dbw.getSizeSigma();b++) {
                    if (dbw.getTransition(i,b)==targetState) {
                        transitionToTargetState |= varsPowerSetOfStatesPre[i] & varsCurrentVermicelliElementCharacterSelectionPost[b];
                    }
                }
            }
            BF_DUMPANDCHECK3(*this,transitionRelationToNonStarVermicelli,"powerSetOfStatesPre currentVermicelliElementPost powerSetOfStatesPost","/tmp/transitionToNonStarVermicelli-not-finished-",targetState,".dot")
            BF newProduct = !(varsPowerSetOfStatesPost[targetState] ^ transitionToTargetState);
            BF_DUMPANDCHECK3(*this,newProduct,"powerSetOfStatesPre currentVermicelliElementPost powerSetOfStatesPost","/tmp/transitionToNonStarVermicelli-add-",targetState,".dot")
            transitionRelationToNonStarVermicelli &= newProduct;
        }
        BF_DUMPANDCHECK(*this,transitionRelationToNonStarVermicelli,"powerSetOfStatesPre currentVermicelliElementPost powerSetOfStatesPost","/tmp/transitionToNonStarVermicelli.dot")

        // 3.2 Compute the transition relation for transitions to a star-vermicelli position.
        // For this, we start with the non-star Vermicelli transition relation, and saturate the post
        // states. We do this in a fixed point loop in situ, i.e., without a third copy of powerSetOfStates variables.
        BF prefixedPointVermicelliStarTransition = mgr.constantTrue();

        // Start with one-jump transition
        BF fixedPointVermicelliStarTransition = transitionRelationToNonStarVermicelli.ExistAbstractSingleVar(varVermicelliTypeSelectorPost);
        fixedPointVermicelliStarTransition &= varVermicelliTypeSelectorPost;

        // Add self-loops
        for (int i=0;i<dbw.getNofStates();i++) {
            BF oldCases = (fixedPointVermicelliStarTransition & varsPowerSetOfStatesPost[i]).ExistAbstract(cubePowerSetOfStatesPost);
            fixedPointVermicelliStarTransition = fixedPointVermicelliStarTransition.ExistAbstractSingleVar(varsPowerSetOfStatesPost[i]) & !(varsPowerSetOfStatesPost[i] ^ (oldCases | varsPowerSetOfStatesPre[i]));
        }

        // Add the transitive closure
        uint round_number = 0;
        while (prefixedPointVermicelliStarTransition!=fixedPointVermicelliStarTransition) {

            // Continue fixed point computation
            round_number++;
            prefixedPointVermicelliStarTransition = fixedPointVermicelliStarTransition;
            for (int i=0;i<dbw.getNofStates();i++) {
                for (int b=0;b<dbw.getSizeSigma();b++) {
                    int target = dbw.getTransition(i,b);
                    BF oldCases = (fixedPointVermicelliStarTransition & varsPowerSetOfStatesPost[target]).ExistAbstract(cubePowerSetOfStatesPost);
                    BF newCases = (fixedPointVermicelliStarTransition & varsPowerSetOfStatesPost[i] & varsCurrentVermicelliElementCharacterSelectionPlusTypePost[b]).ExistAbstract(cubePowerSetOfStatesPost);
                    fixedPointVermicelliStarTransition = fixedPointVermicelliStarTransition.ExistAbstractSingleVar(varsPowerSetOfStatesPost[target]) & !(varsPowerSetOfStatesPost[target] ^ (oldCases | newCases));
                }
            }
        }

        BF_DUMPANDCHECK(*this,fixedPointVermicelliStarTransition,"powerSetOfStatesPre currentVermicelliElementPost powerSetOfStatesPost","/tmp/transitionToStarVermicelli.dot")

        // 4. For every anti-chain, perform a loop with:
        // 4.1 Compute transition relations (for every input alphabet character) for vermicelli-so-far-representing automata
        // 4.2 Compute for increasing vermicelli-cost the transitive closure of the graph. Stop once we found a new vermicelli
        // 4.3 If we have stopped, reverse-compute the vermicelli and try to make it as strong as possible. Update vermicelli-so-far-representing automaton

        // So let's go.
        std::ostringstream finalOutput; // is given out at the end of the computation.
        finalOutput << endComponentDecomposition.size() << " ";

        int current_antichain_element = -1;
        for (std::vector<std::pair<std::set<uint>,BF> >::iterator endComponentIterator = endComponentDecomposition.begin();endComponentIterator!=endComponentDecomposition.end();endComponentIterator++) {

            // Allocate data structures
            current_antichain_element += 1;
            std::cerr << "Working on antichain element: " << current_antichain_element << "\n";

            // Copy its details to the output stream
            for (int i=0;i<dbw.getSizeSigma();i++) {
                finalOutput << ((endComponentIterator->first.count(i)!=0)?"1":"0");
            }
            finalOutput << " ";


            VermicelliSoFarAutomaton vermicelliSoFar(*this);

            bool found_new_vermicelli = true;
            while (found_new_vermicelli) {

                std::cerr << "Intermediate vermicelli: \n";
                vermicelliSoFar.print();

                // 4.1 Compute transition relations (for every input alphabet character) for vermicelli-so-far-representing automata
                std::vector<BF> vermicelliSoFarTransitionRelations;
                for (int b=0;b<dbw.getSizeSigma();b++) vermicelliSoFarTransitionRelations.push_back(vermicelliSoFar.getTransitionRelation(b));

                // 4.2 Compute for increasing vermicelli-cost the transitive closure of the graph. Stop once we found a new vermicelli.
                std::vector<BF> previousLevels;
                BF sumOfPreviousLevels = mgr.constantFalse();
                previousLevels.push_back(initialState & vermicelliSoFar.getInitialState());
                bool abort_searching_for_longer_vermicellis = false;
                while (!abort_searching_for_longer_vermicellis) {

                    BF_DUMPANDCHECK3(*this,previousLevels.back(),"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/mainloop",current_antichain_element,"-all.dot")
                    BF thisLevelStared = previousLevels.back() & varVermicelliTypeSelectorPre;

                    // 4.2.1. Saturate first-component/vermicelli-so-far-wise (in the pre-variables) the current Level
                    // for star-elements
                    BF previousFixedPoint = mgr.constantFalse();
                    round_number = 0;
                    while (thisLevelStared != previousFixedPoint) {

                        BF_DUMPANDCHECK5(*this,thisLevelStared,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/mainloop",current_antichain_element,"-",round_number,"-stared.dot")
                        previousFixedPoint = thisLevelStared;

                        for (int b=0;b<dbw.getSizeSigma();b++) {
                            BF transitionsForThisSymbol = mgr.constantFalse();
                            for (int post=0;post<dbw.getNofStates();post++) {
                                BF postCases = mgr.constantFalse();
                                for (int pre=0;pre<dbw.getNofStates();pre++) {
                                    if (dbw.getTransition(pre,b)==post) {

                                        // Encode pre in the "pre" variables
                                        BF currentCase = thisLevelStared;
                                        uint pre_encoding = stateBFEncoder.getAssignment(pre);
                                        for (int j=0;j<stateBFEncoder.getNumberOfBits();j++) {
                                            if (((pre_encoding>>j)&1)==0) {
                                                currentCase &= !varsSingleStatePre[j];
                                            } else {
                                                currentCase &= varsSingleStatePre[j];
                                            }
                                        }

                                        // Select alphabet symbol and abstract from pre state
                                        currentCase &= varsCurrentVermicelliElementCharacterSelectionPre[b];
                                        currentCase = currentCase.ExistAbstract(cubeVarsSingleStatePre);
                                        postCases |= currentCase;
                                    }
                                }

                                // Encode post in the "post" variables
                                uint post_encoding = stateBFEncoder.getAssignment(post);
                                for (int j=0;j<stateBFEncoder.getNumberOfBits();j++) {
                                    if (((post_encoding>>j)&1)==0) {
                                        postCases &= !varsSingleStatePre[j];
                                    } else {
                                        postCases &= varsSingleStatePre[j];
                                    }
                                }

                                transitionsForThisSymbol |= postCases;
                            }

                            // Now connect with the transition for the automaton representing the vermicelli-so-far
                            BF transitionsVermicelliAutomaton = (vermicelliSoFarTransitionRelations[b] & transitionsForThisSymbol).ExistAbstract(mgr.computeCube(varsAutomatonRepresentingVermicelliSoFarPre)).SwapVariables(mgr.computeVarVector(varsAutomatonRepresentingVermicelliSoFarPost),mgr.computeVarVector(varsAutomatonRepresentingVermicelliSoFarPre));
                            thisLevelStared |= transitionsVermicelliAutomaton;
                        }
                    }
                    BF_DUMPANDCHECK(*this,thisLevelStared,"singleStatePre powerSetOfStatesPre currentVermicelliElementPre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/stared-saturated.dot")

                    // Connect with non-stared version
                    BF thisLevel = thisLevelStared | (previousLevels.back() & !varVermicelliTypeSelectorPre);
                    BF_DUMPANDCHECK(*this,thisLevel,"singleStatePre powerSetOfStatesPre currentVermicelliElementPre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/stared-and-unstared.dot")

                    // Update previous levels
                    previousLevels.back() = thisLevel;

                    // 4.2.2. Find vermicelli on this level
                    // -> The following contraints apply:
                    //   Not needed: * The standard encoding of the pre-state must be in the antichain
                    //   * There should be no element active in the power set that is not rejecting
                    //   * The Vermicelli-so-far-automaton must not accept.
                    BF_DUMPANDCHECK(*this,endComponentIterator->second,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/endcomponentiterator.dot")
                    BF_DUMPANDCHECK(*this,vermicelliSoFar.getAcceptingCombinationsAsPre(),"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/acceptingcombinations.dot")

                    BF_DUMPANDCHECK(*this,vermicelliSoFar.getTransitionRelation(0),"varsAutomatonRepresentingVermicelliSoFarPre varsAutomatonRepresentingVermicelliSoFarPost","/tmp/vermicelli-so-far-aut-trans0.dot")
                    BF_DUMPANDCHECK(*this,vermicelliSoFar.getTransitionRelation(1),"varsAutomatonRepresentingVermicelliSoFarPre varsAutomatonRepresentingVermicelliSoFarPost","/tmp/vermicelli-so-far-aut-trans1.dot")

                    BF newVermicelliFound = thisLevel & (!vermicelliSoFar.getAcceptingCombinationsAsPre());
                    for (int i=0;i<dbw.getNofStates();i++) {
                        BF_DUMPANDCHECK(*this,newVermicelliFound,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/vermicelli-pre.dot")
                        BF thisState = mgr.constantTrue();
                        uint thisStateEncoding = stateBFEncoder.getAssignment(i);
                        for (int j=0;j<stateBFEncoder.getNumberOfBits();j++) {
                            if (((thisStateEncoding >> j)&1)==0) {
                                thisState &= !varsSingleStatePre[j];
                            } else {
                                thisState &= varsSingleStatePre[j];
                            }
                        }
                        if ((thisState & endComponentIterator->second)==mgr.constantFalse()) {
                            newVermicelliFound &= !varsPowerSetOfStatesPre[i];
                        }
                    }
                    BF_DUMPANDCHECK(*this,newVermicelliFound,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/vermicelli-pre.dot")

                    if (newVermicelliFound!=mgr.constantFalse()) {

                        // We've found a new vermicelli.
#ifndef NDEBUG
                        std::cerr << "We've found a new vermicelli!\n";
#endif

                        // Write previous levels
                        for (uint i=0;i<previousLevels.size();i++) {
                            BF_DUMPANDCHECK3(*this,previousLevels[i],"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/previousLevel-",i,".dot")
                        }

                        // We start by determinising the very last element (except for the rejecting end-loop)
                        BF_DUMPANDCHECK(*this,newVermicelliFound,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/vermicelli-found.dot")

                        BF lastVermicelliElement = newVermicelliFound.ExistAbstract(cubeVarsSingleStatePre);

                        // Now we trace-back. "Clean up the vermicelli" as we go along.
                        std::vector<BF> cleanedUpVermicelli;
                        cleanedUpVermicelli.resize(previousLevels.size()-1);
                        for (int level=previousLevels.size()-1;level>0;level--) {

                            // Determinise the currently last vermicelli element
                            BF_DUMPANDCHECK(*this,lastVermicelliElement,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/vermicelli-last-element-before-det.dot")
                            lastVermicelliElement = determinise(lastVermicelliElement,varsCurrentVermicelliElementCharacterSelectionPlusTypePre,true);
                            lastVermicelliElement = determinise(lastVermicelliElement,varsPowerSetOfStatesPre,true);
                            lastVermicelliElement = lastVermicelliElement.ExistAbstract(cubeVarsSingleStatePre);
                            BF_DUMPANDCHECK(*this,lastVermicelliElement,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/vermicelli-found-last-element.dot")

                            // .ExistAbstract(mgr.computeCube(varsAutomatonRepresentingVermicelliSoFarPre));
                            // BF_DUMPANDCHECK(*this,cleanedOne.ExistAbstract(cubePowerSetOfStatesPre),"currentVermicelliElementPre","/tmp/vermicelli-cleaned-one.dot");
                            cleanedUpVermicelli[level-1] = lastVermicelliElement.ExistAbstract(cubePowerSetOfStatesPre).ExistAbstract(mgr.computeCube(varsAutomatonRepresentingVermicelliSoFarPre));
                            BF_DUMPANDCHECK(*this,lastVermicelliElement,"currentVermicelliElementPre powerSetOfStatesPre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/vermicelli-cleaned-one.dot")

                            // Inverse saturate the vermicelli-so-far stuff
                            BF onlyVermicelliSoFar = lastVermicelliElement.ExistAbstract(cubePowerSetOfStatesPre).ExistAbstract(cubeCurrentVermicelliElementCharacterSelectionPlusTypePre);
                            BF_DUMPANDCHECK(*this,onlyVermicelliSoFar,"varsAutomatonRepresentingVermicelliSoFarPre","/tmp/only-vermicelli-so-far.dot")
                            bool isStarType = ((lastVermicelliElement & varVermicelliTypeSelectorPre)!=mgr.constantFalse());
                            BF reachablePre = mgr.constantFalse();
                            BF reachablePrePrevious;
                            do {
                                reachablePrePrevious = reachablePre;
                                BF reachablePreNext = mgr.constantFalse();
                                for (int b=0;b<dbw.getSizeSigma();b++) {
                                    if ((lastVermicelliElement & varsCurrentVermicelliElementCharacterSelectionPre[b])!=mgr.constantFalse()) {
                                        reachablePreNext |= ((reachablePre | onlyVermicelliSoFar).SwapVariables(mgr.computeVarVector(varsAutomatonRepresentingVermicelliSoFarPost),mgr.computeVarVector(varsAutomatonRepresentingVermicelliSoFarPre)) & vermicelliSoFar.getTransitionRelation(b)).ExistAbstract(mgr.computeCube(varsAutomatonRepresentingVermicelliSoFarPost));
                                    }
                                }
                                reachablePre = reachablePreNext;
                                BF_DUMPANDCHECK(*this,reachablePre,"varsAutomatonRepresentingVermicelliSoFarPre","/tmp/reachable-pre.dot");

                            } while ((isStarType) & (reachablePrePrevious!=reachablePre));

                            // Compute transitions from the previous level
                            BF postPowerset = lastVermicelliElement.ExistAbstract(mgr.computeCube(varsAutomatonRepresentingVermicelliSoFarPre))
                                                                   .SwapVariables(vectorCurrentVermicelliElementCharacterSelectionPlusTypePost,vectorCurrentVermicelliElementCharacterSelectionPlusTypePre)
                                                                   .SwapVariables(vectorPowerSetOfStatesPost,vectorPowerSetOfStatesPre);

                            BF_DUMPANDCHECK(*this,postPowerset,"currentVermicelliElementPost powerSetOfStatesPost","/tmp/post-powerset.dot");

                            BF possiblePres = postPowerset & (transitionRelationToNonStarVermicelli | fixedPointVermicelliStarTransition);
                            BF_DUMPANDCHECK(*this,possiblePres,"currentVermicelliElementPre currentVermicelliElementPost powerSetOfStatesPre powerSetOfStatesPost","/tmp/post-powerset.dot");

                            lastVermicelliElement = possiblePres.ExistAbstract(cubePowerSetOfStatesPost).ExistAbstract(cubeCurrentVermicelliElementCharacterSelectionPlusTypePost);
                            BF_DUMPANDCHECK(*this,previousLevels[level-1],"currentVermicelliElementPre powerSetOfStatesPre varsAutomatonRepresentingVermicelliSoFarPre singleStatePre","/tmp/previousLevel.dot");
                            lastVermicelliElement &= previousLevels[level-1] & reachablePre;
                        }

                        // Add original vermicelli
                        vermicelliSoFar.add_vermicelli(cleanedUpVermicelli);

#ifdef EXPERIMENTAL_VERMICELLI_EXTENSION
                        if (cleanedUpVermicelli.size()>0) {
                            // Add (current EndComponent)* as element to speed-up the overall computation
                            // if there is a final element that accepts more that
                            // the end component, however, don't do it.

                            for (auto it = cleanedUpVermicelli.begin();it!=cleanedUpVermicelli.end();it++) {
                                BF_DUMPANDCHECK(*this,*it,"currentVermicelliElementPre","/tmp/testing123.dot");
                            }

                            BF finalElement = mgr.constantTrue();
                            bool superSeededByFinalElement = true;
                            bool isSafetyLosingState = true;
                            for (int i=0;i<dbw.getSizeSigma();i++) {
                                finalElement &= (endComponentIterator->first.count(i)!=0)?varsCurrentVermicelliElementCharacterSelectionPlusTypePre[i]:!varsCurrentVermicelliElementCharacterSelectionPlusTypePre[i];
                                if (endComponentIterator->first.count(i)==0) {
                                    superSeededByFinalElement &= ((cleanedUpVermicelli[cleanedUpVermicelli.size()-1] & varsCurrentVermicelliElementCharacterSelectionPlusTypePre[i])!=mgr.constantFalse());
                                } else {
                                    isSafetyLosingState = false;
                                }
                            }

                            superSeededByFinalElement &= ((cleanedUpVermicelli[cleanedUpVermicelli.size()-1] & varsCurrentVermicelliElementCharacterSelectionPlusTypePre[dbw.getSizeSigma()])!=mgr.constantFalse());

                            finalElement &= varsCurrentVermicelliElementCharacterSelectionPlusTypePre[dbw.getSizeSigma()];
                            cleanedUpVermicelli.push_back(finalElement);
                            if ((!superSeededByFinalElement) && isSafetyLosingState) {
                                //std::cerr << "Add extended vermicelli\n";
                                vermicelliSoFar.add_vermicelli(cleanedUpVermicelli);
                            }
                        }
#endif


                        // Debugging: Print out new vermicelli
#ifndef NDEBUG
                        for (uint i=0;i<cleanedUpVermicelli.size();i++) {
                            std::cerr << "e" << i << ":";
                            BF currentElement = cleanedUpVermicelli[i];
                            for (uint i=0;i<varsCurrentVermicelliElementCharacterSelectionPre.size();i++) {
                                if ((currentElement & varsCurrentVermicelliElementCharacterSelectionPre[i])!=mgr.constantFalse()) {
                                    std::cerr << i << " ";
                                }
                            }
                            if ((currentElement & varVermicelliTypeSelectorPre)!=mgr.constantFalse()) {
                                std::cerr << "*\n";
                            } else {
                                std::cerr << "t\n";
                            }
                        }
#endif

                        // Add to vermicelli-so-far-representing automaton
                        abort_searching_for_longer_vermicellis = true;

                    } else {

                        // 4.2.3. Peform transition to the next level
                        std::cerr << "Perform transition to the next level\n";
                        BF_DUMPANDCHECK(*this,thisLevel,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/transition-pre.dot")

                        BF postStates = mgr.constantFalse();
                        for (int b=0;b<dbw.getSizeSigma();b++) {

                            // Apply a few parts of the overall graph transition relation:
                            // - Vermicelli element
                            BF thisCharPost = thisLevel & varsCurrentVermicelliElementCharacterSelectionPost[b];
                            // - Vermicelli-so-far-automataton
                            thisCharPost &= vermicelliSoFarTransitionRelations[b];

                            // - single copy of the reference automaton
                            BF singleTransitions = mgr.constantFalse();
                            for (int i=0;i<dbw.getNofStates();i++) {

                                BF thisAddition = mgr.constantTrue();
                                uint predecessorEncoding = stateBFEncoder.getAssignment(i);
                                for (uint j=0;j<varsSingleStatePre.size();j++) {
                                    bool neg = ((predecessorEncoding >> j) & 1)==0;
                                    thisAddition &= neg?(!varsSingleStatePre[j]):varsSingleStatePre[j];
                                }
                                uint successorEncoding = stateBFEncoder.getAssignment(dbw.getTransition(i,b));
                                for (uint j=0;j<varsSingleStatePre.size();j++) {
                                    bool neg = ((successorEncoding >> j) & 1)==0;
                                    thisAddition &= neg?(!varsSingleStatePost[j]):varsSingleStatePost[j];
                                }

                                singleTransitions |= thisAddition;
                            }
                            thisCharPost &= singleTransitions;

                            BF_DUMPANDCHECK(*this,thisCharPost,"powerSetOfStatesPre powerSetOfStatesPost currentVermicelliElementPre currentVermicelliElementPost singleStatePre singleStatePost varsAutomatonRepresentingVermicelliSoFarPre varsAutomatonRepresentingVermicelliSoFarPost","/tmp/thischar-post.dot")

                            // - Transitions in the power set -> Done externally here
                            /*
                            for (int t=0;t<dbw.getNofStates();t++) {
                                BF possibleCasesThisTarget = mgr.constantFalse();
                                for (int i=0;i<dbw.getNofStates();i++) {
                                    if (dbw.getTransition(i,b)==t) {
                                        BF thisAddition = mgr.constantTrue();
                                        uint predecessorEncoding = stateBFEncoder.getAssignment(i);
                                        for (uint j=0;j<varsSingleStatePre.size();j++) {
                                            bool neg = ((predecessorEncoding >> j) & 1)==0;
                                            thisAddition &= neg?(!varsSingleStatePre[j]):varsSingleStatePre[j];
                                        }
                                        possibleCasesThisTarget |= thisAddition;
                                    }
                                }
                                thisCharPost &= !(varsPowerSetOfStatesPost[t] ^ possibleCasesThisTarget);
                            }*/

                            postStates |= thisCharPost;
                        }

                        // Add already saturated transitions in the power set component.
                        postStates &= (transitionRelationToNonStarVermicelli | fixedPointVermicelliStarTransition);

                        BF_DUMPANDCHECK(*this,postStates,"characters powerSetOfStatesPre powerSetOfStatesPost currentVermicelliElementPre currentVermicelliElementPost singleStatePre singleStatePost varsAutomatonRepresentingVermicelliSoFarPre varsAutomatonRepresentingVermicelliSoFarPost","/tmp/poststates-raw.dot")

                        postStates = postStates.ExistAbstract(mgr.computeCube(varsAutomatonRepresentingVermicelliSoFarPre))
                                                                              .ExistAbstract(cubeCurrentVermicelliElementCharacterSelectionPlusTypePre)
                                                                              .ExistAbstract(cubePowerSetOfStatesPre)
                                                                              .ExistAbstract(cubeVarsSingleStatePre);

                        BF_DUMPANDCHECK(*this,postStates,"powerSetOfStatesPost currentVermicelliElementPost singleStatePost varsAutomatonRepresentingVermicelliSoFarPost","/tmp/poststates-semiraw.dot")

                        // Finally, swap pre against Post
                        postStates = postStates.SwapVariables(mgr.computeVarVector(varsAutomatonRepresentingVermicelliSoFarPost),mgr.computeVarVector(varsAutomatonRepresentingVermicelliSoFarPre));
                        postStates = postStates.SwapVariables(vectorCurrentVermicelliElementCharacterSelectionPlusTypePost,vectorCurrentVermicelliElementCharacterSelectionPlusTypePre);
                        postStates = postStates.SwapVariables(vectorPowerSetOfStatesPost,vectorPowerSetOfStatesPre);
                        postStates = postStates.SwapVariables(vectorVarsSingleStatePost,vectorVarsSingleStatePre);
                        BF_DUMPANDCHECK(*this,postStates,"powerSetOfStatesPre currentVermicelliElementPre singleStatePre varsAutomatonRepresentingVermicelliSoFarPre","/tmp/transition-pre.dot")

                        // Everything found? Check this...
                        if (sumOfPreviousLevels >= postStates) {
                            abort_searching_for_longer_vermicellis = true;
                            found_new_vermicelli = false;
                        } else {
                            previousLevels.push_back(postStates);
                            sumOfPreviousLevels |= postStates;
                        }
                    }
                }
            }

            // Print out the vermicelli found
            std::cerr << "End component vermicelli: ";
            for (std::set<uint>::const_iterator it = endComponentIterator->first.begin();it!=endComponentIterator->first.end();it++) {
                std::cerr << *it << " ";
            }
            std::cerr << "\n";
            vermicelliSoFar.print();
            finalOutput << vermicelliSoFar.getOutputEncoding();
        }

        std::vector<std::string> propNames = dbw.getPropositionNames();
        std::cout << propNames.size() << " ";
        for (uint i=0;i<propNames.size();i++) {
            std::cout << propNames[i] << " ";
        }
        std::cout << finalOutput.str() << "\n";
    }
};


int main(int argN, const char **args) {
    try {
        if (argN<2) {
            std::cerr << "Error: No input file given!\n";
            return 100;
        }

        ConverterAlgorithm algorithm(args[1]);
        algorithm.computeRejectingEndTuples();
        algorithm.computeVermicelli();

        return 0;
    } catch (const char *c) {
        std::cerr << "Error: " << c << "\n";
        return 100;
    } catch (int i) {
        std::cerr << "Integer error: " << i<< "\n";
        return 100;
    } catch (BFDumpDotException e) {
        std::cerr << "BFDumpDotException: " << e.getMessage()<< "\n";
        return 100;
    }
}
