// A program that checks if a given deterministic Buechi automaton
// represents a language that is in the intersection of the ones expressible by
// ACTL and LTL
//
// (c) 2011 Rüdiger Ehlers

#include "buechi.hpp"
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>

int main(int argN, char **args) {
    if (argN<2) {
        std::cerr << "Error: No input file given!\n";
        return 100;
    }
    DBW dba(args[1]);

    // The following algorithm assumes that all states in the DBA are reachable. Thus,
    // check that all states *are* actually reachable. Otherwise complain.
    {
        std::set<int> doneList;
        std::set<int> todoList;

        doneList.insert(0);
        todoList.insert(0);
        while (todoList.size()>0) {
            int nextOne = *todoList.begin();
            todoList.erase(todoList.begin());
            for (int b=0;b<dba.getSizeSigma();b++) {
                int succ = dba.getTransition(nextOne,b);
                if (doneList.count(succ)==0) {
                    doneList.insert(succ);
                    todoList.insert(succ);
                }
            }
        }

        if (dba.getNofStates()!=(int)doneList.size()) {
            std::cerr << "Error: Not all states in the deterministic Büchi automaton are reachable!\nThis tool assumes that this is the case.\n";
            return 100;
        }
    }

    // We now start with criterion (*) from the paper "The common fragment of ACTL and LTL" and check
    // that it holds.
    //
    // Here, we do not build the transitive closure of some relation, but rather
    // compute the language containments we have to check directly.

    for (int p=0;p<dba.getNofStates();p++) {
        for (int q=0;q<dba.getNofStates();q++) {

            if (p!=q) {

                std::set<boost::tuple<int,int,int> > doneList; // First: State, Second: Signature so far
                std::set<boost::tuple<int,int,int> > todoList;

                todoList.insert(boost::make_tuple(p,q,p));
                doneList.insert(boost::make_tuple(p,q,p));

                while (todoList.size()>0) {

                    // Read next tuple
                    boost::tuple<int,int,int> thisTuple = *todoList.begin();
                    todoList.erase(todoList.begin());

                    // Iterate over the alphabet
                    for (int b=0;b<dba.getSizeSigma();b++) {
                        int target1 = dba.getTransition(thisTuple.get<0>(),b);
                        int target2 = dba.getTransition(thisTuple.get<1>(),b);

                        boost::tuple<int,int,int> nextTupleA = boost::make_tuple(target1,target2,thisTuple.get<2>());
                        if (doneList.count(nextTupleA)==0) {
                            todoList.insert(nextTupleA);
                            doneList.insert(nextTupleA);
                        }

                        boost::tuple<int,int,int> nextTupleB = boost::make_tuple(target1,target2,dba.getTransition(thisTuple.get<2>(),b));
                        if (doneList.count(nextTupleB)==0) {
                            todoList.insert(nextTupleB);
                            doneList.insert(nextTupleB);
                        }
                    }
                }

                if (doneList.count(boost::make_tuple(p,q,q))) {
                    // std::cout << "I would now have to check: " << p << " >= " << q << " (when viewing the automaton as Büchi automaton)\n";
                    if (!dba.checkStateLanguageSubsumption(q,p)) {
                        std::cerr << "This automaton is not in the intersection of ACTL and LTL! Reason (1)\n";
                        return 1;
                    }
                }
            }
        }
    }

    // Now check criterion (**) from the paper
    for (int p=0;p<dba.getNofStates();p++) {
        if (!(dba.isAcceptingState(p))) {
            // Check which letters lead to a self-loop without visiting an accepting
            // state (in the Büchi semantics). For this, we successively compute pairs with the state
            // as the first element, and one of the letters that occured as the second one.
            std::set<boost::tuple<int,int> > todoList;
            std::set<boost::tuple<int,int> > doneList;
            todoList.insert(boost::make_tuple(p,-1));
            doneList.insert(boost::make_tuple(p,-1));

            while (todoList.size()>0) {

                // Read next tuple
                boost::tuple<int,int> thisTuple = *todoList.begin();
                todoList.erase(todoList.begin());

                for (int b=0;b<dba.getSizeSigma();b++) {
                    int target1 = dba.getTransition(thisTuple.get<0>(),b);
                    if (!(dba.isAcceptingState(target1))) {

                        boost::tuple<int,int> t1 = boost::make_tuple(target1,thisTuple.get<1>());
                        boost::tuple<int,int> t2 = boost::make_tuple(target1,b);

                        if (doneList.count(t1)==0) {
                            doneList.insert(t1);
                            todoList.insert(t1);
                        }

                        if (doneList.count(t2)==0) {
                            doneList.insert(t2);
                            todoList.insert(t2);
                        }
                    }
                }
            }

            // Collect the letters that may occur on a loop:
            std::set<int> letters;
            for (std::set<boost::tuple<int,int> >::iterator it = doneList.begin();it!=doneList.end();it++) {
                if (it->get<0>()==p) {
                    int letter = it->get<1>();
                    if (letter!=-1) letters.insert(letter);
                }
            }

            // Make sure that there is no Büchi-accepting lasso starting from p when using only letters
            // in the set "letters". In the next few lines, a non-optimal construction is used, but for
            // the automata dealt with here, it shouldn't make a difference
            doneList.clear();
            todoList.clear();
            doneList.insert(boost::make_tuple(p,0));
            todoList.insert(boost::make_tuple(p,0));

            while (todoList.size()>0) {

                // Read next tuple
                boost::tuple<int,int> thisTuple = *todoList.begin();
                todoList.erase(todoList.begin());

                for (std::set<int>::const_iterator it = letters.begin();it!=letters.end();it++) {
                    int target = dba.getTransition(thisTuple.get<0>(),*it);
                    boost::tuple<int,int> nextTuple = boost::make_tuple(target,thisTuple.get<1>()+(dba.isAcceptingState(target)?1:0));

                    if (doneList.count(nextTuple)==0) {
                        doneList.insert(nextTuple);
                        todoList.insert(nextTuple);

                        if (nextTuple.get<1>()>dba.getNofStates()) {
                            // Found a witness that it is not in the intersection
                            std::cerr << "This automaton is not in the intersection of ACTL and LTL! Reason (2)\n";
                            return 1;
                        }
                    }
                }
            }
        }
    }
    return 0;
}
