#!/usr/bin/python
# -*- coding: ISO-8859-15 -*-
#
# Takes a specification file and converts all LTLXml parts into Vermicelli.
# 
# Assumes that the XML file has already been processed by "parseFormulasInXMLSpecFile.py",
# so there are no illegal atomic proposition usages, etc.

import math
import os
import sys
import resource
import subprocess
import signal
import binascii
import tempfile, copy
from parser import Parser
from xml import dom
import xml
from xml.dom.minidom import Node, parse, parseString
from xml.dom.pulldom import SAX2DOM

#======================================================
# Obtain parts to the Tools
#======================================================
basepath =  os.path.realpath(__file__)
basepath = basepath[0:basepath.rfind("/Tools/")]
print >>sys.stderr,"Base path: "+basepath

#======================================================
# Reading the input file
#======================================================
if len(sys.argv)<2:
    print "Error: You must provide an input file name."
    sys.exit(1)

try:
    xmlDoc = parse(open(sys.argv[1]))
except OSError,e:
    print >>sys.stderr, "Error: Could not read input file: " + sys.argv[1]
    sys.exit(1)
except IOError,e:
    print >>sys.stderr,"Error: Could not read input file: " + sys.argv[1]
    sys.exit(1)
except xml.parsers.expat.ExpatError: # TODO: In which package is ExpatError actually contained?
    print >>sys.stderr,"Error: Cannot read XML input file. Check with 'xmllint' for validity."
    sys.exit(1)

#======================================================
# Temporary directory handling
#======================================================
def cleanUpTempDir(tempdir):
    if not os.environ.has_key("NOCLEANUP"):
        for myfile in ["formula.ltl","rabin_automaton","buechi_automaton","buechi_automaton_min","vermicelli.txt"]:
            try:        
                os.remove(tempdir+os.sep+myfile)
            except OSError:
                # Ok, file probably wasn't there. Then it's all right...
                pass
        os.rmdir(tempdir)

# Create temp dir
tempdir = tempfile.mkdtemp("")

# =============================================================
# Convert LTL formula into prenex form
# =============================================================
def filterEntities(u):
    res = []
    for a in u:
        if a.nodeType==a.ELEMENT_NODE:
            res.append(a)
    return res

def ltlToPrenexForm(u):
    filtered = filterEntities(u.childNodes)
    if (u.tagName=="G") or (u.tagName=="F") or (u.tagName=="X"):
        return u.tagName+" "+ltlToPrenexForm(filtered[0])
    if (u.tagName=="Not"):
        return "! "+ltlToPrenexForm(filtered[0])
    if (u.tagName=="U"):
        return "U "+ltlToPrenexForm(filtered[0])+" "+ltlToPrenexForm(filtered[1])
    if (u.tagName=="WU"):
        return "W "+ltlToPrenexForm(filtered[0])+" "+ltlToPrenexForm(filtered[1])
    if (u.tagName=="Iff"):
        return "e "+ltlToPrenexForm(filtered[0])+" "+ltlToPrenexForm(filtered[1])
    if (u.tagName=="And"):
        parts = []
        for a in filtered:
            parts.append(ltlToPrenexForm(a))
        if len(parts)==0:
            return "t"
        res = ""
        for i in xrange(0,len(parts)-1):
            res += "& "
        return res+" ".join(parts)
    if (u.tagName=="Or"):
        parts = []
        for a in filtered:
            parts.append(ltlToPrenexForm(a))
        if len(parts)==0:
            return "f"
        res = ""
        for i in xrange(0,len(parts)-1):
            res += "| "
        return res+" ".join(parts)
    if (u.tagName=="Xor"):
        parts = []
        for a in filtered:
            parts.append(ltlToPrenexForm(a))
        if len(parts)==0:
            print "Error: Xor nodes must not be empty!"
            sys.exit(1)
        res = ""
        for i in xrange(0,len(parts)-1):
            res += "^ "
        return res+" ".join(parts)
    if (u.tagName=="True"):
        return "t"
    if (u.tagName=="False"):
        return "f"
    if (u.tagName=="Var"):
        res = ""
        for a in u.childNodes:
            if a.nodeType==a.TEXT_NODE:
                res += a.wholeText
        return "v"+binascii.hexlify(res)
    else:
        print >>sys.stderr,"Error: Unknown LTL XML node type:" + u.tagName
        sys.exit(1)


# =============================================================
# Convert to Vermicelli set - the cheap and incomplete method
# using ltl2ba 
# =============================================================
def convertToLTL2BAForm(u):
    filtered = filterEntities(u.childNodes)
    if (u.tagName=="G"):
        return "[] "+convertToLTL2BAForm(filtered[0])
    if (u.tagName=="F"):
        return "<> "+convertToLTL2BAForm(filtered[0])
    if (u.tagName=="X"):
        return "X "+convertToLTL2BAForm(filtered[0])
    if (u.tagName=="Not"):
        return "! "+convertToLTL2BAForm(filtered[0])
    if (u.tagName=="U"): 
        return "( "+convertToLTL2BAForm(filtered[0])+" U "+convertToLTL2BAForm(filtered[1])+" )"
    if (u.tagName=="Iff"): 
        return "( "+convertToLTL2BAForm(filtered[0])+" <-> "+convertToLTL2BAForm(filtered[1])+" )"
    if (u.tagName=="WU"): 
        return "( "+convertToLTL2BAForm(filtered[0])+" w "+convertToLTL2BAForm(filtered[1])+" )"
    if (u.tagName=="True"):
        return "true"
    if (u.tagName=="False"):
        return "false"
    if (u.tagName=="And"):
        parts = []
        for a in filtered:
            parts.append(convertToLTL2BAForm(a))
        return "( "+" && ".join(parts)+" )"
    if (u.tagName=="Or"):
        parts = []
        for a in filtered:
            parts.append(convertToLTL2BAForm(a))
        return "( "+" || ".join(parts)+" )"
    if (u.tagName=="Var"):
        res = ""
        for a in u.childNodes:
            if a.nodeType==a.TEXT_NODE:
                res += a.wholeText
        return "v"+binascii.hexlify(res)
    else:
        print >>sys.stderr,"Error: Unknown LTL XML node type (2):" + u.tagName
        sys.exit(1)    

def getAPs(u):
    filtered = filterEntities(u.childNodes)
    if (u.tagName=="Var"):
        res = ""
        for a in u.childNodes:
            if a.nodeType==a.TEXT_NODE:
                res += a.wholeText
        return frozenset(["v"+binascii.hexlify(res)])
    else:
        result = set([])
        for a in filtered:
            result |= getAPs(a)
        return frozenset(result)


def computeVermicelli(states,initState,endState):
    if (initState==endState):
        return "1 0"
    transitions = []
    selfLoop = ""
    for (a,b) in states[initState].items():
        if a!=initState:
            result = computeVermicelli(states,a,endState)
            if result != "":
                transitions.append("["+b.replace(" ","-")+"]t "+result)
        else:
            selfLoop = "["+b.replace(" ","-")+"]* 0"
    if len(transitions)!=0:
        if selfLoop!="":
            return "0 1 "+selfLoop+" "+str(len(transitions))+" "+" ".join(transitions)
        else:
            return "0 "+str(len(transitions))+" "+" ".join(transitions)
    else:
        return ""

def convertToVermicelliSetIncompleteMethod(u):

    # Generate LTL formula as used by ltl2ba
    ltl2baForm = convertToLTL2BAForm(filterEntities(u.childNodes)[0])
    print >>sys.stderr,ltl2baForm

    # Call ltl2ba
    ltl2baCmd = basepath+"/Tools/ltl2ba-1.1/ltl2ba -f \"!("+ltl2baForm+")\" > "+tempdir+os.sep+"buechi_automaton"
    print >>sys.stderr," -- "+ltl2baCmd
    sys.stdout.flush()
    return_code = os.system(ltl2baCmd)

    if return_code>0:
        print "Error: Buechi automaton could not be built!"
        cleanUpTempDir(tempdir)
        sys.exit(100)

    # Sum up atomic propositions
    apList = list(getAPs(u))

    # Read in ltl2ba result
    neverclaim = open(tempdir+os.sep+"buechi_automaton","r")
    lines = neverclaim.readlines()
    neverclaim.close()

    if not lines[0].startswith("never"):
        print >>sys.stderr, "Unexpected line start in a result by ltl2ba - Did not find never claim."
        print >>sys.stderr, "File: "+tempdir+os.sep+"buechi_automaton"
        sys.exit(1)

    # Parse never claim (without interpreting the transition conditions)
    states = {}
    current_state = None
    for line in lines[1:]:
        try:
            possible_label_name = line[0:line.index(":")].strip()
            if len(possible_label_name) > 0:
                current_state = possible_label_name
                states[current_state] = []
            possible_transition = line[line.index(":")+1:]
            if possible_transition[0]==':':
                # All transitions start with "::"
                possible_transition = possible_transition[1:].strip()
                try:
                    split_point = possible_transition.index(" -> goto ")
                    transition_condition = possible_transition[0:split_point]
                    transition_target = possible_transition[split_point+9:]
                    states[current_state].append((transition_condition,transition_target))
                except ValueError:
                    print >>sys.stderr, "Unexpected line start in a result by ltl2ba - Cannot identify transition: "+line.strip()
                    print >>sys.stderr, "File: "+tempdir+os.sep+"buechi_automaton"
                    sys.exit(1)
        except ValueError:
            # No ":" found. - Must be a command line
            if line.strip()=="skip":
                states[current_state].append(("(1)",current_state))
            elif line.strip()=="if" or line.strip()=="fi;" or line.strip()=="}" or line.strip()=="{" or line.strip()=="false;":
                # We can ignore these lines
                pass
            else:
                print >>sys.stderr,"Unexpected line in a result by ltl2ba: "+line.strip()
                print >>sys.stderr, "File: "+tempdir+os.sep+"buechi_automaton"
                sys.exit(1)

    # Now check if there is a loop other than a self_loop
    transitiveClosure = set([])
    for a in states.keys():
        transitiveClosure.add((a,a))
    nextSet = frozenset([])
    while len(nextSet)<>len(transitiveClosure):
        nextSet = transitiveClosure
        transitiveClosure = set([])
        for (a,b) in nextSet:
            transitiveClosure.add((a,b))
            for (k,c) in states[b]:
                if (c==a) and (a<>b):
                    # Found non-trivial loop
                    return None
                transitiveClosure.add((a,c))

    # Check the assumption that there are no multiple edges between a pair of states
    # and convert to a dictionary
    for (a,b) in states.items():
        checker = set([])
        newDict = {}
        for (c,d) in b:
            if d in checker:
                print >>sys.stderr, "Illegal double-transition output by ltl2ba"
                print >>sys.stderr, "File: "+tempdir+os.sep+"buechi_automaton"
                sys.exit(1)
            checker.add(d)
            newDict[d] = c
        states[a] = newDict
        

    
    # Compute vermicelli set:
    # 0: List Atomic propositions
    line = str(len(apList))
    for a in apList:
        line +=" "+a

    # 1: Iterate over the vermicelli end components
    vermicelli = []
    for endState in states.keys():
        if (endState in states[endState].keys()) and (endState.find("accept")!=-1):
            # This state has self-loops, so list the incoming transitions
            foundInitState = False
            for initState in states.keys():
                if initState.find("init")!=-1:
                    if foundInitState:
                        print >>sys.stderr, "Output by ltl2ba has more than one initial state."
                        print >>sys.stderr, "File: "+tempdir+os.sep+"buechi_automaton"
                        sys.exit(1)
                    foundInitState = True
                    vermicelli.append("["+states[endState][endState].replace(" ","-")+"] "+computeVermicelli(states,initState,endState))

    line += " "+str(len(vermicelli))+" "+" ".join(vermicelli)

    # Return vermicelli
    vermicelliXMLElement = xmlDoc.createElement("VermicelliSet")
    vermicelliXMLElement.appendChild(xmlDoc.createTextNode(line))
    return vermicelliXMLElement


# =============================================================
# Convert to Vermicelli set - the general way 
# =============================================================
def convertToVermicelliSetFullMethod(node):

    # Write file that we start with
    prenexForm = ltlToPrenexForm(filterEntities(node.childNodes)[0])
    ltlFile = open(tempdir+os.sep+"formula.ltl","w")
    ltlFile.write(prenexForm)
    ltlFile.close()


    # LTL->Rabin
    ltl2dstarCmd = basepath+"/Tools/ltl2dstar-0.5.1/src/ltl2dstar --ltl2nba=spin:"+basepath+"/Tools/ltl2ba-1.1/ltl2ba --automata=rabin --output=automaton "+tempdir+os.sep+"formula.ltl "+tempdir+os.sep+"rabin_automaton"
    print >>sys.stderr," -- "+ltl2dstarCmd
    sys.stdout.flush()
    return_code = os.system(ltl2dstarCmd)

    if return_code>0:
        print "Error: Rabin automaton could not be built!"
        cleanUpTempDir(tempdir)
        sys.exit(100)

    # Rabin->DBA
    rabin2dbaCmd = basepath+"/Tools/DBABuilder/buechinizer "+tempdir+os.sep+"rabin_automaton > "+tempdir+os.sep+"buechi_automaton"
    print >>sys.stderr," -- " + rabin2dbaCmd
    sys.stdout.flush()
    return_code = os.system(rabin2dbaCmd)

    if return_code>256:
        print "Error: Error Buechinizing deterministic Rabin automaton."
        cleanUpTempDir(tempdir)
        sys.exit(100)
    elif return_code==256:
        print "Error: A specification part is not representable as deterministic Büchi automaton. Such specifications lie beyond the scope of this tool set."
        print "The XML representation of this specification part is: "+node.toxml()
        cleanUpTempDir(tempdir)
        sys.exit(100)

    # Log as example that we need the vermicelli-based construction.
    # f = open("/tmp/complicated.txt","a")
    # f.write(prenexForm+"\n")
    # f.close()
    # sys.exit(1)

    # Minimize DBA using some simple construction
    buechiMinCmd = basepath+"/Tools/DBABuilder/minimize.py --timeout=30 "+tempdir+os.sep+"buechi_automaton > "+tempdir+os.sep+"buechi_automaton_min"
    print >>sys.stderr," -- " + buechiMinCmd
    sys.stdout.flush()
    return_code = os.system(buechiMinCmd)
    if return_code>0:
        print "Error: Could not minimize Buechi automaton!"
        cleanUpTempDir(tempdir)
        sys.exit(100)

    # Test if convertable
    uvwCheckCmd = basepath+"/Tools/InIntersectionTester/tester "+tempdir+os.sep+"buechi_automaton_min"
    print >>sys.stderr," -- " + uvwCheckCmd
    sys.stdout.flush()
    return_code = os.system(uvwCheckCmd)
    representableAsUVW = True

    if (return_code/256)==1:
        print "Error: A part of the specification is not representable as UVW"
        cleanUpTempDir(tempdir)
        sys.exit(1)
    elif return_code==100*256:
        print "Fatal Error: DBA is not valid."
        cleanUpTempDir(tempdir)
        sys.exit(100)
    elif return_code==0:
        pass
        # Everything ok
    else:
        print "Error: Cannot interpret 'tester' return value"
        print "Error value: "+str(return_code)
        cleanUpTempDir(tempdir)
        sys.exit(100)

    # Convert!
    uvwConvertCmd = basepath+"/Tools/DBAtoVermicelliSet/converter "+tempdir+os.sep+"buechi_automaton_min > "+tempdir+os.sep+"vermicelli.txt"
    print >>sys.stderr," -- " + uvwConvertCmd
    sys.stdout.flush()
    return_code = os.system(uvwConvertCmd)

    if return_code>0:
        print "Fatal Error: DBAtoVermicelliSet Converter crashed."
        cleanUpTempDir(tempdir)
        sys.exit(100)
    
    # Add vermicelli
    vermicelliXMLElement = xmlDoc.createElement("VermicelliSet")
    vermicelliFile = open(tempdir+os.sep+"vermicelli.txt","r")
    line = vermicelliFile.readline().strip()
    if (len(vermicelliFile.readline())>0):
        print "Fatal Error: DBAtoVermicelliSet tool produced more than one line of output. This is unexpected!"
        cleanUpTempDir(tempdir)
        sys.exit(100)
    vermicelliXMLElement.appendChild(xmlDoc.createTextNode(line.replace("\"","")))

    return vermicelliXMLElement


# Process the input File
changedSomething = True
while changedSomething:
    ltl_nodes = xmlDoc.getElementsByTagName('LTLXML')
    changedSomething = False
    for n in ltl_nodes:
        for u in n.parentNode.childNodes:
            if (u.nodeName=="LTLXML"):
                vermicelliSet = convertToVermicelliSetIncompleteMethod(u)
                if vermicelliSet==None: # Couldn't convert with the incomplete method
                    vermicelliSet = convertToVermicelliSetFullMethod(u)
                n.parentNode.replaceChild(vermicelliSet,u)
                changedSomething = True
                break
        if changedSomething:
            break

# Output the result
cleanUpTempDir(tempdir)
print xmlDoc.toxml()
sys.exit(0)

