#!/usr/bin/python
# -*- coding: ISO-8859-15 -*-
#
# Takes a specification file and generates NuSMV verification obligations.
# Performs the conversion from Mealy- to Moore-semantics
#
# Assumes that the XML file has already been processed by "parseFormulasInXMLSpecFile.py",
# so there are no illegal atomic proposition usages, etc.

import math
import os
import sys
import resource
import subprocess
import signal
import binascii
import tempfile, copy
from parser import Parser
from xml import dom
import xml
from xml.dom.minidom import Node, parse, parseString
from xml.dom.pulldom import SAX2DOM

#======================================================
# Reading the input file
#======================================================
if len(sys.argv)<2:
    print "Error: You must provide an input file name."
    sys.exit(1)

try:
    xmlDoc = parse(open(sys.argv[1]))
except OSError,e:
    print >>sys.stderr, "Error: Could not read input file: " + sys.argv[1]
    sys.exit(1)
except IOError,e:
    print >>sys.stderr,"Error: Could not read input file: " + sys.argv[1]
    sys.exit(1)
except xml.parsers.expat.ExpatError: # TODO: In which package is ExpatError actually contained?
    print >>sys.stderr,"Error: Cannot read XML input file. Check with 'xmllint' for validity."
    sys.exit(1)

# =============================================================
# Convert to NuSMV spec
# =============================================================
def filterEntities(u):
    res = []
    for a in u:
        if a.nodeType==a.ELEMENT_NODE:
            res.append(a)
    return res

def convertToNUSMVForm(u):
    filtered = filterEntities(u.childNodes)
    if (u.tagName=="G"):
        return "G "+convertToNUSMVForm(filtered[0])
    if (u.tagName=="F"):
        return "F "+convertToNUSMVForm(filtered[0])
    if (u.tagName=="X"):
        return "X "+convertToNUSMVForm(filtered[0])
    if (u.tagName=="Not"):
        return "! "+convertToNUSMVForm(filtered[0])
    if (u.tagName=="U"): 
        return "( "+convertToNUSMVForm(filtered[0])+" U "+convertToNUSMVForm(filtered[1])+" )"
    if (u.tagName=="Iff"): 
        return "( "+convertToNUSMVForm(filtered[0])+" <-> "+convertToNUSMVForm(filtered[1])+" )"
    if (u.tagName=="WU"): 
        return "( "+convertToNUSMVForm(filtered[0])+" w "+convertToNUSMVForm(filtered[1])+" )"
    if (u.tagName=="True"): 
        return "TRUE"
    if (u.tagName=="False"): 
        return "FALSE"
    if (u.tagName=="And"):
        parts = []
        for a in filtered:
            parts.append(convertToNUSMVForm(a))
        return "( "+" & ".join(parts)+" )"
    if (u.tagName=="Or"):
        parts = []
        for a in filtered:
            parts.append(convertToNUSMVForm(a))
        return "( "+" | ".join(parts)+" )"
    if (u.tagName=="Var"):
        res = ""
        for a in u.childNodes:
            if a.nodeType==a.TEXT_NODE:
                res += a.wholeText
        if res in inputs:
            return "in_"+res
        elif res in outputs:
            return "X out_"+res
        else:
            print >>sys.stderr,"Error: The atomic proposition " + res + " used in the specification in neither declared as input nor output."
            sys.exit(1)
    else:
        print >>sys.stderr,"Error: Unknown LTL XML node type (2):" + u.tagName
        sys.exit(1)    

def getAPs(u):
    filtered = filterEntities(u.childNodes)
    aps = []
    for v in filtered:
        if (v.tagName=="Bit"):
            res = ""
            for a in v.childNodes:
                if a.nodeType==a.TEXT_NODE:
                    res += a.wholeText
            aps.append(res)
    return aps

# =============================================
# Process a specification group
# =============================================
def recurseAnalyseSpecification(node):
    if node.nodeName=="PropertyGroup":
        if node.getAttribute("type")=="Assumption-Guarantee":
            groups = []
            for a in node.childNodes:
                if a.nodeType==a.ELEMENT_NODE:
                    groups.append(recurseAnalyseSpecification(a))
            if len(groups)!=2:
                print >>sys.stderr, "Error analysing PropertyGroup: Assume-Guarantee Group has <>2 elements."
                sys.exit(1)
            complete = []
            for g in groups[1]:
                result = "((TRUE "
                for a in groups[0]:
                    result += " & ("+a+")"
                result += ") -> ("+g+"))"
                complete.append(result)
            return complete
        elif node.getAttribute("type")=="Conjunction":
            complete = []
            for a in node.childNodes:
                if a.nodeType==a.ELEMENT_NODE:
                    complete.extend(recurseAnalyseSpecification(a))
            return complete
        else:
            print >>sys.stderr, "Error analysing PropertyGroup: Unknown type : "+node.getAttribute("type")
            sys.exit(1)
    elif node.nodeName=="Specification":
        complete = []
        for a in node.childNodes:
            if a.nodeType==a.ELEMENT_NODE:
                complete.extend(recurseAnalyseSpecification(a))
        return complete
    elif node.nodeName=="LTLXML":
        return [convertToNUSMVForm(filterEntities(node.childNodes)[0])]
    else:
        print "Node type cannot be interpreted (while generating NuSMV specification: " + node.nodeName
        print node

# Process the input File
inputs = getAPs(xmlDoc.getElementsByTagName('Inputs')[0])
outputs = getAPs(xmlDoc.getElementsByTagName('Outputs')[0])

# Check that we are really working with Mealy-mode semantics - the others are unsupported right now
semanticsNode = xmlDoc.getElementsByTagName('Semantics')
if len(semanticsNode)!=1:
    print >>sys.stderr, "Error: Could not find precisely one semantics node in the problem instance."
    sys.exit(1)
semanticsNode = semanticsNode[0]
try:
    semanticsType = semanticsNode.attributes["type"].value
except KeyError:
    print >>sys.stderr, "Error: Semantics node has no type attribute!"
    sys.exit(1)
if semanticsType!="Mealy":
    print >>sys.stderr, "Error: The Bassist part 'generateNuSMVVerificationObligations' currently only supports Mealy-type semantics, but '"+str(semanticsType)+"' semantics type was found instead."
    sys.exit(1)

# Now convert the specification
specificationNode = xmlDoc.getElementsByTagName('Specification')
specificationParts = recurseAnalyseSpecification(specificationNode[0])

for a in specificationParts:
    print "LTLSPEC "+a

