#!/usr/bin/python
#
# Parse all LTL formulas in a given XML input specification file into XML format.
# Checks sanity of input/output atomic proposition set

import math
import os
import sys
import resource
import subprocess
import signal
from parser import Parser
from xml import dom
from lxml import etree
import xml
from xml.dom.minidom import Node, parse, parseString
from xml.dom.pulldom import SAX2DOM
from xml.parsers.expat import ExpatError
from re import match
import StringIO

#======================================================
# Obtain parts to the Tools
#======================================================
basepath =  os.path.realpath(__file__)
basepath = basepath[0:basepath.rfind("/Tools/")]
print >>sys.stderr,"Base path: "+basepath

#======================================================
# Reading the input file
#======================================================
if len(sys.argv)<2:
    print >>sys.stderr, "Error: You must provide an input file name."
    sys.exit(1)

# First, load the DTD file
try:
    dtdFilename = basepath+"/HelperData/LinearTimeSynthesisProblem.dtd"
    f = open(dtdFilename,"r")
    dtdFile = "\n".join(f.readlines())
    f.close()
except OSError,e:
    print >>sys.stderr, "Error: Could not read the DTD file for syntax checking of the input instance (OSError)."
    sys.exit(1)
except IOError,e:
    print >>sys.stderr, "Error: Could not read the DTD file for syntax checking of the input instance (IOError)."
    sys.exit(1)

try:
    # First, parse and check the DTD
    lxmlParser = etree.XMLParser(dtd_validation=False)
    current_tree = etree.parse(sys.argv[1],lxmlParser)
    dtd = etree.DTD(StringIO.StringIO(dtdFile))
    if not dtd.validate(current_tree):
        print >>sys.stderr, "Error. The given XML file is not valid linear-time synthesis problem specification."
        print >>sys.stderr, dtd.error_log
        sys.exit(1)

    # Now, get the tree for usage in the following
    xmlDoc = parse(open(sys.argv[1]))
except etree.XMLSyntaxError, e:
    print >>sys.stderr, "Error: The given XML file is not valid: \n"+str(e)
    sys.exit(1)
except OSError,e:
    print >>sys.stderr, "Error: Could not read input file: " + sys.argv[1]
    sys.exit(1)
except IOError,e:
    print >>sys.stderr, "Error: Could not read input file: " + sys.argv[1]
    sys.exit(1)
except xml.parsers.expat.ExpatError: 
    print >>sys.stderr, "Error: Cannot read XML input file. Check with 'xmllint' for validity."
    sys.exit(1)

# Allocate global parser
p = Parser()

# =====================================================
# Lexer for the PSL formulas
# =====================================================
def tokenize(str):

    res = []
    while str:
        # Ignoring stuff        
        if str[0].isspace() or (str[0]=='\n'):
            str = str[1:]
            continue

        if str.startswith("FALSE"):
            str = str[5:]
            res.append(("FALSE",))
            continue

        if str.startswith("TRUE"):
            str = str[4:]
            res.append(("TRUE",))
            continue

        m = match('[a-zA-Z0-9_.]+', str)
        if m:
            # Special case: ["X","F","U","W","0","1"]
            if m.end()==1:
                if str[0] in ["X","F","G","U","W","0","1"]:
                    res.append((str[0],))
                    str = str[1:]
                else:
                    res.append(('id', m.group(0)))
                    str = str[m.end(0):]
            else:
                res.append(('id', m.group(0)))
                str = str[m.end(0):]
            continue

        res.append((str[0],))
        str = str[1:]
    return res

# =====================================================
# Simplify the specifications
# =====================================================
def clean_tree(tree):
    """ Cleans a parse Tree, i.e. removes brackets and so on """
    if tree[0] in p.terminals:
        return tree
    if (tree[0]=="Brackets"):
        return clean_tree(tree[2])
    elif (tree[0]=="Implication") and (len(tree)==2):
        return clean_tree(tree[1])
    elif (tree[0]=="Atomic") and (len(tree)==2):
        return clean_tree(tree[1])
    elif (tree[0]=="Conjunction") and (len(tree)==2):
        return clean_tree(tree[1])
    elif (tree[0]=="Biimplication") and (len(tree)==2):
        return clean_tree(tree[1])
    elif (tree[0]=="Disjunction") and (len(tree)==2):
        return clean_tree(tree[1])
    elif (tree[0]=="Xor") and (len(tree)==2):
        return clean_tree(tree[1])
    elif (tree[0]=="BinaryTemporalFormula") and (len(tree)==2):
        return clean_tree(tree[1])
    elif (tree[0]=="UnaryFormula") and (len(tree)==2):
        return clean_tree(tree[1])
    elif (tree[0]=="AtomicFormula"):
        if len(tree)!=2:
            print >>sys.stderr, "Error: AtomicFormula must have length 2"
            sys.exit(1)
        return clean_tree(tree[1])
    elif (tree[0]=="Implication"):
        return [tree[0],clean_tree(tree[1]),clean_tree(tree[3])]
    elif (tree[0]=="Conjunction"):
        return [tree[0],clean_tree(tree[1]),clean_tree(tree[3])]
    elif (tree[0]=="Biimplication"):
        return [tree[0],clean_tree(tree[1]),clean_tree(tree[3])]
    elif (tree[0]=="Disjunction"):
        return [tree[0],clean_tree(tree[1]),clean_tree(tree[3])]
    elif (tree[0]=="Xor"):
        return [tree[0],clean_tree(tree[1]),clean_tree(tree[3])]
    elif (tree[0]=="BinaryTemporalFormula"):
        return [tree[0],clean_tree(tree[1]),clean_tree(tree[2]),clean_tree(tree[3])]
    elif (tree[0]=="UnaryFormula"):
        return [tree[0],clean_tree(tree[1]),clean_tree(tree[2])]
    elif (tree[0]=="BinaryTemporalOperator"):
        # Remove the "superfluous indirection"
        return clean_tree(tree[1])
    elif (tree[0]=="UnaryTemporalOperator"):
        # Remove the "superfluous indirection"
        return clean_tree(tree[1])
    elif (tree[0]=="Assignment"):
        # Flatten "id" case
        A = [tree[0],tree[1][1]]
        A.extend(tree[2:])
        return A
    else:
        A = [tree[0]]
        for x in tree[1:]:
           A.append(clean_tree(x))
        return A

def flatten_as_much_as_possible(tree):
    """ Flattens nested disjunctions/conjunctions """
    # Ground case?
    if len(tree)==1:
        return tree
    if (type(tree)==type("A")) or (type(tree)==type(u"A")): # TODO: How to do this in the way intended?
        return tree
    newTree = []
    for a in tree:
        newTree.append(flatten_as_much_as_possible(a))
    tree = newTree

    # Conjunction
    if (tree[0]=="Conjunction"):
        parts = [tree[0]]
        for a in tree[1:]:
            if a[0]=="Conjunction":
                parts.extend(a[1:])
            else:
                parts.append(a)
        return parts

    # Disjunction
    if (tree[0]=="Disjunction"):
        parts = [tree[0]]
        for a in tree[1:]:
            if a[0]=="Disjunction":
                parts.extend(a[1:])
            else:
                parts.append(a)
        return parts

    # Xor
    if (tree[0]=="Xor"):
        parts = [tree[0]]
        for a in tree[1:]:
            if a[0]=="Xor":
                parts.extend(a[1:])
            else:
                parts.append(a)
        return parts

    # Every other case
    return tree


# =====================================================
# Parse the Specification to XML form
# =====================================================
def specToXML(tree):
    """ Cleans a parse Tree, i.e. removes brackets and so on """
    if (tree[0]=="Brackets"):
        print >>sys.stderr, "Error: Brackets should have been parsed away already."
        sys.exit(1)
    elif (tree[0]=="Implication"):
        if len(tree)!=3:
            print >>sys.stderr, "Error: Implication doesn't have correct size."
            sys.exit(1)
        newNode = xmlDoc.createElement("Or")
        newNode2 = xmlDoc.createElement("Not")
        newNode.appendChild(newNode2)
        newNode2.appendChild(specToXML(tree[1]))
        newNode.appendChild(specToXML(tree[2]))
        return newNode
    elif (tree[0]=="Biimplication"):
        if len(tree)!=3:
            print >>sys.stderr, "Error: Biimplication doesn't have correct size."
            sys.exit(1)
        newNode = xmlDoc.createElement("Iff")
        newNode.appendChild(specToXML(tree[1]))
        newNode.appendChild(specToXML(tree[2]))
        return newNode
    elif (tree[0]=="Conjunction"):
        newNode = xmlDoc.createElement("And")
        for a in tree[1:]:
            newNode.appendChild(specToXML(a))
        return newNode
    elif (tree[0]=="Disjunction"):
        newNode = xmlDoc.createElement("Or")
        for a in tree[1:]:
            newNode.appendChild(specToXML(a))
        return newNode
    elif (tree[0]=="UnaryFormula"):
        if len(tree)!=3:
            print >>sys.stderr, "Error: Unary subformula should only work with unary temporal operators or negation!"
            sys.exit(1)
        newNode = None
        if tree[1][0]=="GloballyOperator":
            newNode = xmlDoc.createElement("G")
        if tree[1][0]=="NotOperator":
            newNode = xmlDoc.createElement("Not")
        if tree[1][0]=="FinallyOperator":
            newNode = xmlDoc.createElement("F")
        if tree[1][0]=="NextOperator":
            newNode = xmlDoc.createElement("X")
        if (newNode==None):
            print >>sys.stderr, "Error: Unary subformula uses unknown temporal operator."
            print >>sys.stderr, tree[1]
            sys.exit(1)
        newNode.appendChild(specToXML(tree[2]))
        return newNode
    elif (tree[0]=="BinaryTemporalFormula"):
        if len(tree)!=4:
            print "Error: Unary subformula should only work with unary temporal operators or negation!"
            sys.exit(1)
        newNode = None
        if tree[2][0]=="UntilOperator":
            newNode = xmlDoc.createElement("U")
        if tree[2][0]=="WeakUntilOperator":
            newNode = xmlDoc.createElement("WU")
        if (newNode==None):
            print >>sys.stderr, "Error: BinaryTemporalOperator subformula uses unknown temporal operator."
            print >>sys.stderr, tree[1]
            sys.exit(1)
        newNode.appendChild(specToXML(tree[1]))
        newNode.appendChild(specToXML(tree[3]))
        return newNode
    elif tree[0]=="FALSE":
        return xmlDoc.createElement("False")
    elif tree[0]=="TRUE":
        return xmlDoc.createElement("True")
    elif tree[0]=="Assignment":
        if len(tree)==2:
            newNode = xmlDoc.createElement("Var")
            newNode.appendChild(xmlDoc.createTextNode(tree[1]))
            return newNode
        elif len(tree)==4:
            newNode = xmlDoc.createElement("Var")
            newNode.appendChild(xmlDoc.createTextNode(tree[1]))
            if tree[3]==("1",):
                return newNode
            elif tree[3]==("0",):
                newNode2 = xmlDoc.createElement("Not")
                newNode2.appendChild(newNode)
                return newNode2
            else:
                print >>sys.stderr, "Error: Couldn't parse old-style assignment"
                print >>sys.stderr, "Can't identify value: " + str(tree[3])
                sys.exit(1)
        else:
            print >>sys.stderr, "Error: Could not identify assignment type"
            sys.exit(1)
    else:
        print >>sys.stderr, "Error: specToXML: Don't know what to do with: " + str(tree)
        sys.exit(1)


# =====================================================
# The Parsing function
# =====================================================
def parseLTL(u):
    # Sum up all Text node
    ltlTxt = ""
    for c in u.childNodes:
        if c.nodeType == c.TEXT_NODE:
            ltlTxt += c.wholeText

    # Parse LTL formula
    # print ltlTxt

    try:
        input = tokenize(ltlTxt)
        tree = p.parse(input)

    except p.ParseErrors, exception:
        for t,e in exception.errors:
            if t[0] == p.EOF:
                print >>sys.stderr, "Formula end not expected here"
                continue

            found = repr(t[0])
            if len(e) == 1:
                print >>sys.stderr, "Expected %s, but found %s " %(repr(e[0]), found)
            else:
                print >>sys.stderr, "Could not parse %s, "%found
                print >>sys.stderr, "Wanted a token of one of the following forms: "+", ".join([ repr(s) for s in e ])
        sys.exit(1)

    # Convert to a tree
    cleaned_tree = clean_tree(tree)
    # print cleaned_tree
    simplified_tree = flatten_as_much_as_possible(cleaned_tree)
    # print simplified_tree
    parsedLTLFormula = specToXML(simplified_tree)
    # print parsedLTLFormula


    # Replace by new parsed LTL formula
    newNode = xmlDoc.createElement("LTLXML")
    newNode.appendChild(parsedLTLFormula)
    newNode.setAttribute("PositionInOriginalXMLFile","Undefined")
    # print dir(newNode)
    # print dir(xmlDoc)
    # PositionInOriginalXMLFile
    return newNode

# Get Input&Output variables
atomicPropositions = []
atomicPropositionsXML = xmlDoc.getElementsByTagName('Bit')
for u in atomicPropositionsXML:
    name = ""
    for c in u.childNodes:
        if c.nodeType == c.TEXT_NODE:
            name += c.wholeText
    atomicPropositions.append(name.strip())

# Check if the list of input&output variables is sane
listOfDisallowedAPNames = ["G","F","X","W","U","R","FALSE","TRUE","0","1"]
for a in atomicPropositions:
    u = a.upper()
    if u in listOfDisallowedAPNames:
        print >>sys.stderr, "Error in input file: '"+a+"' must not be used as a name for an input or output atomic proposition."
        print >>sys.stderr, "Disallowed atomic proposition names are, in any capitalization mode: " + str(listOfDisallowedAPNames)
        sys.exit(1)
    m = match('[a-zA-Z0-9_.]+', a)
    if m.end()<>len(a):
        print >>sys.stderr, "Error in input file: '"+a+"' is not a valid atomic proposition name. Atomic proposition names must only consist of numbers, underscores, points and upper- or lower-case letters without special characters."
        sys.exit(1)


# Process the input File
changedSomething = True
while changedSomething:
    ltl_nodes = xmlDoc.getElementsByTagName('LTL')
    changedSomething = False
    for n in ltl_nodes:
        for u in n.parentNode.childNodes:
            if (u.nodeName=="LTL"):
                n.parentNode.replaceChild(parseLTL(u),u)
                changedSomething = True
                break
        if changedSomething:
            break

# Output the result
print xmlDoc.toxml()
sys.exit(0)

