// A program that computes a UVW representation of a given
// vermicelli set.
//
// (c) 2011/2012 Rüdiger Ehlers

#include "buechi.hpp"
#include "BF.h"
#include "bddDump.h"
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>
#include <tr1/unordered_map>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <fstream>
#include <map>
#include <boost/algorithm/string.hpp>

//=============================================================
// A special class for managing pairs of states
// Interface-compatible to std::set<std::pair<uint,uint> >
//=============================================================
class UintPairSetWithMaximum : public boost::noncopyable {
    uint nofElements;
    uint nofCombinationsSet;
    bool* elements;
    std::pair<uint,uint>* combinationsSet;


public:
    UintPairSetWithMaximum(uint _nof) : nofElements(_nof) {
        elements = new bool[_nof*_nof];
        nofCombinationsSet = 0;
        combinationsSet = new std::pair<uint,uint>[_nof*_nof];
        memset(elements,0,sizeof(bool)*_nof*_nof);
    }
    ~UintPairSetWithMaximum() {
        delete[] elements;
        delete[] combinationsSet;
    }
    inline int count(std::pair<uint,uint> p) {
        return elements[p.first*nofElements+p.second];
    }
    inline void insert(std::pair<uint,uint> p) {
        uint pointer = p.first*nofElements+p.second;
        if (!elements[pointer]) {
            combinationsSet[nofCombinationsSet++] = p;
            elements[pointer] = true;
        }
    }
    inline uint size() {
        return nofCombinationsSet;
    }

    void clear() {
        nofCombinationsSet = 0;
        memset(elements,0,sizeof(bool)*nofElements*nofElements);
    }

    void replaceAllReferencesOfBy(uint from, uint to) {
        for (int i=0;i<(int)nofCombinationsSet;i++) {
            bool divert = (combinationsSet[i].first==from) || (combinationsSet[i].second==from);
            if (divert) {
                std::pair<uint,uint> newPair = combinationsSet[i];
                if (newPair.first==from) newPair.first = to;
                if (newPair.second==from) newPair.second = to;
                elements[combinationsSet[i].first*nofElements+combinationsSet[i].second] = false;
                if (elements[newPair.first*newPair.second]) {
                    // Copy back
                    combinationsSet[i--] = combinationsSet[--nofCombinationsSet];
                } else {
                    combinationsSet[i] = newPair;
                    elements[newPair.first*nofElements+newPair.second] = true;
                }
            }
        }
    }


    // Iterating
    class iterator {
        uint pointer;
        UintPairSetWithMaximum *parent;
        friend class UintPairSetWithMaximum;
        iterator(uint p,UintPairSetWithMaximum *_parent) : pointer(p), parent(_parent) {}

    public:
        // The assignment and relational operators are straightforward
        iterator& operator=(const iterator& a) {
            parent = a.parent;
            pointer = a.pointer;
            return *this;
        }

        bool operator==(const iterator &a) {
            return a.pointer==pointer;
        }

        bool operator!=(const iterator &a) {
            return a.pointer != pointer;
        }

        iterator& operator++() {
            pointer++;
            return *this;
        }

        iterator operator++(int) {
            iterator newOne(*this);
            pointer++;
            return newOne;
        }

        std::pair<uint,uint> &operator*() {
            return parent->combinationsSet[pointer];
        }

        // Return the address of the value referred to.
        std::pair<uint,uint> *operator->() {
            return &(parent->combinationsSet[pointer]);
        }
    };
    iterator begin() {
        return iterator(0,this);
    }
    iterator end() {
        return iterator(nofCombinationsSet,this);
    }

};


//=============================================================
// Algorithm container class
//=============================================================
#define VERBOSITY_LEVEL 0

class ConverterAlgorithm : public VariableInfoContainer {
private:

    // BF stuff
    BFManager mgr;
    std::vector<BF> vars;
    std::vector<std::string> varNames;
    std::vector<uint> varsAPs; // One variable per alphabet symbol

    // Tool functions
    static std::string toString(int i) {
        std::ostringstream os;
        os << i;
        return os.str();
    }

    //=================================
    // Main function - computes the UVW
    //=================================
    struct UVWState {
        bool isAccepting; // During initial Automaton building from vermicelli, misuse this flag is "incoming edge should also go to vermicelli end element"
        bool isInitial;
        bool isActive; // Set to false after removing a state
        std::vector<std::pair<uint,BF> > transitions;
    public:
        UVWState() : isAccepting(false), isInitial(false), isActive(true) {}
    };

    struct UVW {
        // Data
        ConverterAlgorithm *parent;
        std::vector<UVWState> states;

        UVW(ConverterAlgorithm *_parent) : parent(_parent) {}

        // Obtain a final self-looping state. Used for the "ends" of the vermicelli
        uint getSelfLoopingFinalState(BF inputs) {
            assert(inputs.getHashCode()>1);
            for (uint i=0;i<states.size();i++) {
                UVWState &state = states[i];
                if (state.isAccepting) {
                    if ((state.transitions.size()==1) && (state.transitions[0].first==i) && (state.transitions[0].second==inputs)) {
                        return i;
                    }
                }
            }
            // Create new state
            uint newState = states.size();
            states.push_back(UVWState());
            states[newState].isAccepting = true;
            states[newState].transitions.push_back(std::pair<uint,BF>(newState,inputs));
            BF_DUMPANDCHECK(*parent,inputs,"APs","/tmp/selfLoopInputs.dot")
            return newState;
        }

        UVW(std::vector<UVW> &uvws) : parent(uvws[0].parent) {
            std::vector<uint> offset;

            offset.push_back(0);
            for (uint i=1;i<uvws.size();i++) {
                uint next_offset = offset.back() + uvws[i-1].states.size();
                offset.push_back(next_offset);
            }
            for (uint i=0;i<uvws.size();i++) {
                int targetOffset = offset[i];
                for (uint j=0;j<uvws[i].states.size();j++) {
                    states.push_back(UVWState());
                    UVWState &source = uvws[i].states[j];
                    UVWState &target = states.back();
                    target.isAccepting = source.isAccepting;
                    target.isActive = source.isActive;
                    assert(source.isActive);
                    target.isInitial = source.isInitial;
                    for (auto it = source.transitions.begin();it!=source.transitions.end();it++) {
                        target.transitions.push_back(std::pair<uint,BF>(it->first+targetOffset,it->second));
                    }
                }
            }
        }

        // Joins self-looping states
        // Asserts that non-active states have been removed.
        void joinOnlySelfLoopingStatesWithSameLabelSet(BFManager &mgr) {
            std::vector<bool> onlySelfLoopingFinalStates;
            std::vector<BF> selfLoopingLabels;
            for (uint i=0;i<states.size();i++) {
                assert(states[i].isActive);
                bool isSelfLooping = states[i].isAccepting & (!states[i].isInitial);
                BF selfLoopingLabel = mgr.constantFalse();
                for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                    if(it->first==i) {
                        selfLoopingLabel |= it->second;
                    } else {
                        isSelfLooping = false;
                    }
                }
                selfLoopingLabels.push_back(selfLoopingLabel);
                onlySelfLoopingFinalStates.push_back(isSelfLooping);
            }
            std::map<size_t,uint> remappingBF;
            std::map<uint,uint> remapping;
            for (uint i=0;i<states.size();i++) {
                if (onlySelfLoopingFinalStates[i]) {
                    size_t hash = selfLoopingLabels[i].getHashCode();
                    if (remappingBF.count(hash)==0) {
                        remappingBF[hash] = i;
                    } else {
                        states[remappingBF[hash]].isInitial |= states[i].isInitial;
                    }
                }
            }
            for (uint i=0;i<states.size();i++) {
                if (!onlySelfLoopingFinalStates[i]) {
                    remapping[i] = i;
                } else {
                    assert(remappingBF.count(selfLoopingLabels[i].getHashCode())>0);
                    remapping[i] = remappingBF[selfLoopingLabels[i].getHashCode()];
                    if (remapping[i]!=i) states[i].isActive = false;
                }
            }

            for (uint i=0;i<states.size();i++) {
                std::map<uint,BF> tran;
                for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                    if (tran.count(remapping[it->first])==0) {
                        tran[remapping[it->first]] = it->second;
                    } else {
                        tran[remapping[it->first]] |= it->second;
                    }
                }
                states[i].transitions.clear();
                for (auto it = tran.begin();it!=tran.end();it++) {
                    states[i].transitions.push_back(std::pair<uint,BF>(it->first,it->second));
                    assert(states[it->first].isActive);
                }
            }
        }



        std::string getBFAsTreeString(BF whatBF, std::vector<BF> &whichVars, uint level) {
            if (level==whichVars.size()) {
                if ((whatBF==parent->mgr.constantFalse())) throw "Internal error in getBFAsTreeString ";
                if ((whatBF!=parent->mgr.constantTrue())) throw "Illegal support in getBFAsTreeString ";
                return "";
            }
            BF truePart = (whatBF & whichVars[level]).ExistAbstractSingleVar(whichVars[level]);
            BF falsePart = (whatBF & !whichVars[level]).ExistAbstractSingleVar(whichVars[level]);
            if (truePart==falsePart) {
                return "*" + getBFAsTreeString(truePart,whichVars,level+1);
            } else if (truePart==parent->mgr.constantFalse()) {
                return "0" + getBFAsTreeString(falsePart,whichVars,level+1);
            } else if (falsePart==parent->mgr.constantFalse()) {
                return "1" + getBFAsTreeString(truePart,whichVars,level+1);
            } else {
                // std::cout << "SPLIT: " << level << std::endl;
                std::string firstPart = getBFAsTreeString(falsePart,whichVars,level+1);
                std::string secondPart = getBFAsTreeString(truePart,whichVars,level+1);
                // std::cout << "F: " << firstPart << ", S: " << secondPart << std::endl;
                return "("+firstPart+","+secondPart+")";
            }
        }

        void print(std::vector<BF> &variablesToCareAbout) {
            std::cerr << "UVW with " << states.size() << " states:\n";
            for (uint i=0;i<states.size();i++) {
                if (states[i].isActive) {
                    std::cerr << " state " << i << ": ";
                    if (states[i].isInitial) std::cerr << "initial ";
                    if (states[i].isAccepting) std::cerr << "accepting ";
                    std::cerr << "\n";
                    for (uint j=0;j<states[i].transitions.size();j++) {
                        std::pair<uint,BF> thisTransition = states[i].transitions[j];
                        if (thisTransition.second!=parent->mgr.constantFalse()) {
                            std::cerr << "   trans. to " << thisTransition.first << ": ";
                            std::cerr << getBFAsTreeString(thisTransition.second,variablesToCareAbout,0) << "\n";
                        }
                    }
                }
            }
        }

        // Returns state Number of generated state
        uint recurseParse(std::istringstream &is, std::vector<BF> varsUsedInThisSet, std::vector<uint> starIncoming, BF selfLoopingLabel, uint vermicelliFinalState) {
            assert(selfLoopingLabel.getHashCode()>1);
            int accepting;
            is >> accepting;
            if (is.fail()) {
                std::cerr << "Error parsing vermicelli (5)\n";
                throw "Aborting.";
            }
            if ((accepting<0) || (accepting>1)) {
                std::cerr << "Error parsing vermicelli (6)\n";
                throw "Aborting.";
            }
            int nofChildren;
            is >> nofChildren;
            if (is.fail()) {
                std::cerr << "Error parsing vermicelli (7)\n";
                throw "Aborting.";
            }

            // Allocate new state
            uint newState = states.size();
            // std::cerr << "NewState: " << newState<< "\n";
            states.push_back(UVWState());
            states[newState].isAccepting = accepting>0;
            if (selfLoopingLabel!=parent->mgr.constantFalse()) {
                states.back().transitions.push_back(std::pair<uint,BF>(states.size()-1,selfLoopingLabel));
                if (accepting>0) {
                    states.back().transitions.push_back(std::pair<uint,BF>(vermicelliFinalState,selfLoopingLabel));
                }
            }

            // Recurse
            starIncoming.push_back(newState);
            for (int i=0;i<nofChildren;i++) {

                std::string transitionString;
                is >> transitionString;
                if (is.fail()) {
                    std::cerr << "Error parsing vermicelli (8)\n";
                    throw "Aborting.";
                }
                // std::cerr << "TransitionString : " << transitionString << "\n";
                BF transitionBF = parent->parseTransitionString(transitionString.substr(0,transitionString.length()-1),varsUsedInThisSet);
                assert(transitionBF.getHashCode()>1);
                assert(transitionBF!=parent->mgr.constantFalse());
                std::vector<uint> nextStarIncoming;
                BF nextSelfLoopingLabel;
                switch (transitionString[transitionString.length()-1]) {
                case '*':
                    nextStarIncoming = starIncoming;
                    nextSelfLoopingLabel = transitionBF;
                    break;
                case 't':
                    nextSelfLoopingLabel = parent->mgr.constantFalse();
                    break;
                default:
                    std::cerr << "Error parsing vermicelli (9)\n";
                    throw "Aborting.";
                }

                uint nextState = recurseParse(is,varsUsedInThisSet,nextStarIncoming,nextSelfLoopingLabel,vermicelliFinalState);
                for (auto it = starIncoming.begin();it!=starIncoming.end();it++) {
                    bool found = false;
                    assert(*it<states.size());
                    for (auto it2 = states[*it].transitions.begin();(it2!=states[*it].transitions.end()) & (!found);it2++) {
                        if (it2->first == nextState) {
                            found = true;
                            assert(it2->second.getHashCode()>1);
                            it2->second |= transitionBF;
                        }
                    }
                    if (!found) states[*it].transitions.push_back(std::pair<uint,BF>(nextState,transitionBF));
                }
                if (states[nextState].isAccepting) {
                    for (auto it = starIncoming.begin();(it!=starIncoming.end());it++) {
                        bool found = false;
                        for (auto it2 = states[*it].transitions.begin();(it2!=states[*it].transitions.end()) & (!found);it2++) {
                            if (it2->first == vermicelliFinalState) {
                                found = true;
                                it2->second |= transitionBF;
                            }
                        }
                        if (!found) {
                            states[*it].transitions.push_back(std::pair<uint,BF>(vermicelliFinalState,transitionBF));
                        }
                    }
                    states[nextState].isAccepting = false;
                }
            }
            return newState;
        }

        //====================================================================================
        // Here, we do a special case a the general minimisation method (below) that does not
        // change the fact that final states are self-looping
        //====================================================================================
        void onlyNonFinalAndNonFinalStateMergingSimpleSimulationBasedMinimisation(std::vector<BF> &variablesToCareAboutWhenPrinting) {
            (void)variablesToCareAboutWhenPrinting;

            // Start by computing the L>= relation
            UintPairSetWithMaximum lgeRelation(states.size());

            // 0. Compute L>= relation of the vermicelli end components
            for (uint i=0;i<states.size();i++) {
                lgeRelation.insert(std::pair<uint,uint>(i,i));
                if (states[i].isAccepting) {
                    for (uint j=0;j<states.size();j++) {
                        if (states[j].isAccepting) {
                            bool process = true;
                            assert(states[i].transitions.size()==1);
                            assert(states[i].transitions[0].first == i);
                            assert(states[j].transitions.size()==1);
                            assert(states[j].transitions[0].first == j);
                            if (process) {
                                if (states[i].transitions[0].second <= states[j].transitions[0].second) {
                                    if (VERBOSITY_LEVEL>1) std::cerr << i << ">=: " << j << "\n";
                                    lgeRelation.insert(std::pair<uint,uint>(i,j));
                                }
                            }
                        }
                    }
                }
            }

            // Saturate the "L>=" relation
            uint oldSizeLgeRelation = 0;
            while (oldSizeLgeRelation != lgeRelation.size()) {
                oldSizeLgeRelation = lgeRelation.size();

                // Update L>= relation
                for (uint i=0;i<states.size();i++) {
                    for (uint j=0;j<states.size();j++) {
                        if (lgeRelation.count(std::pair<uint,uint>(i,j))==0) {
                            // Check if L(i) >= L(j)
                            // For this, iterate over the successors and compute the
                            // respective inputs
                            std::vector<std::pair<uint,BF> > &transitionsI = states[i].transitions;
                            std::vector<std::pair<uint,BF> > &transitionsJ = states[j].transitions;
                            bool outerFound = true;
                            for (auto it = transitionsI.begin();it!=transitionsI.end();it++) {
                                uint succ1 = it->first;
                                BF inputsToFindLeft = it->second;

                                //BF_DUMPANDCHECK(*parent,inputsToFindLeft,"APs","/tmp/left.dot")
                                for (auto it2 = transitionsJ.begin();it2!=transitionsJ.end();it2++) {
                                    if (lgeRelation.count(std::pair<uint,uint>(succ1,it2->first))>0) {
                                        inputsToFindLeft &= !it2->second;
                                    } else if ((it2->first==j) && (succ1==i)) {
                                        // Reduce by non-accepting self-loops.
                                        if (!states[i].isAccepting) {
                                            inputsToFindLeft &= !it2->second;
                                        }
                                    }
                                    //BF_DUMPANDCHECK(*parent,inputsToFindLeft,"APs","/tmp/left.dot")
                                }
                                outerFound &= (inputsToFindLeft == parent->mgr.constantFalse());
                            }
                            if (outerFound) {
                                lgeRelation.insert(std::pair<uint,uint>(i,j));
                                if (VERBOSITY_LEVEL>1) std::cerr << "N:" << i << ">=: " << j << std::endl;
                            }
                        }
                    }
                }
            }

            // Compute reachability relation
            UintPairSetWithMaximum reachabilityRelation(states.size());
            for (uint i=0;i<states.size();i++) reachabilityRelation.insert(std::pair<uint,uint>(i,i));
            uint oldSizeReachabilityRelation = 0;
            while (oldSizeReachabilityRelation != reachabilityRelation.size()) {
                oldSizeReachabilityRelation = reachabilityRelation.size();
                // The following code only works for std::set<std::pair<uint,uint> > replacements with iterators that are stable under insertion
                for (auto it = reachabilityRelation.begin();it!=reachabilityRelation.end();it++) {
                    for (auto it2 = states[it->second].transitions.begin();it2!=states[it->second].transitions.end();it2++) {
                        if (it2->second!=parent->mgr.constantFalse()) reachabilityRelation.insert(std::pair<uint,uint>(it->first,it2->first));
                    }
                }
            }


            // Merge a few states based on simulation equivalence
            for (auto it = lgeRelation.begin();it!=lgeRelation.end();it++) {
                if (VERBOSITY_LEVEL>1) std::cerr << "Working on " << it->first << "," << it->second;
                if (it->first!=it->second) {
                    if (VERBOSITY_LEVEL>1) std::cerr << ".";
                    if (lgeRelation.count(std::pair<uint,uint>(it->second,it->first))>0) {
                        if (VERBOSITY_LEVEL>1) std::cerr << ".";
                        if (((reachabilityRelation.count(*it)>0)) || ((VERBOSITY_LEVEL>1) && (reachabilityRelation.count(std::pair<uint,uint>(it->second,it->first))==0))) {
                            if (VERBOSITY_LEVEL>1) std::cerr << ".";
                            if ((states[it->first].isActive) && (states[it->second].isActive)) {
                                if (VERBOSITY_LEVEL>1) std::cerr << ".";
                                states[it->first].isActive = false;
                                if (VERBOSITY_LEVEL>1) std::cerr << "Moving " << it->first << " into " << it->second << std::endl;
                                states[it->second].isInitial |= states[it->first].isInitial;
                                for (uint i=0;i<states.size();i++) {
                                    for (auto it2 = states[i].transitions.begin();it2!=states[i].transitions.end();it2++) {
                                        if (it2->first==it->first) {
                                            it2->first = it->second;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                if (VERBOSITY_LEVEL>1) std::cerr << std::endl;
            }

            // Recompute reachability relation after merging
            reachabilityRelation.clear();
            for (uint i=0;i<states.size();i++) reachabilityRelation.insert(std::pair<uint,uint>(i,i));
            oldSizeReachabilityRelation = 0;
            while (oldSizeReachabilityRelation != reachabilityRelation.size()) {
                oldSizeReachabilityRelation = reachabilityRelation.size();
                // The following code only works for std::set<std::pair<uint,uint> > replacements with iterators that are stable under insertion
                for (auto it = reachabilityRelation.begin();it!=reachabilityRelation.end();it++) {
                    for (auto it2 = states[it->second].transitions.begin();it2!=states[it->second].transitions.end();it2++) {
                        reachabilityRelation.insert(std::pair<uint,uint>(it->first,it2->first));
                    }
                }
            }

            // Merge equiv-reachable states: (1) Compute inverse transition relation
            // Compute inverse transition relation
            std::vector<std::vector<std::pair<uint,BF> > > inverseTransitionRelation;
            inverseTransitionRelation.resize(states.size());
            for (uint i=0;i<states.size();i++) {
                for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                    uint accessIndex = it->first;
                    inverseTransitionRelation[accessIndex].push_back(std::pair<uint,BF>(i,it->second));
                }
            }

            // (2) Compute initial states
            UintPairSetWithMaximum rgeRelation(states.size());
            for (uint i=0;i<states.size();i++) {
                if (states[i].isActive) rgeRelation.insert(std::pair<uint,uint>(i,i));
            }
            uint oldSizeRgeRelation = 0;
            while (oldSizeRgeRelation != rgeRelation.size()) {
                oldSizeRgeRelation = rgeRelation.size();

                // Update L>= relation
                for (uint i=0;i<states.size();i++) {
                    for (uint j=0;j<states.size();j++) {
                        // std::cerr << "State " << j << ", isActive: " << states[j].isActive << "\n";
                        if (states[i].isActive && states[j].isActive && (states[i].isInitial || !states[j].isInitial)) {

                            // Check proper self-looping
                            // std::cerr << "Initial IS look: " << i << "," << j << std::endl;
                            BF selfLoops = parent->mgr.constantFalse();
                            for (auto it = states[j].transitions.begin();it!=states[j].transitions.end();it++) {
                                if (it->first == j) selfLoops |= it->second;
                            }
                            for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                                if (it->first == i) selfLoops &= !it->second;
                            }

                            if (selfLoops==parent->mgr.constantFalse()) {
                                // std::cerr << "Closer IS look: " << i << "," << j << std::endl;
                                // Check incoming edges
                                bool mismatchFound = false;
                                for (auto it = inverseTransitionRelation[j].begin();it!=inverseTransitionRelation[j].end();it++) {
                                    if (states[it->first].isActive) {
                                        if (it->first!=j) {
                                            // std::cerr << "Even closer look: " << it->first << "\n";
                                            BF inputsLeft = it->second;
                                            for (auto it2 = inverseTransitionRelation[i].begin();it2!=inverseTransitionRelation[i].end();it2++) {
                                                if (rgeRelation.count(std::pair<uint,uint>(it2->first,it->first))>0) inputsLeft &= !it2->second;
                                            }
                                            mismatchFound |= inputsLeft!=parent->mgr.constantFalse();
                                        }
                                    }
                                }
                                if (!mismatchFound) {
                                    rgeRelation.insert(std::pair<uint,uint>(i,j));
                                    if (VERBOSITY_LEVEL>1) std::cerr << "IS: " << i << "," << j << "\n";
                                }
                            }
                        }
                    }
                }
            }

            // (3) Merge equi-reachable states
            for (auto it = rgeRelation.begin();it!=rgeRelation.end();it++) {
                if ((!states[it->first].isAccepting) && (!states[it->second].isAccepting)) {
                    if (it->first != it->second) {
                        if (rgeRelation.count(std::pair<uint,uint>(it->second,it->first))>0) {
                            if (states[it->first].isActive && states[it->second].isActive) {
                                if (reachabilityRelation.count(std::pair<uint,uint>(it->second,it->first))==0) {

                                    // Check that we would not introduce a loop if we merged the states. Example:
                                    // 2->4, 2->15, 4->15
                                    // If 2 and 15 are equi-reachable, we can still not merge to into 15, as we would end up with:
                                    // 4->15, 15->4
                                    bool isMergeable = true;
                                    for (auto it2 = states[it->first].transitions.begin();it2!=states[it->first].transitions.end();it2++) {
                                        if (it2->first!=it->first) // Disregard self-loops - They are OK.
                                            isMergeable &= (reachabilityRelation.count(std::pair<uint,uint>(it2->first,it->first))==0);
                                    }

                                    if ((isMergeable) && (states[it->second].isAccepting == states[it->first].isAccepting)) {

                                        if (VERBOSITY_LEVEL>1) std::cerr << "Merge " << it->first << " into " << it->second << "\n";
                                        // Merge everything to the second state!
                                        states[it->second].isInitial |= states[it->first].isInitial;

                                        // Copy incoming transitions
                                        for (uint j=0;j<states.size();j++) {
                                            if (states[j].isActive) {
                                                for (auto it2 = states[j].transitions.begin();it2!=states[j].transitions.end();it2++) {
                                                    if (it2->first == it->first) {
                                                        it2->first = it->second;
                                                    }
                                                }
                                            }
                                        }

                                        // Copy outgoing transitions
                                        for (auto it2 = states[it->first].transitions.begin();it2!=states[it->first].transitions.end();it2++) {
                                            states[it->second].transitions.push_back(*it2);
                                        }

                                        // Kill old state
                                        states[it->first].isActive = false;

                                        // Update reachability relation - There might be new reachable states.
                                        reachabilityRelation.replaceAllReferencesOfBy(it->first,it->second);
                                        oldSizeReachabilityRelation = reachabilityRelation.size()+1;
                                        while (oldSizeReachabilityRelation != reachabilityRelation.size()) {
                                            oldSizeReachabilityRelation = reachabilityRelation.size();
                                            // The following code only works for std::set<std::pair<uint,uint> > replacements with iterators that are stable under insertion
                                            for (auto it = reachabilityRelation.begin();it!=reachabilityRelation.end();it++) {
                                                for (auto it2 = states[it->second].transitions.begin();it2!=states[it->second].transitions.end();it2++) {
                                                    reachabilityRelation.insert(std::pair<uint,uint>(it->first,it2->first));
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }


        //========================================================================
        // Here we do a simulation-based minimisation, where we only consider single
        // states and not sets of states.
        //
        // We start with the following simulation relation. For two states a and b, we say that
        // L(a) >= L(b) (i.e., the accepted words from a are also accepted from b) iff:
        //
        // forall input, succ1: succ1 in delta(a,input) => U or V with:
        //   U = exist succ2: succ2 in delta(b,input) and L(succ1) >= L(succ2)
        //   V = succ1 = a and input is self-looping for both a and b, and a is not final.
        //
        // This relation only makes sense if final (rejecting) states in the UVW have
        // no outgoing edges yet, so directly after the parsing from the vermicellis!
        // For these final states, we pre-compute the L>= relation
        //
        // Then, we check foe equi-reachable states. We say that R(a) >= R(b)
        // (meaning that a is reachable along the paths to b) iff:
        // - for every m,i, for an i-labelled edge m->b, there exists an i-labelled edge n->a such that
        //   R(n)>=R(m)h
        // - The self-loops that b has, are also loops for a
        // - if b is initial, then a is also initial
        //
        void simpleSimulationBasedMinimisation(std::vector<BF> &variablesToCareAboutWhenPrinting) {

            (void)variablesToCareAboutWhenPrinting;

            // Perform multiple passes. Reason: The "lgeRelation" isn't up-to-date after
            // the forward-equivalent state matching anymore. Thus, the rule in pass 2 (for every states
            // a and b, if whenever a is reached, also b is reached, and the language of a is greater than
            // the one of b, remove b) needs a re-computation of the lgeRelation

            for (uint pass=0;pass<2;pass++) {

                // Start by computing the L>= relation
                UintPairSetWithMaximum lgeRelation(states.size());

                // 0. Compute L>= relation of the vermicelli end components
                for (uint i=0;i<states.size();i++) {
                    lgeRelation.insert(std::pair<uint,uint>(i,i));
                    if (states[i].isAccepting) {
                        for (uint j=0;j<states.size();j++) {
                            if (states[j].isAccepting) {
                                    bool process = true;
                                if (pass==0) {
                                    assert(states[i].transitions.size()==1);
                                    assert(states[i].transitions[0].first == i);
                                    assert(states[j].transitions.size()==1);
                                    assert(states[j].transitions[0].first == j);
                                } else if ((pass==1) && ((states[i].transitions.size()!=1) || (states[i].transitions[0].first != i) || (states[j].transitions.size()!=1) || (states[j].transitions[0].first != j))) {
                                    // The first pass might have converted some rejecting end states to also have other outgoing edges. Ignore these for the second pass.
                                    process = false;
                                }
                                if (process) {
                                    if (states[i].transitions[0].second <= states[j].transitions[0].second) {
                                        if (VERBOSITY_LEVEL>1) std::cerr << i << ">=: " << j << "\n";
                                        lgeRelation.insert(std::pair<uint,uint>(i,j));
                                    }
                                }
                            }
                        }
                    }
                }

                // Saturate the "L>=" relation
                uint oldSizeLgeRelation = 0;
                while (oldSizeLgeRelation != lgeRelation.size()) {
                    oldSizeLgeRelation = lgeRelation.size();

                    // Update L>= relation
                    for (uint i=0;i<states.size();i++) {
                        for (uint j=0;j<states.size();j++) {
                            if (lgeRelation.count(std::pair<uint,uint>(i,j))==0) {
                                // Check if L(i) >= L(j)
                                // For this, iterate over the successors and compute the
                                // respective inputs
                                std::vector<std::pair<uint,BF> > &transitionsI = states[i].transitions;
                                std::vector<std::pair<uint,BF> > &transitionsJ = states[j].transitions;
                                bool outerFound = true;
                                for (auto it = transitionsI.begin();it!=transitionsI.end();it++) {
                                    uint succ1 = it->first;
                                    BF inputsToFindLeft = it->second;

                                    for (auto it2 = transitionsJ.begin();it2!=transitionsJ.end();it2++) {
                                        if (lgeRelation.count(std::pair<uint,uint>(succ1,it2->first))>0) {
                                            inputsToFindLeft &= !it2->second;
                                        } else if ((it2->first==j) && (succ1==i)) {
                                            // Reduce by non-accepting self-loops.
                                            if (!states[i].isAccepting) {
                                                inputsToFindLeft &= !it2->second;
                                            }
                                        }
                                    }
                                    outerFound &= (inputsToFindLeft == parent->mgr.constantFalse());
                                }
                                if (outerFound) {
                                    lgeRelation.insert(std::pair<uint,uint>(i,j));
                                    if (VERBOSITY_LEVEL>1) std::cerr << "N:" << i << ">=: " << j << std::endl;
                                }
                            }
                        }
                    }
                }

                // Compute reachability relation
                UintPairSetWithMaximum reachabilityRelation(states.size());
                for (uint i=0;i<states.size();i++) reachabilityRelation.insert(std::pair<uint,uint>(i,i));
                uint oldSizeReachabilityRelation = 0;
                while (oldSizeReachabilityRelation != reachabilityRelation.size()) {
                    oldSizeReachabilityRelation = reachabilityRelation.size();
                    // The following code only works for std::set<std::pair<uint,uint> > replacements with iterators that are stable under insertion
                    for (auto it = reachabilityRelation.begin();it!=reachabilityRelation.end();it++) {
                        for (auto it2 = states[it->second].transitions.begin();it2!=states[it->second].transitions.end();it2++) {
                            if (it2->second!=parent->mgr.constantFalse()) reachabilityRelation.insert(std::pair<uint,uint>(it->first,it2->first));
                        }
                    }
                }

                if (pass==0) {
                    // Merge a few states based on simulation equivalence
                    for (auto it = lgeRelation.begin();it!=lgeRelation.end();it++) {
                        if (VERBOSITY_LEVEL>1) std::cerr << "Working on " << it->first << "," << it->second;
                        if (it->first!=it->second) {
                            if (VERBOSITY_LEVEL>1) std::cerr << ".";
                            if (lgeRelation.count(std::pair<uint,uint>(it->second,it->first))>0) {
                                if (VERBOSITY_LEVEL>1) std::cerr << ".";
                                if (((reachabilityRelation.count(*it)>0)) || ((VERBOSITY_LEVEL>1) && (reachabilityRelation.count(std::pair<uint,uint>(it->second,it->first))==0))) {
                                    if (VERBOSITY_LEVEL>1) std::cerr << ".";
                                    if ((states[it->first].isActive) && (states[it->second].isActive)) {
                                        if (VERBOSITY_LEVEL>1) std::cerr << ".";
                                        states[it->first].isActive = false;
                                        if (VERBOSITY_LEVEL>1) std::cerr << "Moving " << it->first << " into " << it->second << std::endl;
                                        states[it->second].isInitial |= states[it->first].isInitial;
                                        for (uint i=0;i<states.size();i++) {
                                            for (auto it2 = states[i].transitions.begin();it2!=states[i].transitions.end();it2++) {
                                                if (it2->first==it->first) {
                                                    it2->first = it->second;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        if (VERBOSITY_LEVEL>1) std::cerr << std::endl;
                    }

                    // Recompute reachability relation after merging
                    reachabilityRelation.clear();
                    for (uint i=0;i<states.size();i++) reachabilityRelation.insert(std::pair<uint,uint>(i,i));
                    oldSizeReachabilityRelation = 0;
                    while (oldSizeReachabilityRelation != reachabilityRelation.size()) {
                        oldSizeReachabilityRelation = reachabilityRelation.size();
                        // The following code only works for std::set<std::pair<uint,uint> > replacements with iterators that are stable under insertion
                        for (auto it = reachabilityRelation.begin();it!=reachabilityRelation.end();it++) {
                            for (auto it2 = states[it->second].transitions.begin();it2!=states[it->second].transitions.end();it2++) {
                                reachabilityRelation.insert(std::pair<uint,uint>(it->first,it2->first));
                            }
                        }
                    }
                }


                // Merge equiv-reachable states: (1) Compute inverse transition relation
                // Compute inverse transition relation
                std::vector<std::vector<std::pair<uint,BF> > > inverseTransitionRelation;
                inverseTransitionRelation.resize(states.size());
                for (uint i=0;i<states.size();i++) {
                    for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                        uint accessIndex = it->first;
                        inverseTransitionRelation[accessIndex].push_back(std::pair<uint,BF>(i,it->second));
                    }
                }

                // (2) Compute initial states
                UintPairSetWithMaximum rgeRelation(states.size());
                for (uint i=0;i<states.size();i++) {
                    if (states[i].isActive) rgeRelation.insert(std::pair<uint,uint>(i,i));
                }
                uint oldSizeRgeRelation = 0;
                while (oldSizeRgeRelation != rgeRelation.size()) {
                    oldSizeRgeRelation = rgeRelation.size();

                    // Update L>= relation
                    for (uint i=0;i<states.size();i++) {
                        for (uint j=0;j<states.size();j++) {
                            // std::cerr << "State " << j << ", isActive: " << states[j].isActive << "\n";
                            if (states[i].isActive && states[j].isActive && (states[i].isInitial || !states[j].isInitial)) {

                                // Check proper self-looping
                                // std::cerr << "Initial IS look: " << i << "," << j << std::endl;
                                BF selfLoops = parent->mgr.constantFalse();
                                for (auto it = states[j].transitions.begin();it!=states[j].transitions.end();it++) {
                                    if (it->first == j) selfLoops |= it->second;
                                }
                                for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                                    if (it->first == i) selfLoops &= !it->second;
                                }

                                if (selfLoops==parent->mgr.constantFalse()) {
                                    // std::cerr << "Closer IS look: " << i << "," << j << std::endl;
                                    // Check incoming edges
                                    bool mismatchFound = false;
                                    for (auto it = inverseTransitionRelation[j].begin();it!=inverseTransitionRelation[j].end();it++) {
                                        if (states[it->first].isActive) {
                                            if (it->first!=j) {
                                                // std::cerr << "Even closer look: " << it->first << "\n";
                                                BF inputsLeft = it->second;
                                                for (auto it2 = inverseTransitionRelation[i].begin();it2!=inverseTransitionRelation[i].end();it2++) {
                                                    if (rgeRelation.count(std::pair<uint,uint>(it2->first,it->first))>0) inputsLeft &= !it2->second;
                                                }
                                                mismatchFound |= inputsLeft!=parent->mgr.constantFalse();
                                            }
                                        }
                                    }
                                    if (!mismatchFound) {
                                        rgeRelation.insert(std::pair<uint,uint>(i,j));
                                        if (VERBOSITY_LEVEL>1) std::cerr << "IS: " << i << "," << j << "\n";
                                    }
                                }
                            }
                        }
                    }
                }

                if (pass==0) {

                    // (3) Merge equi-reachable states
                    for (auto it = rgeRelation.begin();it!=rgeRelation.end();it++) {
                        if (it->first != it->second) {
                            if (rgeRelation.count(std::pair<uint,uint>(it->second,it->first))>0) {
                                if (states[it->first].isActive && states[it->second].isActive) {
                                    if (reachabilityRelation.count(std::pair<uint,uint>(it->second,it->first))==0) {

                                        // Check that we would not introduce a loop if we merged the states. Example:
                                        // 2->4, 2->15, 4->15
                                        // If 2 and 15 are equi-reachable, we can still not merge to into 15, as we would end up with:
                                        // 4->15, 15->4
                                        bool isMergeable = true;
                                        for (auto it2 = states[it->first].transitions.begin();it2!=states[it->first].transitions.end();it2++) {
                                            if (it2->first!=it->first) // Disregard self-loops - They are OK.
                                                isMergeable &= (reachabilityRelation.count(std::pair<uint,uint>(it2->first,it->first))==0);
                                        }

                                        if ((isMergeable) && (states[it->second].isAccepting == states[it->first].isAccepting)) {

                                            if (VERBOSITY_LEVEL>1) std::cerr << "Merge " << it->first << " into " << it->second << "\n";
                                            // Merge everything to the second state!
                                            states[it->second].isInitial |= states[it->first].isInitial;

                                            // Copy incoming transitions
                                            for (uint j=0;j<states.size();j++) {
                                                if (states[j].isActive) {
                                                    for (auto it2 = states[j].transitions.begin();it2!=states[j].transitions.end();it2++) {
                                                        if (it2->first == it->first) {
                                                            it2->first = it->second;
                                                        }
                                                    }
                                                }
                                            }

                                            // Copy outgoing transitions
                                            for (auto it2 = states[it->first].transitions.begin();it2!=states[it->first].transitions.end();it2++) {
                                                states[it->second].transitions.push_back(*it2);
                                            }

                                            // Kill old state
                                            states[it->first].isActive = false;

                                            // Update reachability relation - There might be new reachable states.
                                            reachabilityRelation.replaceAllReferencesOfBy(it->first,it->second);
                                            oldSizeReachabilityRelation = reachabilityRelation.size()+1;
                                            while (oldSizeReachabilityRelation != reachabilityRelation.size()) {
                                                oldSizeReachabilityRelation = reachabilityRelation.size();
                                                // The following code only works for std::set<std::pair<uint,uint> > replacements with iterators that are stable under insertion
                                                for (auto it = reachabilityRelation.begin();it!=reachabilityRelation.end();it++) {
                                                    for (auto it2 = states[it->second].transitions.begin();it2!=states[it->second].transitions.end();it2++) {
                                                        reachabilityRelation.insert(std::pair<uint,uint>(it->first,it2->first));
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }

                if (pass==1) {
                    // Remove states for which there exists another state with >= reachability, but <= language
                    for (uint i=0;i<states.size();i++) {
                        for (uint j=0;j<states.size();j++) {
                            if (states[i].isActive) {
                                if (states[j].isActive) {
                                    if ((i!=j) && ((j!=39) || (i==60))) {
                                        if ((rgeRelation.count(std::pair<uint,uint>(j,i))>0) && (lgeRelation.count(std::pair<uint,uint>(i,j))>0)) {
                                            if (reachabilityRelation.count(std::pair<uint,uint>(j,i))==0) {
                                                if ((states[i].isInitial==false) || (states[j].isInitial==true)) {
                                                    if (VERBOSITY_LEVEL>1) std::cerr << "Killing " << i << " in favour of " << j << std::endl;
                                                    states[i].isActive = false;
                                                    for (uint k=0;k<states.size();k++) {
                                                        bool done = false;
                                                        while (!done) {
                                                            done = true;
                                                            for (auto it=states[k].transitions.begin();(it!=states[k].transitions.end()) && done;it++) {
                                                                if (it->first==i) {
                                                                    states[k].transitions.erase(it);
                                                                    done = false;
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                    // Finally, merge transitions with the same goal state
                    for (uint i=0;i<states.size();i++) {
                        if (states[i].isActive) {
                            std::tr1::unordered_map<uint,BF> transitions;
                            for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                                if (transitions.count(it->first)==0) {
                                    transitions[it->first] = it->second;
                                } else {
                                    transitions[it->first] |= it->second;
                                }
                            }
                            states[i].transitions.clear();
                            for (auto it = transitions.begin();it!=transitions.end();it++) {
                                states[i].transitions.push_back(std::pair<uint,BF>(it->first,it->second));
                            }
                        }
                    }
                }

             }
        }

        std::string toParseableString() {
            std::ostringstream os;

            // Write APs
            std::vector<BF> relevantVars;
            os << parent->varsAPs.size() << " ";
            for (uint i=0;i<parent->varsAPs.size();i++) {
                relevantVars.push_back(parent->vars[parent->varsAPs[i]]);
                os << parent->varNames[parent->varsAPs[i]] << " ";
            }

            // Renumber the states such that non-active states are not printed out.
            std::tr1::unordered_map<uint,uint> stateRenumbering;
            for (uint i=0;i<states.size();i++) {
                if (states[i].isActive) {
                    uint nextIndex = stateRenumbering.size();
                    stateRenumbering[i] = nextIndex;
                    std::cerr << "Printing state renumbering: " << i << ": " << nextIndex << std::endl;
                }
            }

            // Write state info
            os << stateRenumbering.size() << " ";
            for (uint i=0;i<states.size();i++) {
                if (states[i].isActive) {
                    os << (states[i].isInitial?"1 ":"0 ");
                    os << (states[i].isAccepting?"1 ":"0 ");
                    os << states[i].transitions.size() << " ";
                    for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                        assert(states[it->first].isActive);
                        assert(stateRenumbering.count(it->first)>0);
                        os << stateRenumbering[it->first] << " ";
                        os << getBFAsTreeString(it->second,relevantVars,0) << " ";
                    }
                }
            }
            os << "END";
            return os.str();
        }

        void removeInactiveStates() {
            // Renumber the states such that non-active states are not printed out.
            std::tr1::unordered_map<uint,uint> stateRenumbering;
            for (uint i=0;i<states.size();i++) {
                if (states[i].isActive) {
                    uint nextIndex = stateRenumbering.size();
                    stateRenumbering[i] = nextIndex;
                }
            }

            std::vector<UVWState> newStates;
            for (uint i=0;i<states.size();i++) {
                if (states[i].isActive) {
                    newStates.push_back(UVWState());
                    newStates.back().isAccepting = states[i].isAccepting;
                    newStates.back().isInitial = states[i].isInitial;
                    newStates.back().isActive = true;
                    for (auto it = states[i].transitions.begin();it!=states[i].transitions.end();it++) {
                        newStates.back().transitions.push_back(std::pair<uint,BF>(stateRenumbering[it->first],it->second));
                    }
                }
            }
            states = newStates;
        }

    };

    BF parseTransitionStringInSPINFormat(std::string what) {

        if (what[what.length()-1]!=']') throw "Error in transition string: String starting with '[' should also end with an ']'.";
        what = what.substr(1,what.length()-2);
        boost::replace_all(what,"-"," ");

        // 1. Check if there are braces around the disjunction (as made by SPOT) and remove them
        int braceLevel = 0;
        bool foundDisjunctionInBraces = false;
        for (uint i=0;i<what.length();i++) {
                switch (what[i]) {
                        case '(':
                                braceLevel += 1;
                                break;
                        case ')':
                                braceLevel -= 1;
                                break;
                        case '|':
                                if (braceLevel>0) foundDisjunctionInBraces = true;
                                break;
                }
                if (braceLevel<0) throw "Error: Unmatched braces in parseTransitionStringInSPINFormat";
        }
        if (braceLevel!=0) throw "Error: Unmatched braces in parseTransitionStringInSPINFormat";
        if (foundDisjunctionInBraces) {
                if ((what[0]!='(') || (what[what.length()-1]!=')')) throw "Error: Too complex braces inparseTransitionStringInSPINFormat";
                what = what.substr(1,what.length()-2);
        }

        // Continue work.
        BF result = mgr.constantFalse();

        // Explode agains "||" operators
        std::vector<std::string> orParts;
        boost::iter_split(orParts, what, boost::first_finder(" || "));

        for (uint i=0;i<orParts.size();i++) {

                // Twiggle with strings
                std::string current = orParts[i];
                if (current[0]=='(') {
                        assert(current[current.length()-1]==')');
                        current = current.substr(1,current.length()-2);
                } else {
                        assert(current[current.length()-1]!=')');
                }

                std::vector<std::string> andParts;
                boost::iter_split(andParts, current, boost::first_finder(" && "));

                BF andPart = mgr.constantTrue();
                for (uint j=0;j<andParts.size();j++) {

                        current = andParts[j];
                        bool negated = current[0]=='!';
                        if (negated) current = current.substr(1,std::string::npos);

                        BF value;

                        if (current=="1") {
                                value = mgr.constantTrue();
                        } else  {
                            // Search for the atomic proposition in the list.
                            bool found=false;
                            for (auto it = varsAPs.begin();it!=varsAPs.end();it++) {
                                if (varNames[*it]==current) {
                                    value = vars[*it];
                                    found = true;
                                }
                            }
                            if (!found) {
                                std::cerr << "Error in parsing the input instance: The atomic proposition that is encoded as " << current << " has not been declared for that vermicelli set.\n Note that this is a tool-chain internal error.";
                                throw "Aborting";
                            }
                        }

                        if (negated) {
                                andPart &= !value;
                        } else {
                                andPart &= value;
                        }
                }
                result |= andPart;
        }

        return result;
    }

    /**
      Parses one of the strings in the Vermicelli tree description that represent a transition condition.
      Two forms are supported: either the one put out by the Rabin2Vermicelli converter, or SPIN-style
      DNF transitions. The latter is handled by a seperate parsing function.
      */
    BF parseTransitionString(std::string transition, std::vector<BF> &relevantVars) {
        if (transition.size()==0) throw "Error: Empty transition string";
        if (transition[0]=='[') return parseTransitionStringInSPINFormat(transition);
        if (transition=="1") return mgr.constantTrue();
        if (transition=="0") return mgr.constantFalse();
        if (transition.length()!=((uint)1 << relevantVars.size())) {
            throw "Error parsing Vermicelli - transition string length is not correct.";
        }
        BF soFar = mgr.constantFalse();
        for (uint i=0;i<transition.length();i++) {
            bool value;
            if (transition[i]=='0') value = false;
            else if (transition[i]=='1') value = true;
            else throw "Error parsing Vermicelli - transition string contains illegal characters";
            if (value) {
                BF thisOne = mgr.constantTrue();
                for (uint j=0;j<relevantVars.size();j++) {
                    thisOne &= (((i >> j)&1)!=0)?relevantVars[j]:!relevantVars[j];
                }
                soFar |= thisOne;
            }
        }
        return soFar;
    }

    std::string computeUVW(std::vector<std::string> &vermicelliSets) {

        std::vector<UVW> uvws;

        for (std::vector<std::string>::iterator it = vermicelliSets.begin();it!=vermicelliSets.end();it++) {

            // Parse variable names
            std::istringstream is(*it);
            int size;
            is >> size;
            if (is.fail()) {
                std::cerr << "Error parsing vermicelli (1)\n";
                throw "Aborting.";
            }
            std::vector<BF> varsUsedInThisSet;
            for (int i=0;i<size;i++) {
                std::string s;
                is >> s;
                if (is.fail()) {
                    std::cerr << "Error parsing vermicelli (2)\n";
                    throw "Aborting.";
                }
                bool found = false;
                for (uint j=0;j<varNames.size();j++) {
                    if (varNames[j]==s) {
                        found = true;
                        varsUsedInThisSet.push_back(vars[j]);
                    }
                }
                if (!found) {
                    vars.push_back(mgr.newVariable());
                    varNames.push_back(s);
                    varsAPs.push_back(vars.size()-1);
                    varsUsedInThisSet.push_back(vars.back());
                }
            }

            // Parse Vermicelli trees for the individual states in the automaton
            uint nofEndComponents;
            is >> nofEndComponents;
            if (is.fail()) {
                std::cerr << "Error parsing vermicelli (3)\n";
                throw "Aborting.";
            }

            uvws.push_back(UVW(this));

            for (uint i=0;i<nofEndComponents;i++) {
                std::string matchingString;
                is >> matchingString;
                if (is.fail()) {
                    std::cerr << "Error parsing vermicelli (4)\n";
                    throw "Aborting.";
                }
                BF selfLoop = parseTransitionString(matchingString,varsUsedInThisSet);
                uint vermicelliEndState = uvws.back().getSelfLoopingFinalState(selfLoop);

                uint initial = uvws.back().recurseParse(is,varsUsedInThisSet,std::vector<uint>(),mgr.constantFalse(),vermicelliEndState);
                uvws.back().states[initial].isInitial = true;
                if (uvws.back().states[initial].isAccepting) {
                    uvws.back().states[initial].isAccepting = false;
                    uvws.back().states[vermicelliEndState].isInitial = true;
                }
            }
            std::cerr << *it << ":\n";
        }

        std::vector<BF> varsToCareAbout;
        for (auto it = varsAPs.begin();it!=varsAPs.end();it++) {
            varsToCareAbout.push_back(vars[*it]);
        }

        // Optimize the individual UVWs
        for (auto it = uvws.begin();it!=uvws.end();it++) {
            it->onlyNonFinalAndNonFinalStateMergingSimpleSimulationBasedMinimisation(varsToCareAbout);
            it->removeInactiveStates();
        }

        // Merge the uvws
        UVW master(uvws);
        master.joinOnlySelfLoopingStatesWithSameLabelSet(mgr);
        master.removeInactiveStates();
        master.simpleSimulationBasedMinimisation(varsToCareAbout);
        master.removeInactiveStates();

        return master.toParseableString();
    }

    bool processXMLTree(xmlNode *root) { // returns true if the root shall be deleted.
        bool shallBeDeleted = false;
        std::vector<xmlNode*> toRemove;
        if (root->type == XML_ELEMENT_NODE) {
            xmlChar* propString = xmlGetProp(root,BAD_CAST("type"));
            if ((xmlStrEqual(root->name,BAD_CAST("PropertyGroup"))) && (xmlStrEqual(propString,BAD_CAST("Conjunction")))) {
                // Sum up all the vermicellis
                std::vector<std::string> vermicelliTrees;
                for (xmlNode *it2 = root->children; it2!=NULL; it2 = it2->next) {
                    if ((it2->type == XML_ELEMENT_NODE) && (xmlStrEqual(it2->name,BAD_CAST("VermicelliSet")))) {
                        std::ostringstream os;
                        for (xmlNode *it3 = it2->children; it3!=NULL; it3 = it3->next) {
                            if (it3->type == XML_TEXT_NODE) {
                                os << it3->content;
                            }
                        }
                        vermicelliTrees.push_back(os.str());
                        std::cerr << os.str() << "\n";
                        toRemove.push_back(it2);
                    } else {
                        std::cerr << "node type: Element, name: " << it2->name << "\n";
                        if (processXMLTree(it2)) {
                            toRemove.push_back(it2);
                        }
                    }
                }

                // Add new processed UVW
                xmlNewTextChild(root->parent,NULL,BAD_CAST("UVW"),BAD_CAST(computeUVW(vermicelliTrees).c_str()));
                shallBeDeleted = true;
            } else {
                // Process lonesome VermicelliSets
                std::vector<std::string> vermicelliTrees;
                if ((root->type == XML_ELEMENT_NODE) && (xmlStrEqual(root->name,BAD_CAST("VermicelliSet")))) {
                    std::ostringstream os;
                    for (xmlNode *it3 = root->children; it3!=NULL; it3 = it3->next) {
                        if (it3->type == XML_TEXT_NODE) {
                            os << it3->content;
                        }
                    }
                    vermicelliTrees.push_back(os.str());
                    std::cerr << os.str() << "\n";
                    xmlNewTextChild(root->parent,NULL,BAD_CAST("UVW"),BAD_CAST(computeUVW(vermicelliTrees).c_str()));
                    shallBeDeleted = true;
                }

                for (xmlNode *it2 = root->children; it2!=NULL;it2 = it2->next) {
                    if (processXMLTree(it2)) {
                        toRemove.push_back(it2);
                    }
                }
            }
            xmlFree(propString);
        }
        for (std::vector<xmlNode*>::iterator it = toRemove.begin();it!=toRemove.end();it++) {
            xmlUnlinkNode(*it);
            xmlFreeNode(*it);
        }
        return shallBeDeleted;
    }

public:

    //==================================================================
    // Setting up functions
    //==================================================================
    ConverterAlgorithm(bool singleUVW, const char *inputFilename) : mgr(4096/*,1.05*/) {

        if (!singleUVW) {
            xmlDoc *xmlDoc = xmlReadFile(inputFilename, NULL, 0);
            if (xmlDoc==NULL) {
                std::cerr << "Error reading XML file: " << inputFilename << "\n";
                throw "Aborting.";
            }

            // Process all Vermicelli
            xmlNode *root = xmlDocGetRootElement(xmlDoc);
            for (xmlNode *it = root; it!=NULL; it = it->next) {
                processXMLTree(it);
            }

            std::cout.flush();
            if (xmlDocDump(stdout,xmlDoc)==-1) {
                std::cerr << "Error dumping resulting XML document.";
                throw "Aborting.";
            }

            // Done.
            xmlFreeDoc(xmlDoc);
        } else {
            // Parse only a single UVW
            std::ifstream is(inputFilename);
            std::string vermicelli;
            std::getline(is,vermicelli);
            if (is.fail()) {
                std::cerr << "Error reading UVW file.\n";
                throw "Aborting.";
            }
            is.close();
            std::vector<std::string> vermicellis;
            vermicellis.push_back(vermicelli);
            std::cout << computeUVW(vermicellis);
        }

    }

    //==================================================================
    // BDD DOT dumping helper functions
    //==================================================================
    virtual void getVariableTypes(std::vector<std::string> &types) const {
        types.push_back("APs");
    }

    virtual void getVariableNumbersOfType(std::string type, std::vector<uint> &nums) const {
        std::vector<BF> const *myvector;
        if (false) {}
        else {
            if (type=="APs") {
                nums.insert(nums.end(),varsAPs.begin(),varsAPs.end());
                return;
            }
            // No hit
            std::cerr << "Error: Could not identify variable type in virtual void getVariableNumbersOfType(std::string type, std::vector<uint> &nums) const:\n" << type << "\n";
            throw "Terminating.";
        }
        for (std::vector<BF>::const_iterator it = myvector->begin();it!=myvector->end();it++) {
            BF thisOne = *it;
            for (uint i=0;i<vars.size();i++) {
                if (thisOne==vars[i]) nums.push_back(i);
            }
        }
    }

    virtual BF getVariableBF(uint number) const {
        return vars[number];
    }

    virtual std::string getVariableName(uint number) const {
        return varNames[number];
    }
};

void showHelp() {
    std::cerr << "Usage: Either: \n   converter --single vermicelliFile\n  converter specFile.xml\n";
}


int main(int argN, const char **args) {
    try {
        if (argN<2) {
            std::cerr << "Error: No input file given!\n";
            return 100;
        }

        std::string filename;
        bool singleUVW = false;

        if (args[1][0]=='-') {
            if (std::string(args[1])=="--single") {
                singleUVW = true;
                if (argN<3) {
                    showHelp();
                    return 100;
                }
                filename = args[2];
            } else {
                showHelp();
                return 100;
            }
        } else {
            filename = args[1];
        }

        LIBXML_TEST_VERSION

        ConverterAlgorithm algorithm(singleUVW,filename.c_str());

        xmlCleanupParser();

        return 0;
    }  catch (const char *c) {
        std::cerr << "Error: " << c << "\n";
        return 100;
    } catch (int i) {
        std::cerr << "Integer error: " << i<< "\n";
        return 100;
    } catch (BFDumpDotException e) {
        std::cerr << "BFDumpDotException: " << e.getMessage()<< "\n";
        return 100;
    }
}
