#!/bin/bash
# This script downloads the libraries that do not come with Bassist and invokes the make process for all libraries and
# tools. Every action performed only affects the directory this script is started from. No system-level alterations will
# be performed.

echo "###############################################"
echo "# Downloading the necessary libraries for the #"
echo "# 'Bassist' synthesis framework               #"
echo "###############################################"

cd LibraryDownloads || (echo "ERROR: Could not switch into the 'LibraryDownloads' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)

if [ -s cudd-2.4.2.tar.gz ]
then
    echo "NOTE: CuDD archive has been downloaded before and is thus not downloaded again."
else
    wget ftp://vlsi.colorado.edu/pub/cudd-2.4.2.tar.gz || (echo "ERROR: Could not download the CUDD 2.4.2 library. Please try to download it manually and save it into the 'library_download' directory"; rm cudd-2.4.2.tar.gz; exit 1)
fi

if [ -s ltl2ba-1.1.tar.gz ]
then
    echo "NOTE: ltl2ba archive has been downloaded before and is thus not downloaded again."
else
    wget http://www.lsv.ens-cachan.fr/~gastin/ltl2ba/ltl2ba-1.1.tar.gz || (echo "ERROR: Could not download the ltl2ba 1.1 LTL-to-Buechi-automaton converter. Please try to download it manually and save it into the 'library_download' directory"; rm ltl2ba-1.1.tar.gz; exit 1)
fi

if [ -s ltl2dstar-0.5.1.tar.gz ]
then
    echo "NOTE: ltl2dstar archive has been downloaded before and is thus not downloaded again."
else
    wget http://www.ltl2dstar.de/down/ltl2dstar-0.5.1.tar.gz || (echo "ERROR: Could not download the ltl2dstar 0.5.1 LTL-to-Rabin-automaton converter. Please try to download it manually and save it into the 'library_download' directory"; rm tl2dstar-0.5.1.tar.gz; exit 1)
fi

cd .. 

echo ""
echo "###############################################"
echo "# Checking if 'picosat' is present and works. #"
echo "###############################################"

echo "p cnf 2 2 
1 -1 0
2 -2 0" | picosat
if [ "$?" -ne "10" ]; then
  echo "Error: Could not find a working copy of picosat. Ubuntu/Debian users can simply install the 'picosat' package. Others need to install picosat manually and make sure that it is in the PATH environment variable."
  exit 1
fi
echo "Result: Picosat if present and works fine."

echo ""
echo "###############################################"
echo "# Extracting and compiling CuDD               #"
echo "###############################################"

echo "Extracting CuDD..."
cd Tools || (echo "ERROR: Could not switch into the 'Tools' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)
tar xzf ../LibraryDownloads/cudd-2.4.2.tar.gz 

echo "Applying patch to Makefile to automatically detect 64-bit architectures..."
patch -p0 < BFAbstractionLibrary/CuddMakefile.patch || (echo "ERROR: Could not patch the CuDD Makefile"; exit 1)

echo "Compiling CUDD..."
cd cudd-2.4.2
make || (echo "ERROR: Cannot compile cudd. Please check the error message and see if packages are missing, etc."; exit 1)
cd ../..

echo ""
echo "###############################################"
echo "# Extracting and compiling ltl2ba             #"
echo "###############################################"

echo "Extracting ltl2ba..."
cd Tools || (echo "ERROR: Could not switch into the 'Tools' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)
tar xzf ../LibraryDownloads/ltl2ba-1.1.tar.gz

echo "Compiling ltl2ba"
cd ltl2ba-1.1
make || (echo "ERROR: Cannot compile ltl2ba. Please check the error message and see if packages are missing, etc."; exit 1)
cd ../..

echo ""
echo "###############################################"
echo "# Extracting and compiling ltl2dstar          #"
echo "###############################################"

echo "Extracting ltl2dstar..."
cd Tools || (echo "ERROR: Could not switch into the 'Tools' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)
tar xzf ../LibraryDownloads/ltl2dstar-0.5.1.tar.gz 

echo "Compiling ltl2dstar"
cd ltl2dstar-0.5.1/src
echo "Note: I'm deleting the boost copy that comes along with ltl2dstar, as it doesn't work on later GCC versions anymore. This requires a current version of boost to be installed."
rm -rf boost
make || (echo "ERROR: Cannot compile ltl2dstar. Please check the error message and see if packages are missing, etc."; exit 1)
cd ../../..

echo ""
echo "###############################################"
echo "# Compiling Bassist part: BDDBackend          #"
echo "###############################################"

cd Tools/BDDBackend || (echo "ERROR: Could not switch into the 'Tools/BDDBackend' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)
qmake Backend.pro || (echo "ERROR: Cannot run 'qmake'. Perhaps the qmake tool is not installed?"; exit 1)
make || (echo "ERROR: Cannot compile the BDD Backend. Please check the error message and see if packages are missing, etc."; exit 1)
cd ../..

echo ""
echo "###############################################"
echo "# Compiling Bassist part: DBABuilder          #"
echo "###############################################"

cd Tools/DBABuilder || (echo "ERROR: Could not switch into the 'Tools/DBABuilder' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)
make || (echo "ERROR: Cannot compile the DBABuilder. Please check the error message and see if packages are missing, etc."; exit 1)
cd ../..

echo ""
echo "###############################################"
echo "# Compiling Bassist part: DBAtoVermicelliSet  #"
echo "###############################################"

cd Tools/DBAtoVermicelliSet || (echo "ERROR: Could not switch into the 'Tools/DBAtoVermicelliSet' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)
qmake Converter.pro || (echo "ERROR: Cannot run 'qmake'. Perhaps the qmake tool is not installed?"; exit 1)
make || (echo "ERROR: Cannot compile the DBAtoVermicelliSet tool. Please check the error message and see if packages are missing, etc."; exit 1)
cd ../..

echo ""
echo "###############################################"
echo "# Compiling Bassist part: InIntersectionTester#"
echo "###############################################"

cd Tools/InIntersectionTester || (echo "ERROR: Could not switch into the 'Tools/InIntersectionTester' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)
qmake Tester.pro || (echo "ERROR: Cannot run 'qmake'. Perhaps the qmake tool is not installed?"; exit 1)
make || (echo "ERROR: Cannot compile the InIntersectionTester tool. Please check the error message and see if packages are missing, etc."; exit 1)
cd ../..

echo ""
echo "###############################################"
echo "# Compiling Bassist part: VermicelliSetsToUVW #"
echo "###############################################"

cd Tools/VermicelliSetsToUVW || (echo "ERROR: Could not switch into the 'Tools/VermicelliSetsToUVW' directory. Please ensure that you start this script from your Bassist installation directory."; exit 1)
qmake Converter.pro || (echo "ERROR: Cannot run 'qmake'. Perhaps the qmake tool is not installed?"; exit 1)
make || (echo "ERROR: Cannot compile the VermicelliSetsToUVW tool. Please check the error message and see if packages are missing, etc."; exit 1)
cd ../..

echo ""
echo "###############################################"
echo "# All Bassist parts and libraries have been   #"
echo "# compiled!                                   #"
echo "###############################################"

read -p "Do you want to test the toolset now? (y/n)? " choice
if [ $choice == "y" ]
then
    Tools/synthesize.py Benchmarks/modifiedMutex.xml
    if [ "$?" -ne "0" ]; then
        echo "ERROR: Sorry, some step in the computation didn't work out. Please observe the error messages to find out more. If the error persists, please contact the author of the Bassist toolset."
        exit 1
    else
        echo ""
        echo "###############################################"
        echo "# Test ran successful.                        #"
        echo "###############################################"
        echo ""
        echo "If you want to check if NuSMV can verify the synthesized design, run, e.g., 'Tools/synthesize.py Benchmarks/modifiedMutex.xml /tmp/model.txt' and then '/path/to/NuSMV /tmp/model.txt'"
        exit 0
    fi
else
    echo "Ok, running no test. You can run it yourself by invoking 'Tools/synthesize.py Benchmarks/modifiedMutex.xml'."
fi


